package com.bria.common

import android.Manifest
import android.app.Application
import android.support.v4.os.UserManagerCompat
import com.bria.common.analytics.Analytics
import com.bria.common.controller.Controller
import com.bria.common.controller.ControllerObservable
import com.bria.common.controller.settings.core.Settings
import com.bria.common.modules.NoncriticalComponentsInitializer
import com.bria.common.permission.PermissionHandler
import com.bria.common.permission.PhoneStatePermissionObservable
import com.bria.common.sdkwrapper.SipPhoneAndroidManager
import com.bria.common.sdkwrapper.SipStackManager
import com.bria.common.sdkwrapper.SipStackManagerInstanceObservable
import com.bria.common.util.Log
import com.bria.common.util.Utils
import com.counterpath.sdk.android.SipPhoneAndroid
import kotlin.concurrent.thread

object ControllerAndSipStackInitializer {
    private val TAG = "ControllerAndSipStackInitializer"

    private var mSipPhoneAndroid: SipPhoneAndroid? = null
    private var mSipStackManager: SipStackManager? = null

    fun initialize(application: Application) {
        // check if "files" and "lib" directories exist
        Log.i(TAG, "isUserUnlocked: " + UserManagerCompat.isUserUnlocked(application))
        val isValidInstallation = Utils.Build.isValidInstallation(application)
        if (!isValidInstallation) {
            Log.e(TAG, "Installation is not valid, not proceeding with service create")
            Analytics.contextSend(application, com.bria.common.analytics.Constants.Events.APP_INVALID_INSTALL)
            return
        }

        Log.i(TAG, "Installation is valid")


        thread(
                isDaemon = true,
                name = "Controller-init-thread") {

            val controller = Controller.create(application)
            Log.d(TAG, "initializeFramework: Controller available")

            ControllerObservable.notifyNewValue(controller)
        }

        thread(
                isDaemon = true,
                name = "SDK-init-thread") {

            // SDK has its own logic for fetching device id, so wait for phone state permission

            Log.i(TAG, "initializeFramework: waiting on PHONE_STATE permission")

            // maybe we already have it
            val phoneStateAccepted = PermissionHandler.checkPermission(application, Manifest.permission.READ_PHONE_STATE)
            PhoneStatePermissionObservable.setNewValue(phoneStateAccepted)

            PhoneStatePermissionObservable
                    .getObservable()
                // wait for phone state
                .filter { x -> x }
                .blockingFirst()

            Log.i(TAG, "initializeFramework: PHONE_STATE permission granted.")

            mSipPhoneAndroid = SipPhoneAndroidManager.create(application, Settings.get(application))

            Log.d(TAG, "initializeFramework: SDK available")

            val controller = ControllerObservable.getObservable().blockingFirst()

            mSipStackManager = SipStackManager(application, controller, mSipPhoneAndroid!!)
            Log.d(TAG, "initializeFramework: SipStackManager initialized")

            controller.onStackManagerInitialized(mSipStackManager!!)

            SipStackManagerInstanceObservable.notifyNewValue(mSipStackManager!!)

            // todo: move this out
            NoncriticalComponentsInitializer.initialize(application, controller, mSipPhoneAndroid!!)
        }
    }

    fun destroy() {
        try {
            // destroy sip stack manager
            if (mSipStackManager != null) {
                Log.d(TAG, "die: Destroying SipStackManager")
                mSipStackManager?.stop()
                Log.d(TAG, "die: Destroyed SipStackManager")
            }

            // destroy the SDK
            if (mSipPhoneAndroid != null) {
                Log.d(TAG, "die: Destroying SDK")
                SipPhoneAndroidManager.destroy(mSipPhoneAndroid!!)
                Log.d(TAG, "die: Destroyed SDK")
            }
        } catch (throwable: Throwable) {
            Log.e(TAG, "Non fatal error during shutdown.", throwable)
        }

    }
}
