package com.bria.common.controller.collaboration.rx.api

import com.counterpath.sdk.SipVccsAccountApi.CrackVCCSURLResult
import com.counterpath.sdk.pb.VccsAccount.VccsAccountSettings
import com.counterpath.sdk.pb.VccsAccount.VccsAccountStateChangedEvent
import io.reactivex.Observable

/**
 * Operations allowed with VCCS/Collab accounts. In most cases, account is equal to a configurable websocket.
 */
interface CollabAccountsSdkApi {

    companion object {
        /**
         * SDK-defined generic error code
         */
        const val K_ERROR = -0x7fffffff

        /**
         * SDK-defined generic success code
         */
        const val K_SUCCESS = 0
    }

    /**
     * Creates a new account.
     * @return ID of the account just created, or [K_ERROR] if something fails
     */
    fun createAccount(): Int

    /**
     * Enables a disabled account.
     * @param id ID of the account you want to enable
     * @param immediate Whether to run 'enable' action immediately or to wait for subscription (default = `false`)
     * @return An observable to track account state changes. Always a new instance
     */
    fun enableAccount(id: Int, immediate: Boolean = false): Observable<VccsAccountStateChangedEvent>

    /**
     * Disables an enabled account.
     * @param id ID of the account you want to disable
     * @param immediate Whether to run 'disable' action immediately or to wait for subscription (default = `false`)
     * @return An observable to track account state changes. Always a new instance
     */
    fun disableAccount(id: Int, immediate: Boolean = false): Observable<VccsAccountStateChangedEvent>

    /**
     * Deletes an account.
     * @param id ID of the account you want to delete
     * @return `True` if account was deleted, `false` otherwise
     */
    fun deleteAccount(id: Int): Boolean

    /**
     * Updates an account using the given settings. For this to work, you need to [disableAccount] first, then [enableAccount] again afterwards.
     * @param id ID of the account you want to update
     * @param settings Settings to apply to the account
     * @return `True` if account was updated, `false` otherwise
     */
    fun updateAccount(id: Int, settings: VccsAccountSettings): Boolean

    /**
     * Parses the given URL to get collaboration data from it.
     * @param url The URL in String form
     * @return A parsed [CrackVCCSURLResult] instance, or `null` if URL was not collaboration-related
     */
    fun parseUrl(url: String): CrackVCCSURLResult?

    /**
     * Observes all changes on all available accounts.
     * @return An observable to track all account changes. Always a new instance
     */
    fun observeAccounts(): Observable<VccsAccountStateChangedEvent>

}
