package com.bria.common.controller.collaboration.rx.api

import android.content.Context
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.EAccountResult
import com.bria.common.controller.collaboration.ICollaborationActions.*
import com.bria.common.controller.collaboration.account.SimpleVccsAccount
import com.bria.common.controller.collaboration.rx.entities.LinkInfo
import com.bria.common.controller.collaboration.rx.entities.SubscriptionRequest
import com.bria.common.controller.collaboration.utils.SimpleVccsConference

/**
 * Organizes collab entities and allows data getters for anything that collab might need from the rest of the system.
 */
interface ConferenceEntityManager {

    /**
     * Creates a new [SimpleVccsConference] object using the given `nullable` link.
     * @param linkInfo If `null`, this will create a 'host mode' conference
     */
    fun createConference(linkInfo: LinkInfo?): SimpleVccsConference

    /**
     * Create a new [SubscriptionRequest] object using for subscription to conference.
     */

    fun createSubscriptionRequest(accountId: Int, bridgeSip: String?, subscriptionCode: String, pinCode: String, participantType: Int): SubscriptionRequest

    /**
     * Creates a new [SimpleVccsAccount] object using the given `nullable` conference.
     * @param conference If `null`, this will create a 'host mode' account
     */
    fun createCollabAccount(conference: SimpleVccsConference?): SimpleVccsAccount

    /**
     * Looks for a primary SIP [Account].
     * @return Either an account, or `null` if no matches were found
     */
    fun findPrimaryAccount(): Account?

    /**
     * Looks for a primary XMPP [Account].
     * @return Either an account, or `null` if no matches were found
     */
    fun getXmppChatAccount(): Account?

    /**
     * Returns the main XMPP account's username, or an empty String if no account is found.
     */
    fun getXmppUserName(): String

    /**
     * Returns the primary account's display name, or an empty String if no account is found.
     */
    fun getDisplayName(): String

    /**
     * Returns a bitmask of participant capabilities, depending on features available.
     * @see PART_CAPABILITY_BRIA
     * @see PART_CAPABILITY_CALL
     * @see PART_CAPABILITY_CHAT
     * @see PART_CAPABILITY_VIDEO
     * @see PART_CAPABILITY_SHARE
     */
    fun getCapabilities(): Int

    /**
     * Returns the primary account's SIP address in format `username@domain`.
     */
    fun getSipAddress(): String

    /**
     * Returns the primary account's SIP username.
     */
    fun getSipUsername(): String

    /**
     * Returns the primary account's domain in format `@domain` (with the '@' sign).
     */
    fun getAtSipDomain(): String

    /**
     * Returns the current device's ID, often wrongly referred to as 'app' ID.
     */
    fun getAppId(context: Context): String

    /**
     * Creates a SIP Account from vccs socket account JSON subscribe response.
     */
    fun createDirectSipAccount(directSipProvisioningJson: String): String?

    /**
     * Remove sipAccount from Account Controller
     */
    fun removeSipAccount(accNickName: String): EAccountResult
}
