package com.bria.common.controller.collaboration.rx.api

import com.bria.common.controller.collaboration.rx.entities.ConferenceConfigFlags
import com.bria.common.controller.collaboration.rx.entities.ConferenceConfigValues
import com.bria.common.controller.collaboration.rx.entities.EventHolder
import com.bria.common.controller.collaboration.rx.entities.SubscriptionRequest
import com.counterpath.sdk.SipVccsConferenceApi
import com.counterpath.sdk.pb.VccsConference.*
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.Single

typealias ParticipantChanges = List<ParticipantStatus>

interface ISipVccsConferenceApi {

    companion object {
        /**
         * SDK-defined generic error code
         */
        const val K_ERROR = -0x7fffffff

        /**
         * SDK-defined generic success code
         */
        const val K_SUCCESS = 0

        /**
         * SDK-defined type for mobile participants
         */
        const val PARTICIPANT_TYPE = SipVccsConferenceApi.ParticipantType_BriaMobile
    }

    /**
     * Queries the conference connection info from the server.
     * @param accountId The account to use
     * @param subscriptionCode The conference's subscription code
     * @return A [Single] to wait for the connection event ([ConferenceConnectionInfoEvent])
     */
    fun getConferenceConnectionInfo(accountId: Int, subscriptionCode: String): Single<EventHolder<ConferenceConnectionInfoEvent>>

    /**
     * Queries the current users' conference list from the server.
     * @param accountId The account to use
     * @return An [Single] to wait for the conferences list ([ConferenceDetails] types)
     */
    fun getConferenceList(accountId: Int): Single<EventHolder<List<ConferenceDetails>>>

    /**
     * Queries the given conference's details.
     * @param accountId The account to use
     * @param conferenceId The conference to query
     * @return A [Single] to wait for conference details ([ConferenceDetails])
     */
    fun getConferenceDetails(accountId: Int, conferenceId: Int): Single<EventHolder<ConferenceDetails>>

    /**
     * Subscribes to the conference with the given conference ID.
     * @param request The request to execute
     * @return A [Single] to wait for subscription details ([SubscribeEvent])
     */
    fun subscribe(request: SubscriptionRequest): Single<EventHolder<SubscribeEvent>>

    /**
     * Unsubscribes from the conference with the given conference bridge number.
     * @param accountId The account to use
     * @param bridgeNumber The conference bridge to unsubscribe from (usually found in conference details)
     * @return A [Completable] to wait for unsubscribe completion
     */
    fun unsubscribe(accountId: Int, bridgeNumber: String): Completable

    /**
     * Locks or unlocks the bridge for new participants.
     * @param accountId The account to use
     * @param conferenceId The conference to lock
     * @param lock Whether to lock or unlock the conference
     * @return A [Completable] to wait for participant lock completion
     */
    fun setParticipantsLock(accountId: Int, conferenceId: Int, lock: Boolean): Completable

    /**
     * Locks or unlocks the bridge mute for all participants in the conference.
     * @param accountId The account to use
     * @param conferenceId The conference to lock mute
     * @param lock Whether to lock or unlock the mute for all participants
     * @return A [Completable] to wait for mute lock completion
     */
    fun setMuteLock(accountId: Int, conferenceId: Int, lock: Boolean): Completable

    /**
     * Mutes or un-mutes all participants in the conference.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param mute Whether to mute or un-mute all participants
     * @return A [Completable] to wait for mute others completion
     */
    fun setMuteOthers(accountId: Int, conferenceId: Int, mute: Boolean): Completable

    /**
     * Enables or disables chime (entry/exit tones) for the given conference.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param chimeEnabled Whether to enable or disable chimes
     * @return A [Completable] to wait for chime status change completion
     */
    fun setChimeStatus(accountId: Int, conferenceId: Int, chimeEnabled: Boolean): Completable

    /**
     * Starts or stops network recording for the given conference.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param record Whether to start or to stop network recording
     * @return A [Completable] to wait for network recording toggle completion
     */
    fun setNetworkRecording(accountId: Int, conferenceId: Int, record: Boolean): Completable

    /**
     * Assigns the prominent video tile to the specified participant.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param participantId Who to set the video floor to
     * @return A [Completable] to wait for video floor assignment completion
     */
    fun setVideoFloorParticipant(accountId: Int, conferenceId: Int, participantId: Int): Completable

    /**
     * Changes the video layout for the given conference.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param layout One of the pre-defined layout variants to change to
     * @return A [Completable] to wait for video layout change completion
     */
    fun setVideoLayout(accountId: Int, conferenceId: Int, layout: VideoLayout): Completable

    /**
     * Kicks the given participant from the conference.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param participantId Who to kick from the conference
     * @return A [Completable] to wait until the participant has been kicked
     */
    fun kickParticipant(accountId: Int, conferenceId: Int, participantId: Int): Completable

    /**
     * Network mutes/un-mutes the given participant in the conference.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param participantId Who to mute/un-mute
     * @param mute Whether to mute or un-mute
     * @return A [Completable] to wait until the participant has been muted/un-muted
     */
    fun muteParticipant(accountId: Int, conferenceId: Int, participantId: Int, mute: Boolean): Completable

    /**
     * Sets recording flag for the given participant in the conference.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param participantId Who to flag for recording
     * @param record Whether to flag with 'recording' or 'not recording'
     * @return A [Completable] to wait until the participant has been flagged
     */
    fun setParticipantRecording(accountId: Int, conferenceId: Int, participantId: Int, record: Boolean): Completable

    /**
     * Retrieves group chat XMPP account information.
     * It is expected that this method will be used more for 'guest VCCS' situations
     * rather than 'home VCCS' situations (when user's pre-configured account is valid).
     * @param accountId The account to use
     * @param conferenceId The conference to query
     * @param guestMode Whether to fetch the 'guest' account info or the 'home' account info
     * @return A [Single] to wait for XMPP account details ([XMPPAccountInfoEvent])
     */
    fun getXmppAccountInfo(accountId: Int, conferenceId: Int, guestMode: Boolean): Single<EventHolder<XMPPAccountInfoEvent>>

    /**
     * Queries the conference for textual and HTML invites, and also fetches the conference URL.
     * @param accountId The account to use
     * @param conferenceId The conference to query
     * @return A [Single] to wait for the invite ([OnQueryConferenceInvite])
     */
    fun getConferenceInvite(accountId: Int, conferenceId: Int): Single<EventHolder<OnQueryConferenceInvite>>

    /**
     * Assigns the screen share command to the specified participant.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param participantId Who to set the screen share to
     * @return A [Completable] to wait for screen share assignment completion
     */
    fun setScreenSharePresenter(accountId: Int, conferenceId: Int, participantId: Int): Completable

    /**
     * Queries the conference for all host settings, wrapped in a single config object.
     * @param accountId The account to use
     * @param conferenceId The conference to query
     * @return A [Single] to wait for the config ([ConferenceConfigEvent])
     */
    fun getConferenceConfig(accountId: Int, conferenceId: Int): Single<EventHolder<ConferenceConfigEvent>>

    /**
     * Sends the conference configuration options to the server.
     * @param accountId The account to use
     * @param conferenceId The conference to use
     * @param values The config values to send. You can ignore unused values (keep them at default values)
     * @param flags An (ordered) set of flags to tell the server which of the values to actually persist. Set to `true` to send
     * @return A [Completable] to wait for conference settings update completion
     */
    fun setConferenceConfig(accountId: Int, conferenceId: Int, values: ConferenceConfigValues, flags: ConferenceConfigFlags): Completable

    /**
     * Continuously observes changes on the participant list in the given conference.
     * @param accountId The account to use
     * @param conferenceId The conference to observe
     * @return An [Observable] to monitor the changes in the form of [ParticipantChanges]
     */
    fun observeParticipantChanges(accountId: Int, conferenceId: Int): Observable<EventHolder<ParticipantChanges>>

    /**
     * Continuously observes changes with the screen share in the given conference.
     * @param accountId The account to use
     * @param conferenceId The conference to observe
     * @return An [Observable] to monitor the changes in the form of [ScreenShareConfigEvent]s
     */
    fun observeScreenShareChanges(accountId: Int, conferenceId: Int): Observable<EventHolder<ScreenShareConfigEvent>>

    /**
     * Continuously observes participants' voice activity in the given conference.
     * @param accountId The account to use
     * @param conferenceId The conference to observe
     * @return An [Observable] to monitor the changes in the form of [OnVoiceActivityChanged] events
     */
    fun observeVoiceActivityChanges(accountId: Int, conferenceId: Int): Observable<EventHolder<OnVoiceActivityChanged>>

}
