package com.bria.common.controller.collaboration.rx.impl

import android.util.Log
import com.bria.common.controller.collaboration.rx.api.CollabAccountsSdkApi
import com.bria.common.controller.collaboration.rx.api.CollabAccountsSdkApi.Companion.K_ERROR
import com.bria.common.controller.collaboration.rx.api.CollabAccountsSdkApi.Companion.K_SUCCESS
import com.bria.common.controller.collaboration.rx.exceptions.CollabAccountException
import com.counterpath.sdk.SipVccsAccountApi
import com.counterpath.sdk.handler.SipVccsAccountHandler
import com.counterpath.sdk.pb.VccsAccount
import com.counterpath.sdk.pb.VccsAccount.VccsAccountStateChangedEvent
import io.reactivex.Observable
import io.reactivex.ObservableEmitter
import io.reactivex.ObservableOnSubscribe
import io.reactivex.plugins.RxJavaPlugins

class CollabAccountsSdkApiImpl(private val accountManager: SipVccsAccountApi?) : CollabAccountsSdkApi {

    private companion object {
        val TAG = "CollabAccountsSdkApiImpl"!!
        const val ERROR_API_MISSING = "SDK module not available"
        const val ERROR_EVENT_MISSING = "Missing event: Error"
        const val ERROR_ACTION_FAILED = "Account action failed"
        const val CHANGE_EVENT_MISSING = "Missing event: Status change"
    }

    init {
        // not sure if this is a good approach
        val originalHandler = RxJavaPlugins.getErrorHandler()
        RxJavaPlugins.setErrorHandler { error ->
            if (error is CollabAccountException) {
                println("Unhandled (non-fatal) RxJava exception: $error")
            } else {
                originalHandler?.accept(error)
            }
        }
    }

    override fun createAccount(): Int = accountManager?.Create() ?: K_ERROR

    override fun enableAccount(id: Int, immediate: Boolean): Observable<VccsAccountStateChangedEvent> {
        // convert the callback-style API to a reactive API
        accountManager?.let {
            val action: SipVccsAccountApi.() -> Int = { Enable(id) }
            if (immediate && K_ERROR == it.action()) return Observable.error(CollabAccountException(id, ERROR_ACTION_FAILED))
            val subscribeCallback = makeCallbacksReactive(it, if (immediate) null else action)
            return Observable.create(subscribeCallback).filter { it.vccsAccountHandle == id }
        }
        // if API is not available, fail reactively
        return Observable.error(CollabAccountException(id, ERROR_API_MISSING))
    }

    override fun disableAccount(id: Int, immediate: Boolean): Observable<VccsAccountStateChangedEvent> {
        // convert the callback-style API to a reactive API
        accountManager?.let {
            val action: SipVccsAccountApi.() -> Int = { Disable(id) }
            if (immediate && K_ERROR == it.action()) return Observable.error(CollabAccountException(id, ERROR_ACTION_FAILED))
            val subscribeCallback = makeCallbacksReactive(it, if (immediate) null else action)
            return Observable.create(subscribeCallback).filter { it.vccsAccountHandle == id }
        }
        // if API is not available, fail reactively
        return Observable.error(CollabAccountException(id, ERROR_API_MISSING))
    }

    override fun observeAccounts(): Observable<VccsAccountStateChangedEvent> {
        // convert the callback-style API to a reactive API
        accountManager?.let {
            return Observable.create(makeCallbacksReactive(it))
        }
        // if API is not available, fail reactively
        return Observable.error(CollabAccountException(K_ERROR, ERROR_API_MISSING))
    }

    override fun updateAccount(id: Int, settings: VccsAccount.VccsAccountSettings): Boolean {
        accountManager?.let {
            if (it.ConfigureDefaultAccountSettings(id, settings) == K_ERROR) return false
            return it.ApplySettings(id) != K_ERROR
        }
        return false
    }

    override fun deleteAccount(id: Int): Boolean = accountManager?.Destroy(id) != K_ERROR

    override fun parseUrl(url: String): SipVccsAccountApi.CrackVCCSURLResult? = accountManager?.CrackVCCSURL(
            url,
            url.startsWith("https://") || url.startsWith("wss://") || url.startsWith("vccs://")
    )

    /* Private helpers */

    /**
     * Creates a new 'on subscribe' action.
     * @param accountManager A non-`null` instance of the accounts API
     * @param action What to do on the API (return [K_ERROR] if something fails). Defaults to `null`
     * @return A subscribe action for when a observer subscribes to the reactive observable
     */
    private fun makeCallbacksReactive(accountManager: SipVccsAccountApi, action: (SipVccsAccountApi.() -> Int)? = null):
            ObservableOnSubscribe<VccsAccountStateChangedEvent> {
        return ObservableOnSubscribe { emitter ->
            // set up a new (callback) listener here; it forwards callbacks through the reactive emitter
            val listener = object : SipVccsAccountHandler {
                override fun onAccountStateChanged(origin: SipVccsAccountApi?, event: VccsAccountStateChangedEvent?): Int {
                    // make sure emitter is still able to emit (this only happens in multi-thread access to origin resource emitter)
                    if (emitter.isDisposed) return K_ERROR
                    // forward callbacks to Rx Observer
                    event?.let {
                        emitter.onNext(it)
                        return K_SUCCESS
                    }
                    // event might be null, in that case emit an error
                    return emitter.onErrorSafe(CollabAccountException(K_ERROR, CHANGE_EVENT_MISSING))
                }

                override fun onError(origin: SipVccsAccountApi?, error: VccsAccount.onErrorEvent?): Int {
                    // make sure emitter is still able to emit (this only happens in multi-thread access to origin resource emitter)
                    if (emitter.isDisposed) return K_ERROR
                    // forward errors to Rx Observer
                    error?.let {
                        return emitter.onErrorSafe(CollabAccountException(it.vccsAccountHandle, it.errorText))
                    }
                    // no error event, deliver that problem too
                    return emitter.onErrorSafe(CollabAccountException(K_ERROR, ERROR_EVENT_MISSING))
                }
            }

            // attach the listener and detach it when unsubscribed
            val registration = accountManager.addHandler(listener)
            emitter.setCancellable {
                registration.removeHandler()
                // TODO we may want to complete here
            }

            // and finally execute the action when subscription is executed
            action?.let {
                if (accountManager.it() == K_ERROR) {
                    emitter.onErrorSafe(CollabAccountException(K_ERROR, ERROR_ACTION_FAILED))
                }
            }
        }
    }

    @Synchronized
    private fun <E> ObservableEmitter<E>.onErrorSafe(exception: Exception): Int {
        return try {
            if (!tryOnError(exception)) {
                Log.d(TAG, "Failed to deliver: $exception")
                K_ERROR
            } else K_SUCCESS
        } catch (error: Throwable) {
            println("${javaClass.simpleName} | Error while emitting exception: $exception")
            K_ERROR
        }
    }

}
