package com.bria.common.controller.collaboration.rx.impl

import android.Manifest
import android.annotation.SuppressLint
import android.content.Context
import android.content.pm.PackageManager
import android.support.v4.content.ContextCompat
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.EAccountResult
import com.bria.common.controller.accounts.IAccountsCtrlActions
import com.bria.common.controller.accounts.IAccountsFilter
import com.bria.common.controller.collaboration.CollaborationController
import com.bria.common.controller.collaboration.ICollaborationActions
import com.bria.common.controller.collaboration.account.SimpleVccsAccount
import com.bria.common.controller.collaboration.rx.api.ConferenceEntityManager
import com.bria.common.controller.collaboration.rx.entities.LinkInfo
import com.bria.common.controller.collaboration.rx.entities.SubscriptionRequest
import com.bria.common.controller.collaboration.utils.SimpleVccsConference
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.ISettingsCtrlActions
import com.bria.common.controller.settings.branding.AccountTemplate
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.controller.settings.branding.EEncryptAudio
import com.bria.common.controller.settings.branding.ESipTransportType
import com.bria.common.util.Utils
import com.counterpath.sdk.SipAccountApi
import com.counterpath.sdk.pb.Account.AccountSettings
import io.reactivex.Observable
import java.util.*

class ConferenceEntityManagerImpl(
        private val context: Context,
        private val settings: ISettingsCtrlActions,
        private val accountsCtrl: IAccountsCtrlActions,
        private val sipAccountApi: SipAccountApi
) : ConferenceEntityManager {

    companion object {
        const val DEFAULT_URL_PORT = 8990
    }

    override fun createSubscriptionRequest(accountId: Int, bridgeSip: String?, subscriptionCode: String, pinCode: String, participantType: Int): SubscriptionRequest {
        return SubscriptionRequest(accountId = accountId, bridgeSip = bridgeSip, subscriptionCode = subscriptionCode, pinCode = pinCode, appId = getAppId(context), sipAddess = getSipAddress(), participantType = participantType, capabilities = getCapabilities())

    }

    override fun createConference(linkInfo: LinkInfo?): SimpleVccsConference {
        // load local credentials
        val localGroup = settings.getStr(ESetting.VccsGroup)
        val localUsername = settings.getStr(ESetting.ProvisioningUsername)
        val localPassword = settings.getStr(ESetting.ProvisioningPassword)
        val localUrl = settings.getStr(ESetting.VccsUrl)

        // and finally create the conference
        return linkInfo?.let {
            // when connecting to other environments, username needs to be empty; nonsense from BRACE-10453
            val username = if (it.groupName == localGroup) localUsername else ""
            SimpleVccsConference
                    .builder()
                    .setWebSocketURL(it.webSocketUrl)
                    .setOriginalURL(it.original)
                    .setServerName(it.serverName)
                    .setPortNumber(it.portNumber)
                    .setGroupName(it.groupName)
                    .setSubscriptionCode(it.conferenceCode)
                    .setUsername(username)
                    .setPassword(localPassword)
                    .build()
        } ?: ({
            SimpleVccsConference
                    .builder()
                    .setGroupName(localGroup)
                    .setServerName(localGroup)
                    .setUsername(localUsername)
                    .setPassword(localPassword)
                    .setWebSocketURL(localUrl)
                    .setPortNumber(DEFAULT_URL_PORT)
                    .setIsHost()
                    .build()
        })()
    }

    override fun createCollabAccount(conference: SimpleVccsConference?): SimpleVccsAccount {
        // load local credentials
        val localGroup = settings.getStr(ESetting.VccsGroup)
        val localUser = settings.getStr(ESetting.ProvisioningUsername)
        val localPass = settings.getStr(ESetting.ProvisioningPassword)
        val localUrl = settings.getStr(ESetting.VccsUrl)
        val displayName = getDisplayName()
        val xmppJid = getXmppUserName()
        // load conference credentials
        val conferenceGroup = conference?.groupName ?: ""
        val conferenceUser = conference?.username ?: ""
        val conferencePass = conference?.password ?: ""
        val conferenceUrl = conference?.webSocketURL ?: ""
        // check if any of these are empty and replace them with local values
        val group = if (conferenceGroup.isBlank()) localGroup else conferenceGroup
        val password = if (conferencePass.isBlank()) localPass else conferencePass
        val webSocketUrl = if (conferenceUrl.isBlank()) localUrl else conferenceUrl
        // when connecting to other environments, username needs to be empty; nonsense from BRACE-10453
        var username = if (conferenceUser.isBlank()) localUser else conferenceUser
        username = if (group == localGroup) username else ""
        // and finally create the account
        return SimpleVccsAccount(group, username, password, displayName, xmppJid, webSocketUrl)
    }

    override fun findPrimaryAccount(): Account? = accountsCtrl.primaryAccount?.let { it } ?: ({
        // no active primary account, just get the first available SIP
        val sipAccounts = accountsCtrl.accountsProvider.getAccounts(IAccountsFilter.SIP)
        Observable.fromIterable(sipAccounts).blockingFirst(null)
    })()

    override fun getXmppChatAccount(): Account? = accountsCtrl.getAccounts(IAccountsFilter.ENABLED_XMPP)?.firstOrNull()

    override fun getXmppUserName(): String = getXmppChatAccount()?.let {
        "${it.getStr(EAccountSetting.UserName)}@${it.getStr(EAccountSetting.Domain)}"
    } ?: ""

    override fun getDisplayName(): String = findPrimaryAccount()?.let {
        val display = it.getStr(EAccountSetting.DisplayName) ?: ""
        if (display.trim().isEmpty()) getSipUsername() else display
    } ?: ""

    override fun getCapabilities(): Int {
        val hasVideo = settings.getBool(ESetting.FeatureVideo) && settings.getBool(ESetting.VideoEnabled)
        val hasScreenshare = settings.getBool(ESetting.ScreenshareEnabled)
        // no screen share capability for now
        var capabilities = ICollaborationActions.PART_CAPABILITY_BRIA or
                ICollaborationActions.PART_CAPABILITY_CALL or
                ICollaborationActions.PART_CAPABILITY_CHAT
        if (hasVideo) {
            capabilities = capabilities or ICollaborationActions.PART_CAPABILITY_VIDEO
        }
        if (hasScreenshare) {
            capabilities = capabilities or ICollaborationActions.PART_CAPABILITY_SHARE
        }
        return capabilities
    }

    override fun getSipAddress(): String = "${getSipUsername()}${getAtSipDomain()}"

    override fun getSipUsername(): String = findPrimaryAccount()?.getStr(EAccountSetting.UserName) ?: ""

    override fun getAtSipDomain(): String = findPrimaryAccount()?.let {
        "@${it.getStr(EAccountSetting.Domain)}"
    } ?: ""

    @SuppressLint("MissingPermission")
    override fun getAppId(context: Context): String =
            if (ContextCompat.checkSelfPermission(context, Manifest.permission.READ_PHONE_STATE) == PackageManager.PERMISSION_GRANTED) {
                Utils.System.getHashedDeviceId(context)
            } else "device_id_permission_not_granted_${Utils.System.getSystemSerialNumber()}"

    override fun createDirectSipAccount(directSipProvisioningJson: String): String? {
        // parse account settings
        val parsedAccountSettings = mutableListOf<AccountSettings>()
        sipAccountApi.decodeProvisioningResponse(directSipProvisioningJson, parsedAccountSettings)
        // check parsing result
        val directSdkAccount = (if (parsedAccountSettings.isEmpty()) null else parsedAccountSettings[0]) ?: return null
        // clone the default/primary account template for creating the direct account
        val collabTemplateName = "Direct VCCS Template"
        val primaryAccount = findPrimaryAccount()
        val template: AccountTemplate
        template = if (primaryAccount != null) {
            AccountTemplate(collabTemplateName, settings.getAccountTemplate(primaryAccount))
        } else {
            AccountTemplate(collabTemplateName, settings.getGenericTemplate(EAccountType.Sip))
        }
        val sipTransportType = ESipTransportType.mapFromSdkValue(directSdkAccount.sipTransportType)
        template.set(EAccountSetting.AccountName, CollaborationController.COLLAB_ACCOUNT_NAME)
        template.set(EAccountSetting.AllowIncomingCalls, false)
        template.set(EAccountSetting.UserName, directSdkAccount.username)
        template.set(EAccountSetting.Domain, directSdkAccount.domain)
        template.set(EAccountSetting.Password, directSdkAccount.password)
        template.set(EAccountSetting.SipTransport, sipTransportType)
        template.set(EAccountSetting.OutProxy, directSdkAccount.outboundProxy)
        template.set(EAccountSetting.AlwaysRouteViaOutboundProxy, true)
        val keys = directSdkAccount.acceptedCertPublicKeys
        if (keys.isNotEmpty()) {
            val keyList = Arrays.asList(*keys)
            template.set(EAccountSetting.CertPublicKeysAccepted, keyList)
        }
        if (sipTransportType == ESipTransportType.TLS) {
            // Enable secure media
            template.set(EAccountSetting.EncryptAudio, EEncryptAudio.Always)
        }

        // create and enable direct SIP account
        return if (accountsCtrl.createAccount(template) == EAccountResult.Success) {
            EAccountResult.Success.resultDetails as String
        } else {
            null
        }
    }

    override fun removeSipAccount(accNickName: String): EAccountResult {
        val sipAccount = accountsCtrl.getAccount(accNickName)
        if (sipAccount != null) {
            return (accountsCtrl.removeAccount(sipAccount))
        }
        return EAccountResult.NotFound
    }

}
