package com.bria.common.controller.collaboration.rx.impl

import com.bria.common.controller.collaboration.ICollaborationActions.PARTICIPANT_TYPE_UNKNOWN
import com.bria.common.controller.collaboration.ICollaborationActions.PARTICIPANT_TYPE_WEB
import com.bria.common.util.Utils
import com.counterpath.sdk.pb.VccsConference.ConferenceDetails
import com.counterpath.sdk.pb.VccsConference.ParticipantStatus
import com.counterpath.sdk.pb.nano.Vccsconference.*
import io.reactivex.Observable

object ParticipantManager {

    const val UNKNOWN_ID = -1

    /**
     * Returns the Participant object associated with the current caller's SIP address.
     *
     * @return {@link ParticipantStatus} reference if such participant is found, {@code null} if call or conference or participant are not
     * available
     */
    var self: ParticipantStatus? = null
        @JvmStatic
        @Synchronized get
        private set

    /**
     * Checks if any of the participants started recording the ongoing conference. Local/current user is not counted.
     *
     * @return {@code True} if at least one participant is recording, {@code false} if none are recording
     */
    var isAnyRecording: Boolean = false
        @JvmStatic
        @Synchronized get
        private set

    /**
     * Tries to find the participant who is currently presenting.
     *
     * @return A {@link ParticipantStatus} object referencing the person who is currently sharing their screen, or {@code null} if not available
     */
    @JvmStatic
    var presenter: ParticipantStatus? = null
        @Synchronized get

    /**
     * Tries to find the participant who is currently moderator.
     *
     * @return True if participant is currently moderator, or {@code null} if not available
     */
    var hasModerators: Boolean = false
        @JvmStatic
        @Synchronized get
        private set

    private var oldPresenter: Int = UNKNOWN_ID

    data class Diff(
            val joined: MutableList<ParticipantStatus> = mutableListOf(),
            val updated: MutableList<ParticipantStatus> = mutableListOf(),
            val left: MutableList<ParticipantStatus> = mutableListOf(),
            var isNewList: Boolean = false,
            var isSelfChanged: Boolean = false,
            var isPresenterChanged: Boolean = false,
            var hasSelfJustJoined: Boolean = false,
            var isAnyRecordingChanged: Boolean = false,
            var moderatorsChanged: Boolean = false,
            var networkMuteChanged: Boolean = false
    )

    private val participants: MutableList<ParticipantStatus> = mutableListOf()

    @Synchronized
    @JvmStatic
    fun getParticipants(withSelf: Boolean = true) = if (withSelf) {
        participants.toMutableList()
    } else {
        participants.filter { (self?.participantNumber ?: UNKNOWN_ID) != it.participantNumber }.toMutableList()
    }

    @Synchronized
    @JvmStatic
    fun updateParticipants(
            updates: MutableList<ParticipantStatus>? = null,
            conference: ConferenceDetails? = null
    ): Diff {
        // remember previous properties
        val oldList = participants.toMutableList()
        val oldSelf = self
        val oldIsAnyRecording = isAnyRecording
        val oldHasModerators = hasModerators

        // find list differences
        val selfId = conference?.socketParticipantNumber ?: UNKNOWN_ID
        val diff = listsDiff(participants, updates, selfId)
        applyChanges(diff, selfId)

        // find changes in properties
        if (oldList.isEmpty() && participants.isNotEmpty()) {
            diff.isNewList = true
        }
        if (!equals(oldSelf, self)) {
            diff.isSelfChanged = true
        }
        val selfAdded = oldSelf == null && self != null && self?.callStatus == CallStatus_InConference
        val selfNowInCall = (oldSelf != null && oldSelf.callStatus != CallStatus_InConference) && (self != null && self?.callStatus == CallStatus_InConference)
        if (selfAdded || selfNowInCall) {
            diff.hasSelfJustJoined = true // TODO some logic overlap possible ^
        }
        if ( presenter != null || oldPresenter != presenter?.participantNumber ?: UNKNOWN_ID) {
            oldPresenter = presenter?.participantNumber ?: UNKNOWN_ID
            diff.isPresenterChanged = true
        }
        if (oldIsAnyRecording != isAnyRecording) {
            diff.isAnyRecordingChanged = true
        }
        if (oldHasModerators != hasModerators) {
            diff.moderatorsChanged = true
        }
        if (selfAdded || selfNowInCall || !Utils.equals(oldSelf?.isMuted, self?.isMuted)) {
            diff.networkMuteChanged = true
        }

        return diff
    }

    @Synchronized
    @JvmStatic
    fun isEmpty() = participants.isEmpty()

    @Synchronized
    @JvmStatic
    fun clear(clearRoles: Boolean = true) {
        participants.clear()
        if (clearRoles) {
            self = null
            presenter = null
            isAnyRecording = false
            hasModerators = false
        }
    }

    /* Private helpers */

    @Synchronized
    @JvmStatic
    fun findById(id: Int) = participants.firstOrNull { id == it.participantNumber }

    @Synchronized
    private fun listsDiff(oldList: MutableList<ParticipantStatus>, updates: MutableList<ParticipantStatus>?, selfId: Int): Diff {
        // tackle edge cases
        if (updates == null || updates.isEmpty()) return Diff()
        if (oldList.isEmpty()) {
            val result = Diff()
            updates.forEach {
                if (it.callStatus == CallStatus_InConference || it.callStatus == CallStatus_None) {
                    result.joined.add(it)
                }
            }
            return result
        }

        // go through the updates to detect differences
        val result = Diff()
        updates.forEach {
            // check if this is me that has changed
            val isSelf = selfId == it.participantNumber
            val oldParticipant = findById(it.participantNumber)

            when (it.callStatus) {
                CallStatus_InConference -> {
                    // check if participant has joined for the first time
                    if (oldParticipant == null) {
                        result.joined.add(it)
                    } else {
                        result.updated.add(it)
                    }
                }
                CallStatus_Kicked, CallStatus_Ended -> {
                    if (isSelf) {
                        // never remove self from list
                        result.updated.add(it)
                    } else if (oldParticipant != null) {
                        // check if participant has left for the first time (they can come back apparently...)
                        if (oldParticipant.callStatus != CallStatus_Ended && oldParticipant.callStatus != CallStatus_Kicked) {
                            result.left.add(it)
                        } else {
                            result.updated.add(it)
                        }
                    }
                }
                CallStatus_LockedOut -> {
                    result.updated.add(it)
                }
                CallStatus_None -> {
                    if (it.geParticipantType() == PARTICIPANT_TYPE_WEB) {
                        // WEB type: web participant is in conference
                        if (oldParticipant == null) {
                            result.joined.add(it)
                        } else {
                            result.updated.add(it)
                        }
                    } else if (it.geParticipantType() == PARTICIPANT_TYPE_UNKNOWN) {
                        // UNKNOWN type: web participant has left the conference
                        if (oldParticipant != null) {
                            result.left.add(it)
                        }
                    } else {
                        // this is server sending garbage data
                        if (oldParticipant != null && !isSelf) {
                            result.left.add(it)
                        }
                    }
                }
            }
        }

        return result
    }

    @Synchronized
    private fun applyChanges(diff: Diff, selfId: Int) {
        // remove all who left
        participants.removeAllParticipants(diff.left)
        // update the ones who were really updated
        val iterator = diff.updated.listIterator()
        while (iterator.hasNext()) {
            val newValue = iterator.next()
            val oldValue = findById(newValue.participantNumber)!!
            if (oldValue equalTo newValue) {
                iterator.remove()
            } else {
                participants[participants.indexOf(oldValue)] = newValue
            }
        }
        // add all who joined
        participants.addAll(diff.joined)

        // update conference roles
        self = findById(selfId)
        if (presenter != null && presenter!!.callStatus == CallStatus_Kicked){
            val presenterID : Int = presenter!!.participantNumber
            presenter = findById(presenterID)
        }

        hasModerators = participants.any { it.isModerator }
        isAnyRecording = participants.filter { self == null || self!!.participantNumber != it.participantNumber }.any { it.isRecording }
    }

    private fun equals(first: ParticipantStatus?, second: ParticipantStatus?): Boolean = if (first == null) second == null else (first equalTo second)

    private infix fun ParticipantStatus.equalTo(other: ParticipantStatus?): Boolean = other != null && participantNumber == other.participantNumber &&
            userId == other.userId && xmppUserId == other.xmppUserId && xmppUsername == other.xmppUsername &&
            sipUsername == other.sipUsername && displayName == other.displayName && emailAddress == other.emailAddress &&
            callSessionID == other.callSessionID && sipAddress == other.sipAddress && recordingURI == other.recordingURI &&
            isRecording == other.isRecording && startTime == other.startTime && endTime == other.endTime &&
            callStatus == other.callStatus && isMuted == other.isMuted && isModerator == other.isModerator &&
            hasVideoFloor == other.hasVideoFloor && capabilites == other.capabilites && geParticipantType() == other.geParticipantType()


    @JvmStatic
    fun isCurrentUserModerator(): Boolean {
        return self != null && self!!.isModerator
    }

    /**
     * Tries to remove the given [ParticipantStatus] from the given participant list by matching its SIP AOR. <font color="#FF0000">**Not thread-safe.**</font>
     *
     * @param toRemove Which participant should be removed (this can be `null`, which returns `false`)
     * @param list Which list to remove the participant from (this can be `null`, which returns `false`)
     *
     * @return `True` if participant was found in the list and successfully removed, `false` otherwise
     */
    @JvmStatic
    fun removeParticipant(toRemove: ParticipantStatus?, list: MutableList<ParticipantStatus>?):Boolean {
        if (list == null || list.isEmpty() || toRemove == null) {
            return false
        }

        val originalSize = list.size
        val filtered = Observable
                .fromIterable(list)
                .filter { status -> status.participantNumber != toRemove.participantNumber }
                .toList()
                .blockingGet()
        list.clear()
        list.addAll(filtered)

        // if size is smaller, we removed something
        return list.size < originalSize
    }

    private fun MutableList<ParticipantStatus>.removeAllParticipants(toRemove: List<ParticipantStatus>) {
        val setIds = toRemove.map { it.participantNumber }.toSet()
        val iterator = listIterator()
        while (iterator.hasNext()) {
            if (iterator.next().participantNumber in setIds) {
                iterator.remove()
            }
        }
    }

}
