package com.bria.common.controller.collaboration.rx.impl

import com.bria.common.controller.collaboration.account.EAccountState.*
import com.bria.common.controller.collaboration.rx.api.CollabAccountsSdkApi
import com.bria.common.controller.collaboration.rx.api.ConferenceEntityManager
import com.bria.common.controller.collaboration.rx.api.ISipVccsConferenceApi
import com.bria.common.controller.collaboration.rx.api.ReactiveCollabApi
import com.bria.common.controller.collaboration.rx.entities.ConferenceConfig
import com.bria.common.controller.collaboration.rx.entities.EventHolder
import com.bria.common.util.reactive.api.SchedulerProvider
import com.counterpath.sdk.pb.VccsConference.ConferenceDetails
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.SingleTransformer
import java.util.*
import java.util.concurrent.TimeUnit

class ReactiveCollabApiImpl(
        private val accountsApi: CollabAccountsSdkApi,
        private val conferenceApi: ISipVccsConferenceApi,
        private val entityManager: ConferenceEntityManager,
        private val schedulers: SchedulerProvider
) : ReactiveCollabApi {

    private companion object {
        private const val CONFERENCE_TIMEOUT = 8L // seconds
        private const val DEATH_TIMEOUT = 2L // seconds
    }

    private val connectedAccounts: MutableSet<Int> = Collections.synchronizedSet(mutableSetOf())

    /* Public API */

    override fun getInviteUrl(): Single<String> = getMyConferenceIds()
        .flatMap { conferenceApi.getConferenceInvite(it.first, it.second) }
        .observeOn(schedulers.computation())
        .map { it.copy(it.data.joinUrl) }
        .compose(disconnectAndCarry())

    override fun getConferenceConfig(): Single<ConferenceConfig> = getMyConferenceIds()
        .flatMap { conferenceApi.getConferenceConfig(it.first, it.second) }
        .observeOn(schedulers.computation())
        .map { it.copy(ConferenceConfig(it.data)) }
        .compose(disconnectAndCarry())

    override fun setConferenceConfig(config: ConferenceConfig): Completable = getMyConferenceIds()
        .flatMap {
            val (values, flags) = config.convertToRequest()
            val event = EventHolder(it.first, it.second, config)
            val setterTask = conferenceApi.setConferenceConfig(event.accountId, event.conferenceId, values, flags)
            setterTask.toSingleDefault(event)
        }
        .compose(disconnectAndCarry())
        .toCompletable()

    override fun disconnectAll() = synchronized(connectedAccounts) {
        connectedAccounts.mapNotNull {
            try {
                disconnectAccount(it).toCompletable().blockingGet(DEATH_TIMEOUT, TimeUnit.SECONDS)
            } catch (ignored: Throwable) {
            }
        }
        connectedAccounts.clear()
        return@synchronized
    }

    /* Private helpers */

    private fun <T> disconnectAndCarry(): SingleTransformer<EventHolder<T>, T> = SingleTransformer { upstream ->
        upstream
            .timeout(CONFERENCE_TIMEOUT, TimeUnit.SECONDS, schedulers.computation())
            .flatMap { disconnectAccount(it.accountId, it.data) }
    }

    // gets account and conference IDs for the current user's primary conference
    private fun getMyConferenceIds(): Single<Pair<Int, Int>> = getConnectedAccountId(true)
        .observeOn(schedulers.io())
        .flatMap { conferenceApi.getConferenceList(it) }
        .observeOn(schedulers.computation())
        .map(findMyConferences)
        .filter { !it.data.isEmpty() }
        .map { Pair(it.accountId, it.data[0].id.toInt()) }
        .toSingle()
        .observeOn(schedulers.io())

    // creates and connects a new collab account
    private fun getConnectedAccountId(errorOnFlakyNetwork: Boolean): Single<Int> {
        return Single
            .fromCallable { entityManager.createCollabAccount(null) }
            .subscribeOn(schedulers.computation())
            .map { account ->
                val accountId = accountsApi.createAccount()
                if (accountId == CollabAccountsSdkApi.K_ERROR) {
                    throw IllegalStateException("Collab account not created")
                }
                account.id = accountId
                connectedAccounts += accountId
                account
            }
            .map {
                if (!accountsApi.updateAccount(it.id, it.sdkSettings)) {
                    throw IllegalStateException("Collab account settings not updated")
                } else it
            }
            .observeOn(schedulers.io())
            .flatMapObservable { accountsApi.enableAccount(it.id) }
            .observeOn(schedulers.computation())
            .distinctUntilChanged { first, second -> first.newState == second.newState }
            .filter {
                val isFailingState = UNREGISTERING.stateId == it.newState || UNREGISTERED.stateId == it.newState
                if (errorOnFlakyNetwork && isFailingState) {
                    throw IllegalStateException("Collab account disconnected")
                } else {
                    REGISTERED.stateId == it.newState
                }
            }
            .map { it.vccsAccountHandle }
            .firstElement()
            .toSingle()
    }

    // disconnects and deletes a previously connected collab account, carrying the original data along the way
    private fun disconnectAccount(id: Int) = disconnectAccount(id, Any())
    private fun <T> disconnectAccount(id: Int, carryData: T): Single<T> {
        return accountsApi
            .disableAccount(id)
            .subscribeOn(schedulers.io())
            .observeOn(schedulers.io())
            .filter { UNREGISTERED.stateId == it.newState }
            .observeOn(schedulers.computation())
            .map {
                val accountId = it.vccsAccountHandle
                connectedAccounts -= accountId
                accountId
            }
            .map {
                if (!accountsApi.deleteAccount(it)) {
                    throw IllegalStateException("Account not deleted")
                } else carryData
            }
            .firstElement()
            .toSingle()
    }

    // removes conferences where current user is not a host
    private val findMyConferences = { conferences: EventHolder<List<ConferenceDetails>> ->
        Observable
            .fromIterable(conferences.data)
            .filter { it.socketModerator }
            .toList()
            .map { conferences.copy(it) }
            .blockingGet()
    }

}
