package com.bria.common.controller.collaboration.rx.impl

import com.bria.common.controller.collaboration.rx.api.ISipVccsConferenceApi
import com.bria.common.controller.collaboration.rx.api.ISipVccsConferenceApi.Companion.K_ERROR
import com.bria.common.controller.collaboration.rx.api.ISipVccsConferenceApi.Companion.K_SUCCESS
import com.bria.common.controller.collaboration.rx.api.ISipVccsConferenceApi.Companion.PARTICIPANT_TYPE
import com.bria.common.controller.collaboration.rx.api.ParticipantChanges
import com.bria.common.controller.collaboration.rx.entities.ConferenceConfigFlags
import com.bria.common.controller.collaboration.rx.entities.ConferenceConfigValues
import com.bria.common.controller.collaboration.rx.entities.EventHolder
import com.bria.common.controller.collaboration.rx.entities.SubscriptionRequest
import com.bria.common.controller.collaboration.rx.exceptions.CollabConferenceException
import com.bria.common.controller.collaboration.rx.rxlisteners.ConferenceListenerCompletable
import com.bria.common.controller.collaboration.rx.rxlisteners.ConferenceListenerObservable
import com.bria.common.controller.collaboration.rx.rxlisteners.ConferenceListenerSingle
import com.counterpath.sdk.SipVccsConferenceApi
import com.counterpath.sdk.pb.VccsConference.*
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.plugins.RxJavaPlugins

class SipVccsConferenceApiImpl(private val conferenceManager: SipVccsConferenceApi?) : ISipVccsConferenceApi {

    private companion object {
        const val ERROR_API_MISSING = "SDK module not available"
        const val ERROR_ACTION_FAILED = "Conference action failed"

        const val EVENT_MISSING = "event is missing"
        const val CONNECTION_EVENT_MISSING = "Conference connection $EVENT_MISSING"
        const val LIST_EVENT_MISSING = "Query list $EVENT_MISSING"
        const val DETAILS_EVENT_MISSING = "Query conference details $EVENT_MISSING"
        const val SUBSCRIBE_EVENT_MISSING = "Subscription $EVENT_MISSING"
        const val UNSUBSCRIBE_EVENT_MISSING = "Unsubscription $EVENT_MISSING"
        const val PARTICIPANTS_LOCK_EVENT_MISSING = "Participants lock $EVENT_MISSING"
        const val MUTE_LOCK_EVENT_MISSING = "Mute lock $EVENT_MISSING"
        const val MUTE_OTHERS_EVENT_MISSING = "Mute others $EVENT_MISSING"
        const val CHIME_CHANGE_EVENT_MISSING = "Chime change $EVENT_MISSING"
        const val NETWORK_RECORDING_EVENT_MISSING = "Network recording $EVENT_MISSING"
        const val ASSIGN_VIDEO_FLOOR_EVENT_MISSING = "Video floor assignment $EVENT_MISSING"
        const val VIDEO_LAYOUT_EVENT_MISSING = "Video layout change $EVENT_MISSING"
        const val PARTICIPANT_KICK_EVENT_MISSING = "Participant kick $EVENT_MISSING"
        const val PARTICIPANT_MUTE_EVENT_MISSING = "Participant mute $EVENT_MISSING"
        const val PARTICIPANT_REC_EVENT_MISSING = "Participant recording $EVENT_MISSING"
        const val XMPP_INFO_EVENT_MISSING = "XMPP account info $EVENT_MISSING"
        const val INVITE_EVENT_MISSING = "Conference invite $EVENT_MISSING"
        const val SCREENSHARE_PRESENTER_EVENT_MISSING = "Screenshare presenter $EVENT_MISSING"
        const val FETCH_CONFIG_EVENT_MISSING = "Fetch conference config $EVENT_MISSING"
        const val SET_CONFIG_EVENT_MISSING = "Set conference config $EVENT_MISSING"
        const val PARTICIPANT_CHANGE_EVENT_MISSING = "Participant list change $EVENT_MISSING"
        const val SCREENSHARE_CHANGE_EVENT_MISSING = "Screenshare config $EVENT_MISSING"
        const val VOICE_ACTIVITY_CHANGE_EVENT_MISSING = "Voice activity $EVENT_MISSING"
    }

    init {
        // not sure if this is a good approach
        val originalHandler = RxJavaPlugins.getErrorHandler()
        RxJavaPlugins.setErrorHandler { error ->
            if (error is CollabConferenceException) {
                println("Unhandled (non-fatal) RxJava exception: $error")
            } else {
                originalHandler?.accept(error)
            }
        }
    }

    /* Public API */

    override fun getConferenceConnectionInfo(accountId: Int, subscriptionCode: String): Single<EventHolder<ConferenceConnectionInfoEvent>> {
        checkNotNull(subscriptionCode, { "Subscription code is null" })

        val singleListener = object : ConferenceListenerSingle<EventHolder<ConferenceConnectionInfoEvent>>(accountId, K_ERROR) {
            override fun onConferenceConnectionInfo(api: SipVccsConferenceApi?, event: OnConferenceConnectionInfo?): Int {
                return reactToSingleSuccess(event, CONNECTION_EVENT_MISSING) {
                    EventHolder(accountId, it.vccsConferenceHandle, it.conferenceConnectionInfoEvent)
                }
            }

            override fun onConferenceConnectionInfoFailure(api: SipVccsConferenceApi?, error: OnConferenceConnectionInfoFailure?): Int {
                return reactToSingleError(error) {
                    "Error [$onConferenceConnectionInfoFailureErrorCode]: $onConferenceConnectionInfoFailureReason"
                }
            }
        }

        return createSingle(singleListener) { api: SipVccsConferenceApi ->
            api.getConferenceConnectionInfo(accountId, subscriptionCode)
        }
    }

    override fun getConferenceList(accountId: Int): Single<EventHolder<List<ConferenceDetails>>> {
        val singleListener = object : ConferenceListenerSingle<EventHolder<List<ConferenceDetails>>>(accountId, K_ERROR) {
            override fun onConferenceListUpdated(api: SipVccsConferenceApi?, event: OnConferenceListUpdated?): Int {
                return reactToSingleSuccess(event, LIST_EVENT_MISSING) {
                    Observable
                        .fromIterable(it.conferenceListUpdatedEvent.conferenceListChange)
                        .filter { change -> change.conferenceDetails != null }
                        .map { change -> change.conferenceDetails }
                        .toList()
                        .map { list -> EventHolder(accountId, conferenceId, list) }
                        .blockingGet()
                }
            }

            override fun onQueryConferenceListFailure(api: SipVccsConferenceApi?, error: OnQueryConferenceListFailure?): Int {
                return reactToSingleError(error) {
                    "Error [$onQueryConferenceListFailureErrorCode]: $onQueryConferenceListFailureReason"
                }
            }
        }

        return createSingle(singleListener) { api: SipVccsConferenceApi ->
            api.QueryConferenceList(accountId)
        }
    }

    override fun getConferenceDetails(accountId: Int, conferenceId: Int): Single<EventHolder<ConferenceDetails>> {
        val singleListener = object : ConferenceListenerSingle<EventHolder<ConferenceDetails>>(accountId, conferenceId) {
            override fun onConferenceListUpdated(api: SipVccsConferenceApi?, event: OnConferenceListUpdated?): Int {
                return reactToSingleSuccess(event, DETAILS_EVENT_MISSING) { listEvent ->
                    withDetails(conferenceId, listEvent) {
                        EventHolder(accountId, conferenceId, it)
                    }
                }
            }

            override fun onQueryConferenceListFailure(api: SipVccsConferenceApi?, error: OnQueryConferenceListFailure?): Int {
                return reactToSingleError(error) {
                    "Error [$onQueryConferenceListFailureErrorCode]: $onQueryConferenceListFailureReason"
                }
            }
        }

        return createSingle(singleListener) { api: SipVccsConferenceApi ->
            api.QueryConferenceDetails(accountId, conferenceId)
        }
    }

    override fun subscribe(request: SubscriptionRequest): Single<EventHolder<SubscribeEvent>> {
        val singleListener = object : ConferenceListenerSingle<EventHolder<SubscribeEvent>>(request.accountId, K_ERROR) {
            override fun onSubscribe(api: SipVccsConferenceApi?, event: OnSubscribe?): Int {
                return reactToSingleSuccess(event, SUBSCRIBE_EVENT_MISSING) {
                    EventHolder(accountId, it.subscribeEvent.vccsConferenceHandle, it.subscribeEvent)
                }
            }

            override fun onSubscribeFailure(api: SipVccsConferenceApi?, error: OnSubscribeFailure?): Int {
                return reactToSingleError(error) {
                    "Error [$subscribeFailureErrorCode]: $subscribeFailureReason"
                }
            }
        }

        return createSingle(singleListener) { api: SipVccsConferenceApi ->
            with(request) {
                api.Subscribe3(accountId, bridgeSip, subscriptionCode, pinCode, appId, sipAor, PARTICIPANT_TYPE, capabilities)
            }
        }
    }

    override fun unsubscribe(accountId: Int, bridgeNumber: String): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, K_ERROR) {
            override fun onUnsubscribe(api: SipVccsConferenceApi?, event: OnUnsubscribe?): Int {
                return reactToCompletableSuccess(event, UNSUBSCRIBE_EVENT_MISSING)
            }

            override fun onUnsubscribeFailure(api: SipVccsConferenceApi?, error: OnUnsubscribeFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$unsubscribeFailureErrorCode]: $unsubscribeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.Unsubscribe(accountId, bridgeNumber)
        }
    }

    override fun setParticipantsLock(accountId: Int, conferenceId: Int, lock: Boolean): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onConferenceListUpdated(api: SipVccsConferenceApi?, event: OnConferenceListUpdated?): Int {
                return reactToCompletableSuccess(event, PARTICIPANTS_LOCK_EVENT_MISSING) { listEvent ->
                    if (accountId != event?.vccsAccountHandle) throw IllegalArgumentException("Invalid account ID")
                    val lockStatusChanged = withDetails(conferenceId, listEvent) { it.participantLocked == lock } ?: false
                    if (!lockStatusChanged) throw IllegalStateException("Participant lock hasn't changed state!")
                }
            }

            override fun onSetConferenceModeFailure(api: SipVccsConferenceApi?, error: OnSetConferenceModeFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onSetConferenceModeFailureErrorCode]: $onSetConferenceModeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.SetParticipantLock(accountId, conferenceId, lock)
        }
    }

    override fun setMuteLock(accountId: Int, conferenceId: Int, lock: Boolean): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onConferenceListUpdated(api: SipVccsConferenceApi?, event: OnConferenceListUpdated?): Int {
                return reactToCompletableSuccess(event, MUTE_LOCK_EVENT_MISSING) { listEvent ->
                    if (accountId != event?.vccsAccountHandle) throw IllegalArgumentException("Invalid account ID")
                    val muteLockChanged = withDetails(conferenceId, listEvent) { it.muteLocked == lock } ?: false
                    if (!muteLockChanged) throw IllegalStateException("Mute lock hasn't changed state!")
                }
            }

            override fun onSetConferenceModeFailure(api: SipVccsConferenceApi?, error: OnSetConferenceModeFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onSetConferenceModeFailureErrorCode]: $onSetConferenceModeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.SetMuteLock(accountId, conferenceId, lock)
        }
    }

    override fun setMuteOthers(accountId: Int, conferenceId: Int, mute: Boolean): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onParticipantListUpdated(api: SipVccsConferenceApi?, event: OnParticipantListUpdated?): Int {
                return reactToCompletableSuccess(event, MUTE_OTHERS_EVENT_MISSING) {
                    val allChangedState = Observable
                        .fromIterable(it.participantListUpdatedEvent.participantListUpdatedEvent)
                        // only work with non-null statuses
                        .filter { change -> change.participantStatus != null }
                        // only count updated statuses
                        .filter { change -> change.changeType == ChangeType_Update }
                        // switch to status instances
                        .map { change -> change.participantStatus }
                        // leave only non-moderators and in-call participants (we're host and not muted)
                        .filter { participant -> !participant.isModerator && participant.callStatus == CallStatus_InConference }
                        // check if all changed state
                        .all { participant -> participant.isMuted == mute }
                        .blockingGet()
                    if (!allChangedState) {
                        throw IllegalStateException("There are participants who haven't changed state!")
                    }
                }
            }

            override fun onSetConferenceModeFailure(api: SipVccsConferenceApi?, error: OnSetConferenceModeFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onSetConferenceModeFailureErrorCode]: $onSetConferenceModeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.SetMuteOthers(accountId, conferenceId, mute)
        }
    }

    override fun setChimeStatus(accountId: Int, conferenceId: Int, chimeEnabled: Boolean): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onConferenceModeUpdated(api: SipVccsConferenceApi?, event: OnConferenceModeUpdated?): Int {
                return reactToCompletableSuccess(event, CHIME_CHANGE_EVENT_MISSING) {
                    if (!it.entryExitTonesEnabledChanged || it.entryExitTonesEnabled != chimeEnabled) {
                        throw IllegalStateException("Chime enabled status hasn't changed!")
                    }
                }
            }

            override fun onSetConferenceModeFailure(api: SipVccsConferenceApi?, error: OnSetConferenceModeFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onSetConferenceModeFailureErrorCode]: $onSetConferenceModeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.SetEntryExitTonesEnabled(accountId, conferenceId, chimeEnabled)
        }
    }

    override fun setNetworkRecording(accountId: Int, conferenceId: Int, record: Boolean): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onConferenceListUpdated(api: SipVccsConferenceApi?, event: OnConferenceListUpdated?): Int {
                return reactToCompletableSuccess(event, NETWORK_RECORDING_EVENT_MISSING) {
                    if (accountId != event?.vccsAccountHandle) throw IllegalArgumentException("Invalid account ID")
                    val recordingStateChanged = withDetails(conferenceId, it) { it.getisRecording() == record } ?: false
                    if (!recordingStateChanged) throw IllegalStateException("Network recording status hasn't changed!")
                }
            }

            override fun onSetConferenceModeFailure(api: SipVccsConferenceApi?, error: OnSetConferenceModeFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onSetConferenceModeFailureErrorCode]: $onSetConferenceModeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.setRecording(accountId, conferenceId, record)
        }
    }

    override fun setVideoFloorParticipant(accountId: Int, conferenceId: Int, participantId: Int): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onParticipantListUpdated(api: SipVccsConferenceApi?, event: OnParticipantListUpdated?): Int {
                return reactToCompletableSuccess(event, ASSIGN_VIDEO_FLOOR_EVENT_MISSING) {
                    val inCallParticipants = Observable
                        .fromIterable(it.participantListUpdatedEvent.participantListUpdatedEvent)
                        // only work with non-null statuses
                        .filter { change -> change.participantStatus != null }
                        // only count updated statuses
                        .filter { change -> change.changeType == ChangeType_Update }
                        // switch to status instances
                        .map { change -> change.participantStatus }
                        // leave only in-call participants
                        .filter { participant -> participant.callStatus == CallStatus_InConference }

                    // ID -1 should strip video floor from everyone, other IDs should give video floor to participants
                    val videoFloorChanged = if (participantId == -1) {
                        inCallParticipants
                            .filter { participant -> participant.hasVideoFloor }
                            .count()
                            .blockingGet() == 0L
                    } else {
                        inCallParticipants
                            .filter { participant -> participant.hasVideoFloor && participant.participantNumber == participantId }
                            .count()
                            .blockingGet() == 1L
                    }
                    if (!videoFloorChanged) {
                        throw IllegalStateException("Video floor hasn't changed")
                    }
                }
            }

            override fun onMixerOptionsCommandFailure(api: SipVccsConferenceApi?, error: OnMixerOptionsCommandFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$subscribeFailureErrorCode]: $subscribeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.setVideoFloorParticipant(accountId, conferenceId, participantId)
        }
    }

    override fun setVideoLayout(accountId: Int, conferenceId: Int, layout: VideoLayout): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onConferenceListUpdated(api: SipVccsConferenceApi?, event: OnConferenceListUpdated?): Int {
                return reactToCompletableSuccess(event, VIDEO_LAYOUT_EVENT_MISSING) { listEvent ->
                    if (accountId != event?.vccsAccountHandle) throw IllegalArgumentException("Invalid account ID")
                    val videoLayoutChanged = withDetails(conferenceId, listEvent) { it.currentVideoLayout == layout.value } ?: false
                    if (!videoLayoutChanged) throw IllegalStateException("Video layout hasn't changed!")
                }
            }

            override fun onMixerOptionsCommandFailure(api: SipVccsConferenceApi?, error: OnMixerOptionsCommandFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$subscribeFailureErrorCode]: $subscribeFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.setVideoLayout(accountId, conferenceId, layout)
        }
    }

    override fun kickParticipant(accountId: Int, conferenceId: Int, participantId: Int): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onParticipantListUpdated(api: SipVccsConferenceApi?, event: OnParticipantListUpdated?): Int {
                return reactToCompletableSuccess(event, PARTICIPANT_KICK_EVENT_MISSING) {
                    val participantKicked = Observable
                        .fromIterable(it.participantListUpdatedEvent.participantListUpdatedEvent)
                        // only work with non-null statuses
                        .filter { change -> change.participantStatus != null }
                        // switch to status instances
                        .map { change -> change.participantStatus }
                        // leave only the matching participant by ID
                        .filter { participant -> participant.participantNumber == participantId }
                        // leave only kicked participants
                        .filter { participant -> participant.callStatus == CallStatus_Kicked }
                        .count()
                        .blockingGet() == 1L
                    if (!participantKicked) {
                        throw IllegalStateException("Participant wasn't kicked")
                    }
                }
            }

            override fun onKickParticipantFailure(api: SipVccsConferenceApi?, error: OnKickParticipantFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onKickParticipantFailureErrorCode]: $onKickParticipantFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.KickParticipant(accountId, conferenceId, participantId)
        }
    }

    override fun muteParticipant(accountId: Int, conferenceId: Int, participantId: Int, mute: Boolean): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onParticipantListUpdated(api: SipVccsConferenceApi?, event: OnParticipantListUpdated?): Int {
                return reactToCompletableSuccess(event, PARTICIPANT_MUTE_EVENT_MISSING) {
                    val participantMuteChanged = Observable
                        .fromIterable(it.participantListUpdatedEvent.participantListUpdatedEvent)
                        // only work with non-null statuses
                        .filter { change -> change.participantStatus != null }
                        // only count updated statuses
                        .filter { change -> change.changeType == ChangeType_Update }
                        // switch to status instances
                        .map { change -> change.participantStatus }
                        // leave only in-call participants
                        .filter { participant -> participant.callStatus == CallStatus_InConference }
                        // leave only the matching participant by ID
                        .filter { participant -> participant.participantNumber == participantId }
                        // all that remains is the one participant we wanted, check mute state
                        .all { participant -> participant.isMuted == mute }
                        .blockingGet()
                    if (!participantMuteChanged) {
                        throw IllegalStateException("Participant wasn't ${if (mute) "muted" else "unmuted"}")
                    }
                }
            }

            override fun onMuteParticipantFailure(api: SipVccsConferenceApi?, error: OnMuteParticipantFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onMuteParticipantFailureErrorCode]: $onMuteParticipantFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.MuteParticipant(accountId, conferenceId, participantId, mute)
        }
    }

    override fun setParticipantRecording(accountId: Int, conferenceId: Int, participantId: Int, record: Boolean): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onParticipantListUpdated(api: SipVccsConferenceApi?, event: OnParticipantListUpdated?): Int {
                return reactToCompletableSuccess(event, PARTICIPANT_REC_EVENT_MISSING) {
                    val participantRecordingChanged = Observable
                        .fromIterable(it.participantListUpdatedEvent.participantListUpdatedEvent)
                        // only work with non-null statuses
                        .filter { change -> change.participantStatus != null }
                        // only count updated statuses
                        .filter { change -> change.changeType == ChangeType_Update }
                        // switch to status instances
                        .map { change -> change.participantStatus }
                        // leave only in-call participants
                        .filter { participant -> participant.callStatus == CallStatus_InConference }
                        // leave only the matching participant by ID
                        .filter { participant -> participant.participantNumber == participantId }
                        // all that remains is the one participant we wanted, check mute state
                        .all { participant -> participant.isRecording == record }
                        .blockingGet()
                    if (!participantRecordingChanged) {
                        throw IllegalStateException("Participant wasn't flagged for recording")
                    }
                }
            }

            override fun onSetIsRecordingFailure(api: SipVccsConferenceApi?, error: OnSetIsRecordingFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onSetIsRecordingFailureErrorCode]: $onSetIsRecordingFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.SetIsRecording(accountId, conferenceId, participantId, record)
        }
    }

    override fun getXmppAccountInfo(accountId: Int, conferenceId: Int, guestMode: Boolean): Single<EventHolder<XMPPAccountInfoEvent>> {
        val singleListener = object : ConferenceListenerSingle<EventHolder<XMPPAccountInfoEvent>>(accountId, conferenceId) {
            override fun onXMPPAccountInfo(api: SipVccsConferenceApi?, event: OnXMPPAccountInfo?): Int {
                return reactToSingleSuccess(event, XMPP_INFO_EVENT_MISSING) {
                    EventHolder(accountId, conferenceId, it.xmppAccountInfoEvent)
                }
            }

            override fun onXMPPAccountInfoFailure(api: SipVccsConferenceApi?, error: OnXMPPAccountInfoFailure?): Int {
                return reactToSingleError(error) {
                    "Error [$onXMPPAccountInfoFailureErrorCode]: $onXMPPAccountInfoFailureReason"
                }
            }
        }

        return createSingle(singleListener) { api: SipVccsConferenceApi ->
            api.GetXMPPAccountInfo(accountId, conferenceId, guestMode)
        }
    }

    override fun getConferenceInvite(accountId: Int, conferenceId: Int): Single<EventHolder<OnQueryConferenceInvite>> {
        val singleListener = object : ConferenceListenerSingle<EventHolder<OnQueryConferenceInvite>>(accountId, conferenceId) {
            override fun onQueryConferenceInvite(api: SipVccsConferenceApi?, event: OnQueryConferenceInvite?): Int {
                return reactToSingleSuccess(event, INVITE_EVENT_MISSING) {
                    EventHolder(accountId, conferenceId, it)
                }
            }

            override fun onQueryConferenceInviteFailure(api: SipVccsConferenceApi?, error: OnQueryConferenceInviteFailure?): Int {
                return reactToSingleError(error) {
                    "Error [$onXMPPAccountInfoFailureErrorCode]: $onXMPPAccountInfoFailureReason"
                }
            }
        }

        return createSingle(singleListener) { api: SipVccsConferenceApi ->
            api.QueryConferenceInvite(accountId, conferenceId)
        }
    }

    override fun setScreenSharePresenter(accountId: Int, conferenceId: Int, participantId: Int): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onSetScreenSharePresenter(api: SipVccsConferenceApi?, event: OnSetScreenSharePresenter?): Int {
                return reactToCompletableSuccess(event, SCREENSHARE_PRESENTER_EVENT_MISSING) {
                    if (it.vccsConferenceParticipantHandle != participantId) {
                        throw IllegalStateException("Screenshare presenter wasn't changed!")
                    }
                }
            }

            override fun onScreenShareCommandFailure(api: SipVccsConferenceApi?, error: OnScreenShareCommandFailure?): Int {
                return reactToCompletableError(error) {
                    "Participant [$vccsConferenceParticipantHandle]: $vccsConferenceParticipantFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.SetScreenSharePresenter(accountId, conferenceId, participantId)
        }
    }

    override fun getConferenceConfig(accountId: Int, conferenceId: Int): Single<EventHolder<ConferenceConfigEvent>> {
        val singleListener = object : ConferenceListenerSingle<EventHolder<ConferenceConfigEvent>>(accountId, conferenceId) {
            override fun onQueryConferenceConfig(api: SipVccsConferenceApi?, event: OnQueryConferenceConfig?): Int {
                return reactToSingleSuccess(event, FETCH_CONFIG_EVENT_MISSING) {
                    EventHolder(accountId, conferenceId, it.conferenceConfigEvent)
                }
            }

            override fun onQueryConferenceConfigFailure(api: SipVccsConferenceApi?, error: OnQueryConferenceConfigFailure?): Int {
                return reactToSingleError(error) {
                    "Error [$onQueryConferenceConfigFailureErrorCode]: $onQueryConferenceConfigFailureReason"
                }
            }
        }

        return createSingle(singleListener) { api: SipVccsConferenceApi ->
            api.queryConferenceConfig(accountId, conferenceId)
        }
    }

    override fun setConferenceConfig(accountId: Int, conferenceId: Int, values: ConferenceConfigValues, flags: ConferenceConfigFlags): Completable {
        val completableListener = object : ConferenceListenerCompletable(accountId, conferenceId) {
            override fun onSetConferenceConfig(api: SipVccsConferenceApi?, event: OnSetConferenceConfig?): Int {
                return reactToCompletableSuccess(event, SET_CONFIG_EVENT_MISSING)
            }

            override fun onSetConferenceConfigFailure(api: SipVccsConferenceApi?, error: OnSetConferenceConfigFailure?): Int {
                return reactToCompletableError(error) {
                    "Error [$onSetConferenceConfigFailureErrorCode]: $onSetConferenceConfigFailureReason"
                }
            }
        }

        return createCompletable(completableListener) { api: SipVccsConferenceApi ->
            api.setConferenceConfig(
                    // conference identifiers
                    accountId, conferenceId,
                    // setting values
                    values.participantPin, values.moderatorPin, values.videoLayout, values.frameRate, values.isModerated,
                    values.overrunMinutes, values.dropBoxToken, values.sendSummaryEmail, values.recordAudioOnly, values.autoRecord,
                    // persistence flags
                    flags.participantPin, flags.moderatorPin, flags.videoLayout, flags.frameRate, flags.isModerated,
                    flags.overrunMinutes, flags.dropBoxToken, flags.sendSummaryEmail, flags.recordAudioOnly, flags.autoRecord
            )
        }
    }

    override fun observeParticipantChanges(accountId: Int, conferenceId: Int): Observable<EventHolder<ParticipantChanges>> {
        val observableListener = object : ConferenceListenerObservable<EventHolder<ParticipantChanges>>(accountId, conferenceId) {
            override fun onParticipantListUpdated(api: SipVccsConferenceApi?, event: OnParticipantListUpdated?): Int {
                return reactToObservableEvent(event, PARTICIPANT_CHANGE_EVENT_MISSING) {
                    Observable
                        .fromIterable(it.participantListUpdatedEvent.participantListUpdatedEvent)
                        // only work with non-null statuses
                        .filter { change -> change.participantStatus != null }
                        // switch to status instances
                        .map { change -> change.participantStatus }
                        .toList()
                        .map { list -> EventHolder(accountId, conferenceId, list) }
                        .blockingGet()
                }
            }
        }

        return createObservable(observableListener) { K_SUCCESS }
    }

    override fun observeScreenShareChanges(accountId: Int, conferenceId: Int): Observable<EventHolder<ScreenShareConfigEvent>> {
        val observableListener = object : ConferenceListenerObservable<EventHolder<ScreenShareConfigEvent>>(accountId, conferenceId) {
            override fun onScreenShareConfigChanged(api: SipVccsConferenceApi?, event: OnScreenShareConfigChanged?): Int {
                return reactToObservableEvent(event, SCREENSHARE_CHANGE_EVENT_MISSING) {
                    EventHolder(accountId, conferenceId, it.screenShareConfigEvent)
                }
            }
        }

        return createObservable(observableListener) { K_SUCCESS }
    }

    override fun observeVoiceActivityChanges(accountId: Int, conferenceId: Int): Observable<EventHolder<OnVoiceActivityChanged>> {
        val observableListener = object : ConferenceListenerObservable<EventHolder<OnVoiceActivityChanged>>(accountId, conferenceId) {
            override fun onVoiceActivityChanged(api: SipVccsConferenceApi?, event: OnVoiceActivityChanged?): Int {
                return reactToObservableEvent(event, VOICE_ACTIVITY_CHANGE_EVENT_MISSING) {
                    EventHolder(accountId, conferenceId, it)
                }
            }
        }

        return createObservable(observableListener) { K_SUCCESS }
    }

    /* Private helpers */

    /**
     * Converts a callback-based handler into a reactive interface using [Observable].
     * @param listener The [Observable]-aware listener capable of emitting Rx events, sub-typed from [ConferenceListenerObservable]
     * @param apiCall Which API call to invoke on the original callback-based API
     * @return A new [Observable] that will properly convert traditional callback events to Rx events
     */
    private fun <E> createObservable(listener: ConferenceListenerObservable<E>, apiCall: (SipVccsConferenceApi) -> Int): Observable<E> {
        val accountId = listener.accountId
        val conferenceId = listener.conferenceId
        conferenceManager?.let { api ->
            // API call succeeded, register the listener
            return Observable.create { e ->
                listener.emitter = e
                val registration = api.addHandler(accountId, listener)
                e.setCancellable {
                    registration.removeHandler()
                    listener.emitter = null
                }
                // check if API call fails
                if (apiCall(api) == K_ERROR) e.tryOnError(CollabConferenceException(accountId, conferenceId, ERROR_ACTION_FAILED))
            }
        }
        // API instance was null
        return Observable.error(CollabConferenceException(accountId, conferenceId, ERROR_API_MISSING))
    }

    /**
     * Converts a callback-based handler into a reactive interface using [Single].
     * @param listener The [Single]-aware listener capable of emitting Rx events, sub-typed from [ConferenceListenerSingle]
     * @param apiCall Which API call to invoke on the original callback-based API
     * @return A new [Single] that will properly convert traditional callback events to Rx events
     */
    private fun <E> createSingle(listener: ConferenceListenerSingle<E>, apiCall: (SipVccsConferenceApi) -> Int): Single<E> {
        val accountId = listener.accountId
        val conferenceId = listener.conferenceId
        conferenceManager?.let { api ->
            // API call succeeded, register the listener
            return Single.create { e ->
                listener.emitter = e
                val registration = api.addHandler(accountId, listener)
                e.setCancellable {
                    registration.removeHandler()
                    listener.emitter = null
                }
                // check if API call fails
                if (apiCall(api) == K_ERROR) e.tryOnError(CollabConferenceException(accountId, conferenceId, ERROR_ACTION_FAILED))
            }
        }
        // API instance was null
        return Single.error(CollabConferenceException(accountId, conferenceId, ERROR_API_MISSING))
    }

    /**
     * Converts a callback-based handler into a reactive interface using [Completable].
     * @param listener The [Completable]-aware listener capable of emitting Rx events, sub-typed from [ConferenceListenerCompletable]
     * @param apiCall Which API call to invoke on the original callback-based API
     * @return A new [Completable] that will properly convert traditional callback events to Rx events
     */
    private fun createCompletable(listener: ConferenceListenerCompletable, apiCall: (SipVccsConferenceApi) -> Int): Completable {
        val accountId = listener.accountId
        val conferenceId = listener.conferenceId
        conferenceManager?.let { api ->
            // API call succeeded, register the listener
            return Completable.create { e ->
                listener.emitter = e
                val registration = api.addHandler(accountId, listener)
                e.setCancellable {
                    registration.removeHandler()
                    listener.emitter = null
                }
                // check if API call fails
                if (apiCall(api) == K_ERROR) e.tryOnError(CollabConferenceException(accountId, conferenceId, ERROR_ACTION_FAILED))
            }
        }
        // API instance was null
        return Completable.error(CollabConferenceException(accountId, conferenceId, ERROR_API_MISSING))
    }

    /**
     * Picks a [ConferenceListChange] from the list updated event (provided the conference ID), and
     * converts it to a return type using the given `converter` function.
     * @param conferenceId The conference ID to look for in the change list
     * @param event The original conference list updated event (this contains all changes and details)
     * @param converter A function to convert from [ConferenceDetails] to a return type of your choosing
     * @param R The return type of your choosing
     * @return A `nullable` variant of the `<R>` return type
     */
    private fun <R : Any?> withDetails(conferenceId: Int, event: OnConferenceListUpdated, converter: (ConferenceDetails) -> R?): R? {
        return if (event.conferenceListUpdatedEvent.conferenceListChangeCount > 0) {
            val details: ConferenceDetails? = Observable
                .fromIterable(event.conferenceListUpdatedEvent.conferenceListChange)
                .filter { it.conferenceDetails.id == conferenceId.toLong() }
                .map { it -> it.conferenceDetails }
                .blockingFirst(null)
            if (details != null) converter(details) else null
        } else null
    }

}
