package com.bria.common.controller.collaboration.rx.rxlisteners

import android.util.Log
import com.bria.common.controller.collaboration.rx.api.ISipVccsConferenceApi.Companion.K_ERROR
import com.bria.common.controller.collaboration.rx.api.ISipVccsConferenceApi.Companion.K_SUCCESS
import com.bria.common.controller.collaboration.rx.exceptions.CollabConferenceException
import com.counterpath.sdk.handler.SipVccsConferenceHandler.SipVccsConferenceAdapter
import io.reactivex.MaybeEmitter

/**
 * A listener adapter derived from [SipVccsConferenceAdapter] that is capable of emitting through an Rx emitter.
 * Whenever a traditional callback happens: if it's a successful one, you should call [reactToMaybeSuccess];
 * if it's an unsuccessful one (i.e. an error), call [reactToMaybeError].
 *
 * @param accountId The account ID to initialize the listener with
 * @param conferenceId The conference ID to initialize the listener with, pass [K_ERROR] if unavailable
 * @param E The emitted event type
 *
 * @see [reactToMaybeSuccess]
 * @see [reactToMaybeError]
 */
open class ConferenceListenerMaybe<E>(val accountId: Int, val conferenceId: Int) : SipVccsConferenceAdapter() {

    private companion object {
        val TAG = "ConferenceListenerMaybe"!!
        const val ERROR_EVENT_MISSING = "Missing event: Error"
        const val ERROR_EVENT_CONVERSION_FAILED = "Event conversion failed"
    }

    @Volatile
    var emitter: MaybeEmitter<E>? = null
        @Synchronized get() = field
        @Synchronized set(emitter) {
            field = emitter
        }

    /**
     * Provides a generic reaction to a successful callback event from a previous API call. Emits to the local emitter.
     * @param event The original event from the API callback, can be `null`
     * @param missingMessage A message to emit when event is not available
     * @param eventConverter A way to convert the original event type into the emittable type, function
     * @return Either [K_SUCCESS] or [K_ERROR], depending on whether the event can be processed
     */
    @Synchronized
    fun <T> reactToMaybeSuccess(event: T?, missingMessage: String, eventConverter: (T) -> E?): Int {
        emitter?.let { emitter ->
            // if disposed, don't handle anything
            if (emitter.isDisposed) return K_ERROR
            // check if event is available
            event?.let {
                // event is available, convert it into the appropriate event type
                val converted = eventConverter(it)
                converted?.let {
                    emitter.onSuccess(it)
                    // MM emit complete?
                    return K_SUCCESS
                }
                // conversion failed, generate a 'conversion failed' error
                return emitter.onErrorSafe(CollabConferenceException(accountId, conferenceId, ERROR_EVENT_CONVERSION_FAILED))
            }
            // event is unavailable, generate a 'missing event' error
            return emitter.onErrorSafe(CollabConferenceException(accountId, conferenceId, missingMessage))
        }
        // emitter is not available
        return K_ERROR
    }

    /**
     * Provides a generic reaction to a failure (unsuccessful callback event) from a previous API call. Emits to the local emitter.
     * @param error The original error event from the API callback, can be `null`
     * @param lazyMessage A way to construct the error message given the original error event, function
     * @return Either [K_SUCCESS] or [K_ERROR], depending on whether the error can be processed
     */
    @Synchronized
    fun <T> reactToMaybeError(error: T?, lazyMessage: T.() -> String): Int {
        emitter?.let { emitter ->
            // if disposed, don't handle anything
            if (emitter.isDisposed) return K_ERROR
            // check if error is available
            error?.let {
                // error is available, construct the message and generate an error
                return emitter.onErrorSafe(CollabConferenceException(accountId, conferenceId, it.lazyMessage()))
            }
            // error is unavailable, generate a 'missing event' error
            return emitter.onErrorSafe(CollabConferenceException(accountId, conferenceId, ERROR_EVENT_MISSING))
        }
        // emitter is not available
        return K_ERROR
    }

    @Synchronized
    private fun MaybeEmitter<E>.onErrorSafe(exception: Exception): Int {
        return try {
            if (!tryOnError(exception)) {
                Log.d(TAG, "Failed to deliver: $exception")
                K_ERROR
            } else K_SUCCESS
        } catch (error: Throwable) {
            println("${javaClass.simpleName} | Error while emitting exception: $exception")
            K_ERROR
        }
    }

}
