package com.bria.common.controller.contacts.buddy.repository

import android.annotation.SuppressLint
import android.text.TextUtils
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.contacts.buddy.VCard
import com.bria.common.controller.contacts.buddy.XmppBuddy
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.util.Log
import com.bria.common.util.im.Participant
import com.counterpath.sdk.XmppVCard
import com.counterpath.sdk.handler.XmppVCardHandler
import com.counterpath.sdk.pb.Xmppvcard
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import io.reactivex.subjects.Subject


class BuddyRepositoryImpl(private val buddyProviderDb: BuddyProviderDb) : BuddyRepository, XmppVCardHandler {

    var observableDataChanged: Subject<Any> = PublishSubject.create()
    val disposable = CompositeDisposable()
    private val TAG = "BuddyRepositoryImpl"

    val mActiveBuddies = HashMap<String, Buddy>()

    override fun addBuddyEntity(buddy: BuddyEntity) {
        disposable.add(Completable.fromAction {
            buddyProviderDb.buddyDao().insert(buddy)
            //buddyProviderDb.phoneDao().insertAll(buddy.phones)
        }.subscribeOn(Schedulers.io())
                .observeOn(Schedulers.io()).subscribe({ observableDataChanged.onNext("Added") }, { throwable -> Log.e(TAG, "OnError $throwable") }))

    }

    override fun deleteBuddy(buddy: BuddyEntity) {
        disposable.add(Completable.fromAction {
            buddyProviderDb.buddyDao().delete(buddy)
        }.subscribeOn(Schedulers.io())
                .observeOn(Schedulers.io()).subscribe({ observableDataChanged.onNext("Deleted") }, { throwable -> Log.e(TAG, "OnError $throwable") }))
    }

    //no phones added
    override fun getBuddies(): List<BuddyEntity> {
        return buddyProviderDb.buddyDao().getAllBuddies()
    }

    override fun filterBuddiesBy(searchString: String): List<BuddyEntity> {
        return buddyProviderDb.buddyDao().getAllBuddiesWith(searchString)
    }

    override fun getBuddyEntityByKey(key: String): BuddyEntity? {
        val list = buddyProviderDb.buddyDao().getBuddyByKey(key)
        if (list.isNotEmpty()) {
            val buddy = list[0]
            buddy.phones = buddyProviderDb.phoneDao().getAllPhonesForBuddy(buddy.key)
            return buddy
        } else return null
    }

    override fun onVCardOperationResult(xmppVcard: XmppVCard?, evt: Xmppvcard.XmppVCardEvents.VCardOperationResultEvent?) {

    }

    // nvtd: track account settings changes and add hardwired flag to all buddies for that account

    @SuppressLint("RxLeakedSubscription", "RxSubscribeOnError")
    override fun onVCardFetched(xmppVcard: XmppVCard?, evt: Xmppvcard.XmppVCardEvents.VCardFetchedEvent?) {
        val acc = xmppVcard?.api?.account;
        val vcardDetail = evt?.detail ?: return
        val userAtDomain = acc?.settings?.username + "@" + acc?.settings?.domain
        val key = BuddyKeyUtils.getNewBuddyKey(Participant.EParticipantType.XMPP, userAtDomain, evt.jid)
        var firstName = vcardDetail.name?.given
        var lastName = vcardDetail.name?.family
        val formattedName = vcardDetail.formattedname
        if (TextUtils.isEmpty(firstName) && !TextUtils.isEmpty(formattedName) && formattedName.contains(" ")) {
            firstName = formattedName?.substring(0, formattedName.indexOf(" "))
        }
        if (TextUtils.isEmpty(lastName) && !TextUtils.isEmpty(formattedName) && formattedName.contains(" ")) {
            lastName = formattedName?.substring(formattedName.indexOf(" "), formattedName.length)
        }

        var email = ""
        for (anEmail in vcardDetail.getEmailListList()) {
            when {
                anEmail.x400 -> email = anEmail.userid
                anEmail.pref -> email = anEmail.userid
                anEmail.internet -> email = anEmail.userid
                anEmail.work -> email = anEmail.userid
                anEmail.home -> email = anEmail.userid
            }
        }
        val phones: ArrayList<PhoneEntity> = ArrayList(vcardDetail.telephoneListList.size)
        for (phone in vcardDetail.telephoneListList) {
            if (TextUtils.isEmpty(phone.number)) {
                continue
            }

            val phoneType: VCard.EPhoneType = when {
                phone.fax -> VCard.EPhoneType.Fax
                phone.pager -> VCard.EPhoneType.PagerVCard
                phone.cell -> VCard.EPhoneType.Cell
                phone.video -> VCard.EPhoneType.Video
                phone.bbs -> VCard.EPhoneType.Bbs
                phone.modem -> VCard.EPhoneType.Modem
                phone.isdn -> VCard.EPhoneType.Isdn
                phone.pcs -> VCard.EPhoneType.Pcs
                phone.home -> VCard.EPhoneType.Home
                phone.work -> VCard.EPhoneType.Work
                phone.voice -> VCard.EPhoneType.Voice

                else -> VCard.EPhoneType.Home
            }
            phones.add(PhoneEntity(buddyKey = key!!, phoneType = phoneType, phoneNumber = phone.number, preferred = phone.pref))
        }


        val buddyEntity = BuddyEntity(key = key,
                firstName = firstName!!,
                lastName = lastName!!,
                formattedName = formattedName!!,
                type = EBuddyType.XMPP,
                organization = vcardDetail.organization.toString(),
                email = email,
                url = vcardDetail.url,
                image = vcardDetail.photo.binval,
                isActive = true
        )
        buddyEntity.phones = phones
        addBuddyEntity(buddyEntity)
    }

    override fun addBuddy(buddy: Buddy) {
        mActiveBuddies[buddy.uniqueIdentifier] = buddy
        addBuddyToStorage(buddy)
    }

    override fun removeBuddyByKey(buddyKey: String) {

    }

    override fun getBuddyByKey(key: String): Buddy? {
        return null
    }

    override fun getObservableOnBuddyListChanged(): Observable<Any> = observableDataChanged


    override fun onError(p0: XmppVCard?, p1: Xmppvcard.XmppVCardEvents.ErrorEvent?) {
    }


    override fun destroy() {
        buddyProviderDb.close()
    }


    private fun addBuddyToStorage(buddy: Buddy) {
        if (buddy is XmppBuddy) {
            val phones = ArrayList<PhoneEntity>()
            buddy.vCard.phoneList.forEach { phone ->
                phones.add(
                        PhoneEntity(buddyKey = buddy.uniqueIdentifier, phoneType = phone.phoneType, preferred = phone.isPreferred, phoneNumber = phone.number))
            }
            val buddyEntity = BuddyEntity(key = buddy.uniqueIdentifier,
                    lastName = buddy.lastName,
                    firstName = buddy.firstName,
                    formattedName = buddy.displayName,
                    type = EBuddyType.XMPP,
                    image = buddy.vCard.rawAvatar,
                    organization = buddy.vCard.organization,
                    email = buddy.vCard.email,
                    url = buddy.vCard.url, isActive = true)
            buddyEntity.phones = phones
            addBuddyEntity(buddyEntity)
        }
    }
}
