package com.bria.common.controller.im.refactoring

import com.bria.common.controller.im.IImRemoteStorage
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.im.refactoring.db.entities.InstantMessageData
import com.bria.common.controller.remotesync.RemoteSyncUtil
import com.bria.common.util.Log
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers

class RemoteStorageImDataObserver(val imData: IImData, private val remoteStorage: IImRemoteStorage) {

    val TAG = "RemoteStorageImDataObserver"

    private val disposables: CompositeDisposable = CompositeDisposable()

    private var conversationsTempList = ArrayList<ImConversationData>()

    init {
        createRemoteSourceObservers()
    }

    fun destroy() {
        disposables.dispose()
    }

    private fun createRemoteSourceObservers() {
        if (RemoteSyncUtil.checkFeature()) {
            disposables.clear()
            disposables.add(imData.observableOnListMessagesAdded.subscribeOn(Schedulers.io())
                    .subscribe({ addedList -> remoteSyncMessage(addedList) }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) }))

            disposables.add(imData.observableOnListMessagesRemoved.subscribeOn(Schedulers.io())
                    .subscribe({ removedList -> remoteRemoveMessage(removedList) }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) }))

            disposables.add(imData.observableOnConversationAdded.subscribeOn(Schedulers.io())
                    .subscribe({ conversation -> remoteAddConversation(conversation) }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) }))

            disposables.add(imData.observableOnConversationUpdated.subscribeOn(Schedulers.io())
                    .subscribe({ conversation -> remoteUpdateConversation(conversation) }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) }))

            disposables.add(imData.observableOnConversationRemove.subscribeOn(Schedulers.io())
                    .subscribe({ conversation -> remoteDeleteConversation(conversation) }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) }))

            disposables.add(imData.observableOnListMessageUpdated.subscribeOn(Schedulers.io())
                    .subscribe({ updatedList -> remoteUpdateMessage(updatedList) }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) }))

            disposables.add(imData.observableOnMessagesRead.subscribeOn(Schedulers.io())
                    .subscribe({ list -> remoteMarkMessagesRead(list) }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) }))
        }
    }

    private fun remoteSyncMessage(listIm: List<InstantMessageData>) {
        listIm.filter { it.syncMe() && it.isSyncable }.forEach { remoteStorage.syncNewMessage(it) }
    }

    private fun remoteRemoveMessage(listIm: List<InstantMessageData>) {
        listIm.forEach { im ->
            if (im.syncMe()) {
                im.deleted = true
                remoteStorage.updateMessage(im)
            }
        }
    }

    private fun remoteMarkMessagesRead(listIm: List<InstantMessageData>) {
        listIm.asSequence().filter { it.syncMe() }.forEach { remoteStorage.updateMessage(it) }
    }

    private fun remoteUpdateConversation(conversation: ImConversationData) {
        if (conversation.syncMe()) {
            remoteStorage.updateImConversation(conversation)
        }
    }

    private fun remoteAddConversation(conversation: ImConversationData) {
        if (conversation.syncMe()) {
            remoteStorage.updateImConversation(conversation)
        }
    }

    private fun remoteDeleteConversation(conversation: ImConversationData) {
        if (conversation.syncMe()) {
            conversation.isDeleted = true
            remoteStorage.updateImConversation(conversation)
        }
    }

    private fun remoteUpdateMessage(listIm: List<InstantMessageData>) {
        listIm.asSequence().filter { it.syncMe() && it.isSyncable }
                .forEach {
                    if (!it.isAlreadySynced) {
                        remoteStorage.syncNewMessage(it)
                    } else {
                        remoteStorage.updateMessage(it)
                    }
                }
    }
}
