package com.bria.common.controller.im.refactoring.chatroom

import android.content.Context
import android.text.TextUtils
import android.util.Log
import com.bria.common.controller.accounts.IAccountsCtrlActions
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.ChatApi
import com.bria.common.controller.im.refactoring.ChatApi.LOCAL_MESSAGE_SEPARATOR
import com.bria.common.controller.im.refactoring.ChatApi.PARTICIPANT_LOCAL_JOINED_EVENT_ID
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomCache.Companion.MEMBERS_LIST_TYPE
import com.bria.common.controller.im.refactoring.db.entities.ImStatusData
import com.bria.common.controller.im.refactoring.roomdb.ChatRoomRepo
import com.bria.common.controller.im.refactoring.roomdb.LocalRoomInfo
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoomState
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.sdkwrapper.SipStackManager
import com.bria.common.uireusable.datatypes.ExtendedConversationData
import com.bria.common.util.ThreadExecutors
import com.bria.common.util.im.Participant
import com.bria.common.util.im.ParticipantsSet
import com.counterpath.sdk.XmppAccount
import com.counterpath.sdk.XmppApiMultiUserChat
import com.counterpath.sdk.XmppMultiUserChat
import com.counterpath.sdk.handler.XmppMultiUserChatHandler
import com.counterpath.sdk.pb.Xmppmultiuserchat
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import java.text.SimpleDateFormat
import java.util.*

data class ChatRoomCacheItem(var chatRoom: ChatRoom, var xmppChat: XmppMultiUserChat, var bookmarked: Boolean = false)

class ChatRoomCache {

    companion object {
        const val MEMBERS_LIST_TYPE = 3
    }

    val items = mutableSetOf<ChatRoomCacheItem>()

    fun getChatRoom(xmppChat: XmppMultiUserChat): ChatRoom? = items.firstOrNull { xmppChat == it.xmppChat }?.chatRoom

    fun updateChatRoom(chatRoom: ChatRoom) = items.firstOrNull { chatRoom.chatKey == it.chatRoom.chatKey }?.apply {
        this.chatRoom = chatRoom.copy()
    }

    fun getXmppChat(chatRoomKey: String): XmppMultiUserChat? = items.firstOrNull { chatRoomKey == it.chatRoom.chatKey }?.xmppChat

    fun add(xmppChat: XmppMultiUserChat, chatRoom: ChatRoom, bookmarked: Boolean = false) {
        items.add(ChatRoomCacheItem(chatRoom, xmppChat, bookmarked))
    }

    fun remove(xmppChat: XmppMultiUserChat) {
        items.firstOrNull { xmppChat == it.xmppChat }?.let { items.remove(it) }
    }

    fun remove(chatRoom: ChatRoom) {
        items.firstOrNull { chatRoom == it.chatRoom }?.let { items.remove(it) }
    }

    fun getAllNonBookmarked() = items.filter { !it.bookmarked }
    fun deleteChatRoom(room: ChatRoom) {
        val delete = mutableListOf<ChatRoomCacheItem>()
        items.firstOrNull { chatRoomCacheItem ->
            chatRoomCacheItem.chatRoom.id == room.id
        }?.let {
            delete.add(it)
        }
        items.removeAll(delete)
    }

    fun setBookmarked(name: String) {
        items.filter { it.chatRoom.name == name }.forEach { it.bookmarked = true }
    }
}

class ChatRoomApiImpl(val chatApi: ChatApi, val chatRoomRepo: ChatRoomRepo, val context: Context, val accountController: IAccountsCtrlActions) : ChatRoomApi {

    private val TAG = "ChatRoomApi"

    private val cache = ChatRoomCache()

    private val disposables = CompositeDisposable()

    private val tempMessages = mutableMapOf<Int, Message>()
    private val alreadySavedMessages = mutableListOf<String>()
    private val bookmarks = mutableListOf<Xmppmultiuserchat.RoomBookmark>()

    private val mServerForAccount = HashMap<String, String>()

    private val stackManager: SipStackManager?
        get() = chatApi.stackManager


    init {
        disposables.add(chatRoomRepo.getOnRoomAddedObservable().mergeWith(chatRoomRepo.getOnRoomUpdatedObservable()).subscribe({ chatRoomList ->
            chatRoomList.forEach { room ->
                cache.updateChatRoom(room)
            }
        }, {}))
        disposables.add(chatRoomRepo.getOnRoomAddedObservable().subscribe({
            it.forEach {
                val time = System.currentTimeMillis()
                if (!it.isUserOwner()) {
                    val message1 = chatRoomRepo.getLastMessageForChatRoom(it.id).blockingGet()
                    if (message1 != null) {
                        val message = Message(chatId = it.id, creationTime = time,
                                modificationTime = time,
                                status = ImStatusData.LOCAL,
                                text = "$PARTICIPANT_LOCAL_JOINED_EVENT_ID$LOCAL_MESSAGE_SEPARATOR${it.accountId}",
                                externalId = "$time")
                        chatRoomRepo.addMessage(message)
                    }
                }
            }
        }, {}))

        disposables.add(chatRoomRepo.getOnRoomDeletedObservable().subscribe({ chatRoomList ->
            chatRoomList.forEach { room ->
                cache.deleteChatRoom(room)
            }
        }, {}))
    }

    override fun createRoom(chatRoom: ChatRoom) {
        getXmppAccount(chatRoom.accountId)?.let {
            val chatApi = XmppApiMultiUserChat.get(it)
            chatApi.newChat()?.let { xmppChat ->
                xmppChat.join("", it.settings.username, "", "message:20", arrayListOf())
                xmppChat.requestConfigurations()
                cache.add(xmppChat = xmppChat, chatRoom = chatRoom)
                refreshBookmarks(xmppChat.api.account)
            }
        }
    }


    override fun isGroupChatCapable(buddyKey: String): Boolean {
        val accountKey = BuddyKeyUtils.getAccountFromNewBuddyKey(buddyKey)
        val userAtDomain = ImpsUtils.removePortFromDomain(accountKey)
        return mServerForAccount.containsKey(userAtDomain)
    }

    private fun refreshBookmarks(account: XmppAccount) {

        XmppApiMultiUserChat.get(account)?.getRoomBookmarks()
    }

    private fun getXmppAccount(accId: String): XmppAccount? {
        accountController.getAccountForUserAtDomain(accId, EAccountType.Xmpp)?.let {
            return stackManager?.registrationManager?.getXmppAccount(it)
        }
        return null
    }

    override fun sendMessage(chatRoom: ChatRoom, message: String) {
        val id = cache.getXmppChat(chatRoom.chatKey)?.sendMessage(message, "") ?: -1
        if (id == -1) {
            Log.d(TAG, "sendMessage: Chat room not found in cache")
            return
        }
        val time = Date().time
        val msg = Message(
                chatId = chatRoom.id,
                status = ImStatusData.OUTGOING_PENDING,
                text = message,
                remoteAddress = chatRoom.chatKey,
                modificationTime = time,
                creationTime = time)
        tempMessages[id] = msg
    }

    private fun joinBookmarkedRooms() {
        //nvtd implement(get room by name from db and join that room
    }

    override fun joinRooms(roomsJid: List<String>, account: XmppAccount) {
        ThreadExecutors.scheduleLightWork {
            val chatApi = XmppApiMultiUserChat.get(account)
            roomsJid.forEach { jid ->
                val chatRoom = ChatRoom(
                        type = ExtendedConversationData.CHAT_ROOM_TYPE.toLong(),
                        chatKey = jid,
                        accountId = account.getUserAtDomain()
                )
                chatApi.newChat()?.let { xmppChat ->
                    xmppChat.join(chatRoom.chatKey, account.settings.username, "", "message:20", arrayListOf())
                    cache.add(xmppChat = xmppChat, chatRoom = chatRoom)
                    Log.d(TAG, "Room joining: " + chatRoom.chatKey)

                }
            }
        }
    }

    override fun destroyRoom(jid: String) {
        Log.d(TAG, "destroyRoom: ${jid}")
        cache.getXmppChat(jid)?.destroy("", "", "")
    }

    override fun leaveRoom(jid: String) {
        Log.d(TAG, "destroyRoom: ${jid}")
        cache.getXmppChat(jid)?.leave("")
    }

    override fun getChatHandler(): XmppMultiUserChatHandler = handler

    val handler: XmppMultiUserChatHandler = object : XmppMultiUserChatHandler {

        override fun onRoomBookmarksReceived(chat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.RoomBookmarksReceivedEvent?) {
            evt?.let { evt ->
                bookmarks.clear()
                bookmarks.addAll(evt.bookmarksList)
                bookmarks.forEach {
                    cache.setBookmarked(it.name)
                }
                bookmarkNonBookmarkedRooms(chat.api)
            }
            joinBookmarkedRooms()
        }

        override fun onServiceAvailability(chat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.ServiceAvailabilityEvent) {
            ///nvtd SDK UPDATE
            chat.api.getRoomBookmarks()
            chat.api.getRoomList()
            chat.api.newChat()
            mServerForAccount[chat.api.account.getUserAtDomain()] = evt.service
            joinPrivateChatRoomsFromStorage()
        }

        fun bookmarkNonBookmarkedRooms(api: XmppApiMultiUserChat) {
            val nonBookmarked = mutableListOf<Xmppmultiuserchat.RoomBookmark>()
            cache.getAllNonBookmarked().takeIf { it.isNotEmpty() }?.let { toBeAdded ->
                toBeAdded.forEach {
                    val nanoBookmark = com.counterpath.sdk.pb.nano.Xmppmultiuserchat.RoomBookmark()
                    nanoBookmark.autojoin = true
                    nanoBookmark.jid = it.xmppChat.api.account.getUserAtDomain()
                    nanoBookmark.name = it.chatRoom.name
                    nonBookmarked.add(Xmppmultiuserchat.RoomBookmark(nanoBookmark))
                }
            }

            if (nonBookmarked.isNotEmpty()) {
                api.setRoomBookmarks(nonBookmarked.toList())

                api.getRoomBookmarks()
            }
        }

        override fun onMultiUserChatConfigurationRequested(chat: XmppMultiUserChat,
                                                           evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatConfigurationRequestedEvent) {

            val chatRoom = cache.getChatRoom(chat)
            Log.d(TAG, "onMultiUserChatConfigurationRequested: " + evt.configurations.instructions)
            evt.xmppDataForm.fields.forEach {
                val values = it.values.joinToString(separator = "||")
                Log.d(TAG, "label: ${it.label} ;name: ${it.name}; required: ${it.required}; type: ${it.type}; values:${values}")
            }

            evt.xmppDataForm.fields.firstOrNull { it.name == "muc#roomconfig_persistentroom" }?.let { field ->
                if (field.values.isNotEmpty() && field.values[0] != "1") {
                    field.values[0] = "1"
                }
            }
            evt.xmppDataForm.fields.firstOrNull { it.name == "muc#roomconfig_roomname" }?.let { field ->
                if (field.values.isNotEmpty()) {
                    field.values[0] = chatRoom?.name ?: ""
                }
            }

            evt.xmppDataForm.fields.firstOrNull { it.name == "muc#roomconfig_roomdesc" }?.let { field ->
                if (field.values.isNotEmpty()) {
                    field.values[0] = chatRoom?.description ?: ""
                }
            }
            evt.xmppDataForm.fields.firstOrNull { it.name == "muc#roomconfig_publicroom" }?.let { field ->
                if (field.values.isNotEmpty()) {
                    field.values[0] = if (chatRoom?.privateRoom!! == 0) "1" else "0"
                }
            }
            chat.setConfigurations(evt.xmppDataForm)
        }

        override fun onMultiUserChatRoomStateChanged(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatRoomStateChangedEvent) {
            ThreadExecutors.scheduleLightWork {
                val state = msg.state
                cache.getChatRoom(chat)?.let {
                    chatRoomRepo.getRoomByRoomKey(it.chatKey)?.let { chatRoom ->
                        chatRoom.name = state.name
                        chatRoom.creationTime = formatTZStringDate(state.creation).time
                        chatRoom.description = state.description
                        chatRoom.privateRoom = if (state.isPublic) 0 else 1
                        chatRoomRepo.updateChatRoom(chatRoom)
                        chat.requestList(MEMBERS_LIST_TYPE)
                    }
                }
            }
        }

        override fun onRoomListRetrieved(account: XmppAccount, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.RoomListRetrievedEvent) {
            ThreadExecutors.scheduleLightWork {
                val roomDetails = StringBuilder("\n")
                roomDetails.append("RoomListRetrieved:\n ")
                val roomListInfos = mutableListOf<LocalRoomInfo>()
                val roomsInDb = chatRoomRepo.getAllChatRooms().blockingGet()
                val roomKeysInDb = roomsInDb.map { it.chatKey }
                Log.d(TAG, "ROOM_KEYS_IN_DB ${roomKeysInDb.joinToString()}")
                val chatApi = XmppApiMultiUserChat.get(account)
                msg.roomsList.map { roomListItem ->
                    roomDetails.append(" Room  name").append(roomListItem.name).append(" JID: ").append(roomListItem.jid).append("\n")
                    roomListInfos.add(LocalRoomInfo(roomListItem.name, account.getUserAtDomain(), ImpsUtils.removeDomainFromAddress(roomListItem.jid)))
                    roomListItem
                }.filter {
                    val jid = ImpsUtils.removeDomainFromAddress(it.jid)
                    roomKeysInDb.contains(jid) && cache.getXmppChat(jid) == null
                }
                        .map { room ->
                            val roomJid = ImpsUtils.removeDomainFromAddress(room.jid)
                            Log.d(TAG, "Room jid without Address: $roomJid")
                            roomsInDb.firstOrNull {
                                it.chatKey == roomJid
                            }
                        }
                        .forEach { chatRoom ->
                            chatRoom?.let { chatRoom ->
                                chatApi.newChat()?.let { xmppChat ->
                                    xmppChat.join(chatRoom.chatKey, account.settings.username, "", "message:20", arrayListOf())
                                    cache.add(xmppChat = xmppChat, chatRoom = chatRoom)
                                    Log.d(TAG, "Room joining: " + chatRoom.chatKey)

                                }
                            }

                        }
                Log.d(TAG, roomDetails.toString())
                chatRoomRepo.addRoomInfos(roomListInfos)
            }
        }

        override fun onMultiUserChatReady(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatReadyEvent) {
            //save to db for example
            chat.api?.getRoomBookmarks()
            chat.api?.getRoomList()
            Log.d(TAG, "onMultiUserChatReady: isNewRoom: ${msg.isNewRoom} ; room: ${msg.room}")
            ThreadExecutors.scheduleLightWork {
                cache.getChatRoom(chat)?.let { chatRoom ->
                    //if room is new
                    if (chatRoom.chatKey.isEmpty() || msg.isNewRoom) {
                        chatRoom.chatKey = msg.room
                        chatRoom.invited.forEach { member ->
                            if (!chatRoom.members.containsRemoteAddress(member.remoteAddress)) {
                                chat.invite(member.remoteAddress, "") // nvtd add invite message
                            }
                        }
                        chatRoom.invited = ParticipantsSet()
                    }
                    chatRoomRepo.addChatRoom(chatRoom = chatRoom)
                    chat.getRoomInfo()
                }
                chat.requestList(MEMBERS_LIST_TYPE)
            }
        }

        override fun onSendMessageSuccess(chat: XmppMultiUserChat?, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.SendMessageSuccessEvent) {
            Log.d(TAG, "onSendMessageSuccess ${evt.messageId}")
            tempMessages.get(evt.message)?.let {
                it.externalId = evt.messageId
                it.status = ImStatusData.OUTGOING_SENT
                alreadySavedMessages.add(evt.messageId)
                chatRoomRepo.addMessage(it)
            }
        }

        override fun onSendMessageFailure(chat: XmppMultiUserChat?, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.SendMessageFailureEvent) {
            tempMessages.get(evt.message)?.let {
                it.externalId = Date().time.toString()
                it.status = ImStatusData.OUTGOING_FLAG_FAIL_TO_SEND
                chatRoomRepo.addMessage(it)
            }
        }

        override fun onMultiUserChatNewMessage(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatNewMessageEvent) {
            ThreadExecutors.scheduleLightWork {
                //nvtd add support for collaboration chat
                cache.getChatRoom(chat)?.let {
                    val userName = chat.api.account.settings.username
                    val domain = chat.api.account.settings.domain
                    var remoteAddress = msg.nickname
                    if (!remoteAddress.contains("@")) {
                        remoteAddress = "$remoteAddress@$domain"
                    }
                    if (TextUtils.equals(remoteAddress, userName) || TextUtils.equals(remoteAddress, "$userName@$domain")) {
                        if (msg.messageId in alreadySavedMessages) {
                            alreadySavedMessages.remove(msg.messageId)
                            return@scheduleLightWork
                        }
                        chatRoomRepo.getRoomByRoomKey(key = it.chatKey)?.let { chatRoom ->
                            val msg = Message(
                                    chatId = chatRoom.id,
                                    status = ImStatusData.OUTGOING_SENT,
                                    text = msg.plain,
                                    read = false,
                                    remoteAddress = remoteAddress,
                                    externalId = msg.messageId,
                                    modificationTime = msg.timestamp * 1000,
                                    creationTime = msg.timestamp * 1000)
                            //nvtd recheck this, what if message is received without invitation?
                            chatRoomRepo.addMessage(msg)
                        }
                    } else {
                        chatRoomRepo.getRoomByRoomKey(key = it.chatKey)?.let { chatRoom ->
                            val msg = Message(
                                    chatId = chatRoom.id,
                                    status = ImStatusData.INCOMING_UNREAD,
                                    text = msg.plain,
                                    read = false,
                                    remoteAddress = remoteAddress,
                                    externalId = msg.messageId,
                                    modificationTime = msg.timestamp * 1000,
                                    creationTime = msg.timestamp * 1000)
                            //nvtd recheck this, what if message is received without invitation?
                            chatRoomRepo.addMessage(msg)
                        }
                    }
                }
            }
        }

        override fun onMultiUserChatError(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatErrorEvent?) {
            Log.e(TAG, msg?.error)
        }

        override fun onMultiUserChatInvitationReceived(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatInvitationReceivedEvent) {
            ThreadExecutors.scheduleLightWork {
                Log.d(TAG, "onMultiUserChatInvitationReceived: ${msg.room}")
                val xmppAccount = chat.api.account
                val chatRoom = ChatRoom(
                        jid = msg.jid,
                        chatKey = msg.room,
                        accountId = ImpsUtils.removePortFromDomain(xmppAccount.getUserAtDomain()),
                        creationTime = msg.timeStamp + msg.millisecond,
                        type = ExtendedConversationData.CHAT_ROOM_TYPE.toLong()
                )
                chatRoomRepo.addChatRoom(chatRoom)
                cache.add(chat, chatRoom)
                chat.accept(xmppAccount.settings.username, "message:300", arrayListOf())
                chat.getRoomInfo()
            }
        }

        override fun onParticipantAdded(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantAddedEvent) {
            ThreadExecutors.scheduleLightWork {
                Log.d(TAG, "onParticipantAdded ${msg.nickname} state: ${msg.state}")
            }

            chat.requestList(MEMBERS_LIST_TYPE)
        }

        override fun onParticipantUpdated(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantUpdatedEvent) {
            Log.d(TAG, "onParticipantUpdated ${msg.nickname} state: ${msg.state}")
        }


        override fun onMultiUserChatListRequested(chat: XmppMultiUserChat, args: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatListRequestedEvent) {
            ThreadExecutors.scheduleLightWork {
                if (args.xmppMultiUserChatListType == MEMBERS_LIST_TYPE) {
                    cache.getChatRoom(chat)?.let { cachedChatRoom ->
                        chatRoomRepo.getChatRoom(cachedChatRoom.id)?.let { chatRoom ->
                            Log.d(TAG, "Room : ${cachedChatRoom.name} list of members:")
                            chatRoom.members.clear()
                            args.configurationsListItemsList.forEach { item ->
                                val buddyKey = BuddyKeyUtils.getNewBuddyKey(Participant.EParticipantType.XMPP, chatRoom.accountId, item.jid)
                                chatRoom.members.add(Participant(buddyKey))
                                Log.d(TAG, "jid: ${item.jid}; nick: ${item.nick}; reason: ${item.reason}; aff ${item.affiliation}; role ${item.role} ")
                            }
                            chatRoomRepo.updateChatRoom(chatRoom)
                            cachedChatRoom.members.clear()
                            cachedChatRoom.members.addAll(chatRoom.members)
                        }

                    }
                }
            }

//            cache.getChatRoom(chat)?.let {
//                Log.d(TAG, "Room : ${it.name} list of members: ${args.xmppMultiUserChatListType}")
//
//                args.configurationsListItemsList.forEach { item ->
//                    Log.d(TAG, "jid: ${item.jid}; nick: ${item.nick}; reason: ${item.reason}; aff ${item.affiliation}; role ${item.role} ")
//                }
//            }
            //nvtd explore this
        }

        //--------------------------------

        override fun onLocalUserLeft(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.LocalUserLeftEvent) {
            Log.d(TAG, "OnLocalUserLeft")
            ThreadExecutors.scheduleLightWork {
                cache.getChatRoom(chat)?.let {
                    chatRoomRepo.getChatRoom(it.id)?.let { chatRoom ->
                        chatRoom.state = ChatRoomState.LEFT
                        chatRoomRepo.updateChatRoom(chatRoom)
                    }
                }
                cache.remove(chat)
            }
        }

        override fun onMultiUserChatSubjectChanged(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatSubjectChangedEvent?) {
            Log.d(TAG, "onMultiUserChatSubjectChanged")
        }

        override fun onParticipantRemoved(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantRemovedEvent?) {
            Log.d(TAG, "onParticipantRemoved")
        }


        override fun onMultiUserChatInvitationDeclined(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatInvitationDeclinedEvent?) {

        }

        override fun onParticipantSelfUpdated(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantSelfUpdatedEvent?) {

        }

        override fun onParticipantChatStateReceived(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantChatStateEvent?) {

        }
    }

    private fun joinPrivateChatRoomsFromStorage() {
        Log.d(TAG, "Join private Chat Rooms")
        disposables.add(chatRoomRepo.getAllPrivateChatRooms().subscribeOn(Schedulers.io()).observeOn(Schedulers.io()).subscribe({ chatRoomList ->
            chatRoomList.forEach { chatRoom ->
                getXmppAccount(chatRoom.accountId)?.let {
                    val chatApi = XmppApiMultiUserChat.get(it)
                    chatApi.newChat()?.let { xmppChat ->
                        xmppChat.join(chatRoom.chatKey, it.settings.username, "", "message:20", arrayListOf())
                        cache.add(xmppChat = xmppChat, chatRoom = chatRoom)
                    }
                }
            }
        }, { Log.e(TAG, "Join private Chat Rooms ERROR!!! ${it.localizedMessage}") }))
    }

    //formats "yyyy-MM-dd'T'HH:mm:ss'Z' date format to string
    private fun formatTZStringDate(stringDate: String): Date {
        val cal = Calendar.getInstance()
        val sdf = SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'")
        sdf.calendar = cal
        return sdf.parse(stringDate)
    }

}

fun XmppAccount.getUserAtDomain(): String {
    return "${this.settings.username}@${ImpsUtils.removePortFromDomain(this.settings.domain)}"
}
