package com.bria.common.controller.im.refactoring.dataprovider

import android.annotation.SuppressLint
import android.content.Context
import com.bria.common.controller.Controllers
import com.bria.common.controller.im.ImpsUtils.getLocalMidnightTime
import com.bria.common.controller.im.articles.ArticleExtractor
import com.bria.common.controller.im.refactoring.IChatApi
import com.bria.common.controller.im.refactoring.adapter.IDataProviderEvents
import com.bria.common.controller.im.refactoring.db.entities.*
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.uireusable.datatypes.MessageListItemData
import com.bria.common.uireusable.datatypes.MessageListItemData.Companion.MESSAGE_TYPE_DATE_DIVIDER
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import com.bria.common.util.ThreadExecutors
import com.bria.common.util.ThreadExecutors.scheduleLightWork
import io.reactivex.Observable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import java.util.*
import java.util.concurrent.atomic.AtomicBoolean
import kotlin.collections.ArrayList

class RxInstantMessageDataProvider(private val mContext: Context, val callback: IDataProviderEvents) {

    companion object {
        private const val TAG = "RxInstantMessageDataProvider"
        private const val ARTICLE_EXTRACTOR_CACHE_SIZE = 50
        private const val NEW_LIST_EVENT: Int = 0
        private const val MESSAGES_ADDED_EVENT: Int = 1
        private const val MESSAGES_REMOVED_EVENT: Int = 2
        private const val MESSAGES_UPDATED_EVENT: Int = 3
    }

    private var list = ArrayList<MessageListItemData>()

    var articleExtractor: ArticleExtractor

    private var conversationId: Long = ImConversationData.INVALID_CONVERSATION_ID

    private var isChatRoom = false

    private val selectedIms = ArrayList<Long>()

    private fun getImData() = Controllers.get().im.imData

    private val chatApi: IChatApi
        get() = Controllers.get().im.chatApi

    private fun numberOfLoadedMessages(): Int {
        var totalMessages = list.count { it.msgType != MessageListItemData.MESSAGE_TYPE_DATE_DIVIDER }
        if (totalMessages < GlobalConstants.REMOTE_SYNC_COUNT) {
            totalMessages = GlobalConstants.REMOTE_SYNC_COUNT
        }
        return totalMessages
    }

    var loading = AtomicBoolean(false)

    init {
        articleExtractor = ArticleExtractor(ARTICLE_EXTRACTOR_CACHE_SIZE, mContext.cacheDir.absolutePath)
    }

    fun clean() {
        Log.d(TAG, "Clean")
        conversationId = ImConversationData.INVALID_CONVERSATION_ID
        clearSelection()
        cleanData()
    }

    private fun cleanData() {
        Log.d(TAG, "CleanData")
        list.clear()
        sourceOfNewConversation.onNext(Pair(NEW_LIST_EVENT, ArrayList()))
    }

    fun setImConversationData(id: Long, hardwiredSession: Boolean, isChatRoom: Boolean) {
        Log.d(TAG, "Set Im Conversation: oldId: $conversationId, new id: $id")
        if (conversationId != id || isChatRoom != this.isChatRoom) {
            conversationId = id
            this.isChatRoom = isChatRoom
        }
        clearSelection()
        reloadConversationMessages()
        this.hardwiredSession = hardwiredSession

    }

    @SuppressLint("CheckResult")
    private fun reloadConversationMessages() {
        loading.set(true)
        callback.onListLoading()
        scheduleLightWork {
            val listOfMessages = if (isChatRoom) {
                roomDbListToUIList(getImData().chatRoomRepo.getNMessageWithOffsetForChat(conversationId, 0, numberOfLoadedMessages()))
            } else {
                dbListToUIList(getImData().getNMessageWithOffsetForConversation(conversationId, 0, numberOfLoadedMessages()))
            }
            sourceOfNewConversation.onNext(Pair(NEW_LIST_EVENT, listOfMessages))
            loading.set(false)
            callback.onListLoaded(listOfMessages.size)
        }
    }

    @SuppressLint("CheckResult")
    fun loadMoreMessages() {
        scheduleLightWork {
            val addedMessages = if (isChatRoom) {
                roomDbListToUIList(getImData().chatRoomRepo.getNMessageWithOffsetForChat(conversationId, calculateOffset(), GlobalConstants.REMOTE_SYNC_COUNT))
            } else {
                dbListToUIList(getImData().getNMessageWithOffsetForConversation(conversationId, calculateOffset(), GlobalConstants.REMOTE_SYNC_COUNT))
            }
            sourceOfLoadedMoreMessages.onNext(Pair(MESSAGES_ADDED_EVENT, addedMessages))
        }
    }

    /*
     * @return -1 if message is not found, else returns message position
     */
    private fun getMessagePosition(messageList: List<MessageListItemData>, messageId: Long): Int =
            messageList.indices.firstOrNull { messageId == messageList[it].id } ?: -1

    private fun addMessageToTheList(messageList: MutableList<MessageListItemData>, message: MessageListItemData): Pair<MessageListItemData, Int> {
        val result = calculateMessagePosition(messageList, message)
        if (!result.first) { // if message is not there already
            if (result.second == -1) {
                messageList.add(message)
            } else {
                messageList.add(result.second, message)
            }
        }
        return Pair(message, result.second)
    }

    @Synchronized
    private fun addMessagesToList(inputList: List<MessageListItemData>, outputList: MutableList<MessageListItemData>):
            ArrayList<MessageListItemData> {
        val tempDateSeparators = HashSet<Long>()
        val addedMessages = ArrayList<MessageListItemData>()
        inputList.filter { !it.deleted }.forEach { uiMessage ->
            addMessageToTheList(outputList, uiMessage)
            addedMessages.add(uiMessage)
            tempDateSeparators.add(getLocalMidnightTime(uiMessage.time))
        }
        tempDateSeparators
                .filter { getMessagePosition(outputList, it) == -1 }
                .map { MessageListItemData.buildDateDivider(it, mContext) }
                .forEach { addMessageToTheList(outputList, it) }
        return addedMessages
    }

    @Synchronized
    private fun removeMessagesFromList(inputList: List<MessageListItemData>, outputList: MutableList<MessageListItemData>) {
        for (imData in inputList) {
            val position = getMessagePosition(outputList, imData.id)
            if (position > 0) outputList.removeAt(position)
            if (selectedIms.contains(imData.id)) selectedIms.remove(imData.id)
        }
    }

    @Synchronized
    private fun updateMessages(inputList: List<MessageListItemData>, outputList: MutableList<MessageListItemData>) {
        inputList.map { Pair(it, getMessagePosition(outputList, it.id)) }
                .filter { it.second > 0 }.map { Pair(it.first, it.second) }.forEach { (uiMessage, position) ->
                    uiMessage.oldMessage = outputList.removeAt(position)
                    uiMessage.oldMessage?.let {
                        uiMessage.selected = it.selected
                    }
                    outputList.add(position, uiMessage)
                }
    }

    private fun dbListToUIList(oldLIst: List<InstantMessageData>) = oldLIst.map{MessageListItemData(it, mContext)}

    private fun roomDbListToUIList(oldLIst: List<Message>): List<MessageListItemData> {
        return oldLIst.map { MessageListItemData(it, mContext) }
    }


    private val sourceOfChatRoomAddedMessages = getImData().chatRoomRepo.getOnMessagesAddedObservable().subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_ADDED_EVENT, filteredMessages) }

    private val sourceOfChatRoomRemovedMessages = getImData().chatRoomRepo.getOnMessagesRemovedObservable().subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_REMOVED_EVENT, filteredMessages) }

    private val sourceOfChatRoomUpdatedMessages = getImData().chatRoomRepo.getOnMessagesUpdatedObservable().subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    private val sourceOfChatRoomReadMessages = getImData().chatRoomRepo.getOnMessagesReadObservable().subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    private val sourceOfAddedMessages = getImData().observableOnListMessagesAdded.subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_ADDED_EVENT, filteredMessages) }

    private val sourceOfRemovedMessages = getImData().observableOnListMessagesRemoved.subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_REMOVED_EVENT, filteredMessages) }

    private val sourceOfUpdatedMessages = getImData().observableOnListMessageUpdated.subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    //assume: all messages belongs to same conversation
    private val sourceOfReadMessages = getImData().observableOnMessagesRead.subscribeOn(Schedulers.computation())
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    private val sourceOfNewConversation = PublishSubject.create<Pair<Int, List<MessageListItemData>>>()
    private val sourceOfLoadedMoreMessages = PublishSubject.create<Pair<Int, List<MessageListItemData>>>()

    private val sourceOfManuallyChangedList = PublishSubject.create<ArrayList<MessageListItemData>>()

    val source : Observable<ArrayList<MessageListItemData>> = sourceOfNewConversation
            .mergeWith(sourceOfLoadedMoreMessages)
            .mergeWith(sourceOfAddedMessages)
            .mergeWith(sourceOfRemovedMessages)
            .mergeWith(sourceOfUpdatedMessages)
            .mergeWith(sourceOfReadMessages)
            .mergeWith(sourceOfChatRoomAddedMessages)
            .mergeWith(sourceOfChatRoomRemovedMessages)
            .mergeWith(sourceOfChatRoomUpdatedMessages)
            .mergeWith(sourceOfChatRoomReadMessages)
            .map {
                Log.d(TAG, "Event: ${it.first}")
                when (it.first) {
                    NEW_LIST_EVENT -> {
                        list.clear();addMessagesToList(it.second, list)
                    }
                    MESSAGES_ADDED_EVENT -> addMessagesToList(it.second, list)
                    MESSAGES_REMOVED_EVENT -> removeMessagesFromList(it.second, list)
                    MESSAGES_UPDATED_EVENT -> updateMessages(it.second, list)
                    else -> {
                        Log.d(TAG, "Else event: ${it.first}")
                        updateMessages(it.second, list)
                    }
                }
                list

            }.mergeWith(sourceOfManuallyChangedList).replay(1).autoConnect()

    /**
     * Uses binary search to find new message position based on message time
     *  @param messageList - message list in which we are searching
     *  @param message - message which position we are looking
     *  @return pair of <Is message in list already, message position>
     */
    private fun calculateMessagePosition(messageList: List<MessageListItemData>, message: MessageListItemData): Pair<Boolean, Int> {
        if (messageList.isEmpty()) return Pair(false, 0)
        if (message > messageList[messageList.size - 1]) return Pair(false, -1) // to the end
        var lo = 0
        var hi = messageList.size - 1
        while (lo <= hi) {
            val mid = (lo + hi) / 2
            val compare = message.compareTo(messageList[mid])
            when {
                compare > 0 -> lo = mid + 1
                compare < 0 -> hi = mid - 1
                else -> return Pair(messageList[mid].id == message.id, mid)
            }
        }
        return Pair(false, hi + 1)
    }

    fun toggleSelection(id: Long) {
        list.firstOrNull { it.id == id }?.let { message ->
            message.selected = !message.selected
            if (message.selected) {
                selectedIms.add(id)
            } else {
                selectedIms.remove(id)
            }
            sourceOfManuallyChangedList.onNext(list)
        }
    }

    @Suppress("unused")
    fun setSelectedIms(ids: LongArray) {
        ids.forEach { imId ->
            list.firstOrNull { it.id == imId }?.apply {
                this.selected = true
                selectedIms.add(this.id)
            }
        }
        sourceOfManuallyChangedList.onNext(list)
    }

    fun isSelected(id: Long): Boolean = list.firstOrNull { it.id == id }?.selected ?: false

    fun clearSelection() {
        list.forEach { im -> im.selected = false }
        selectedIms.clear()
        sourceOfManuallyChangedList.onNext(list)
    }

    fun getSelectedMessagesCount(): Int = selectedIms.size

    /**
     * Used primary for markAsRead Section
     * If last message is visible then mark whole conversation as read
     * @param firstVisibleItemPosition
     * @param lastVisibleItemPosition
     */
    fun setVisibleMessages(firstVisibleItemPosition: Int, lastVisibleItemPosition: Int) {
        val listSize = list.size
        if (listSize > 0 && conversationId != -1L) {
            if (lastVisibleItemPosition >= listSize - 1) {
                if (isChatRoom) {
                    getImData().chatRoomRepo.markReadChatRoom(conversationId)
                } else {
                    getImData().markMessagesRead(conversationId)
                }
            } else {
                val visibleImList = ArrayList<InstantMessageData>()
                if (firstVisibleItemPosition < 0 || lastVisibleItemPosition < 0
                        || list.size <= firstVisibleItemPosition || list.size <= lastVisibleItemPosition) {
                    return
                }
                for (i in firstVisibleItemPosition..lastVisibleItemPosition) {
                    val item = list[i]
                    if (!item.isRead || item.isFileTransferUnread()) {
                        getImData().getMessageById(list[i].id)?.let {
                            it.markRead()
                            visibleImList.add(it)
                        }
                    }
                }
                if (!visibleImList.isEmpty()) {
                    if (isChatRoom) {
                        //getImData().chatRoomRepo(visibleImList)
                        //nvtd
                    } else {
                        getImData().markMessagesRead(visibleImList)
                    }
                }
            }
        }
    }

    private fun calculateOffset(): Int = list.filter { it.msgType != MESSAGE_TYPE_DATE_DIVIDER }.size

    @SuppressLint("CheckResult")
    fun accept(messageListItemData: MessageListItemData) {
        scheduleLightWork {
            val message = getImData().getMessageById(messageListItemData.id)
            chatApi.acceptMessage(message)
        }
    }

    @SuppressLint("CheckResult")
    fun reject(messageListItemData: MessageListItemData) {
        scheduleLightWork {
            val message = getImData().getMessageById(messageListItemData.id)
            chatApi.rejectMessage(message)
        }
    }

    fun retry(messageListItemData: MessageListItemData) {
        scheduleLightWork {
            getImData().getMessageById(messageListItemData.id)?.let {
                chatApi.retryTransfer(it)
            }
        }
    }

    fun isEmpty(): Boolean = list.isEmpty()

    @Suppress("unused")
    @SuppressLint("CheckResult")
    fun incomingFileIsNotImage(messageListItemData: MessageListItemData) {
        if (messageListItemData.status != ImStatusData.INCOMING_FT_RECEIVED_CHECKED_READ) {
            scheduleLightWork {
                val message = getImData().getMessageById(messageListItemData.id)
                chatApi.incomingFileIsNotImage(message)
            }
        }
    }

    @Suppress("unused")
    @SuppressLint("CheckResult")
    fun incomingFileIsImage(messageListItemData: MessageListItemData) {
        if (messageListItemData.status != ImStatusData.INCOMING_IMAGE_TRANSFER_RECEIVED_READ) {
            scheduleLightWork {
                val message = getImData().getMessageById(messageListItemData.id)
                chatApi.incomingFileIsImage(message)
            }
        }
    }

    fun hasTransferMessagesInSelection() = getSelectedMessages().firstOrNull { it.isTransfer() } != null
    fun deleteSelectedMessages() {
        if (isChatRoom) {
            ThreadExecutors.scheduleLightWork {
                getImData().chatRoomRepo.deleteMessages(getSelectedChatRoomMessages())
            }
        } else {
            getImData().removeBulkMessages(getSelectedMessages())
        }
    }
    private fun getSelectedMessages(): List<InstantMessageData> {
        val selected = ArrayList<InstantMessageData>()
        selectedIms.mapNotNullTo(selected) { getImData().getMessageById(it) }
        return selected
    }

    fun getSelectedFileTransferMessages() = getSelectedMessages().filter { it.isTransfer() }
    private fun getSelectedNonFileTransferMessages() = getSelectedMessages().filter { !it.isTransfer() }
    private fun getSelectedChatRoomMessages() : List<Message>{
        val selected = ArrayList<Message>()
        selectedIms.mapNotNullTo(selected) { getImData().chatRoomRepo.getMessageById(it) }
        return selected
    }

    fun getSelectedMessagesAsString(): String {
        return if (isChatRoom) {
            getSelectedChatRoomMessages().joinToString(separator = "\n") { it -> it.text }
        } else {
            getSelectedNonFileTransferMessages().joinToString(separator = "\n") { it -> it.message }
        }
    }

    var hardwiredSession: Boolean = false


}
