package com.bria.common.controller.im.refactoring.filetransfer

import android.content.Context
import com.bria.common.analytics.BIAnalytics
import com.bria.common.controller.Controllers
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.ChatApi
import com.bria.common.controller.im.refactoring.IImData
import com.bria.common.controller.im.refactoring.ImDataObserverAdapter
import com.bria.common.controller.im.refactoring.db.entities.*
import com.bria.common.controller.im.refactoring.filetransfer.FileTransferInfo.Companion.PENDING
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.mdm.Factories
import com.bria.common.util.Log
import com.bria.common.util.RecordingUtils
import com.counterpath.sdk.XmppAccount
import com.counterpath.sdk.XmppApiFileTransfer
import com.counterpath.sdk.XmppFileTransfer
import com.counterpath.sdk.XmppFileTransferItem
import com.counterpath.sdk.handler.XmppFileTransferHandler
import com.counterpath.sdk.pb.Xmppfiletransfer
import io.reactivex.disposables.Disposable
import io.reactivex.subjects.PublishSubject
import java.io.File
import java.util.*
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.TimeUnit
import kotlin.collections.HashMap


@Suppress("UNUSED_PARAMETER") //nvtd work in progress,
class FileTransferApi(private val chatApi: ChatApi, private val context: Context) : XmppFileTransferHandler {

    private val TAG = "FileTransferApi"
    private val FILE_READ_TAG = 0x01

    private val outgoingTransfers: LinkedList<FileTransferInfo> = LinkedList()
    private val incomingTransfers: LinkedList<FileTransferInfo> = LinkedList()

    private val progressEvent = PublishSubject.create<ConcurrentHashMap<Int, Int>>()
    private val activeProgresses = ConcurrentHashMap<Int, Int>()

    private val imObserverAdapter = object : ImDataObserverAdapter() {
        override fun onMessageAdded(messageId: Long?, im: InstantMessageData?) {
            arrayOf(outgoingTransfers, incomingTransfers).forEach { list ->
                list.forEach { transfer ->
                    transfer.messages.filterValues { im?.externalId == it.externalId }.forEach { it.value.id = im?.id }
                }
            }
        }

        override fun onOwnerIsChanged() {
            super.onOwnerIsChanged()
            fixUnfinishedFileTransfers()
        }

        override fun onConversationAdded(conversation: ImConversationData) {
            arrayOf(outgoingTransfers, incomingTransfers).forEach { list ->
                list.filter { it.imConversationData.id == -1L && it.imConversationData.remoteKey == conversation.remoteKey }
                        .forEach { it.imConversationData.id = conversation.id }
            }
        }

        override fun onConversationRemoved(conversation: ImConversationData?) {
            if (conversation == null) {
                return
            }
            outgoingTransfers.filter { it.imConversationData.id == conversation.id }.forEach { it.fileTransfer.end() }
            incomingTransfers.filter { it.imConversationData.id == conversation.id }.forEach { it.fileTransfer.reject() }
        }
    }


    init {
        getImData().attachObserver(imObserverAdapter)
        fixUnfinishedFileTransfers()
    }


    fun sendImagesAsSingleTransfer(selectedFiles: Array<String>, imConversationData: ImConversationData): Boolean =
            sendFilesAsSingleTransfer(selectedFiles, imConversationData, true)

    fun sendFilesAsSingleTransfer(selectedFiles: Array<String>, imConversationData: ImConversationData, areFilesImages: Boolean = false): Boolean {
        val fileTransfer = prepareFileTransfer(imConversationData.accountId, imConversationData.remoteKey) ?: return false
        var item: XmppFileTransferItem? = null
        val messages = HashMap<Int, InstantMessageData>()
        selectedFiles.forEach {
            val file = File(it)
            if (file.exists()) {
                item = fileTransfer.newXmppFileTransferItem(file)
            }
            val time = Date().time
            val externalID = item?.let { getFileTransferExternalId(fileTransfer.handle(), it.handle()) } ?: UUID.randomUUID().toString()
            val status = (if (areFilesImages) ImStatusData.OUTGOING_IMAGE_TRANSFER_PREFIX else ImStatusData.OUTGOING_FT_PENDING) or
                    (if (file.exists()) ImStatusData.OUTGOING_FLAG_PENDING else ImStatusData.OUTGOING_FT_NOT_EXISTS)
            val messageText = "File: ${file.name} Size: ${FileTransferUtils.readableFileSize(file.length())}"
            val message = InstantMessageData.InstantMessageDataBuilder()
                    .setMessage(messageText).setConversationId(imConversationData.id).setFilePath(it).setTime(time).setModTime(time)
                    .setRemoteAddress(imConversationData.remoteKey)
                    .setStatus(status)
                    .setFileSize(file.length())
                    .setExternalId(externalID)
                    .createInstantMessageData()
            getImData().addMessage(message)
            item?.let { itemInfo -> messages.put(itemInfo.handle(), message) }
        }

        outgoingTransfers.add(FileTransferInfo(imConversationData, fileTransfer, messages, PENDING, 0, null))
        fileTransfer.start()
        return true
    }

    private fun prepareFileTransfer(accountId: String, remoteKey: String): XmppFileTransfer? {
        val account = Controllers.get().accounts.getAccountForUserAtDomain(accountId, EAccountType.Xmpp)
        if (account == null || !account.isRegistered) {
            return null
        }
        val xmppAccount = chatApi.stackManager?.registrationManager?.getXmppAccount(account) ?: return null
        val fileTransferApi = XmppApiFileTransfer.get(xmppAccount)
        val fileTransfer: XmppFileTransfer? = fileTransferApi.newXmppFileTransfer()
        BuddyKeyUtils.getImAdressFromNewBuddyKey(remoteKey).let {
            if (it.isNotEmpty()) {
                fileTransfer?.addParticipant(it)
            }
        }

        return fileTransfer
    }


    override fun onNewFileTransferEvent(transfer: XmppFileTransfer, msg: Xmppfiletransfer.XmppFileTransferEvents.NewFileTransferEvent) {
        Log.d(TAG, "onNewFileTransferEvent: ")
        if (msg.fileTransferType == Xmppfiletransfer.XmppFileTransferType_Outgoing) {
            BIAnalytics.get().reportFireTransferSent()
        } else if (msg.fileTransferType == Xmppfiletransfer.XmppFileTransferType_Incoming) {
            BIAnalytics.get().reportFireTransferReceived()
        }
        if (msg.fileTransferType == Xmppfiletransfer.XmppFileTransferType_Incoming) {
            val xmppAccount = transfer.api.account
            val accUserAtDomain = "${xmppAccount.settings.username}@${xmppAccount.settings.domain}"
            val buddyKey = BuddyKeyUtils.getNewBuddyKey(EAccountType.Xmpp, accUserAtDomain, ImpsUtils.getRemoteAddress(msg.remoteAddress))
            var conversation = getImData().getConversationByRemoteKeyAndType(buddyKey, ImConversationData.IM_XMPP)
            var hasConversation = true
            if (conversation == null) {
                hasConversation = false
                //create conversation in db
                conversation = ImConversationData.ImConversationDataBuilder()
                        .setAccountId(accUserAtDomain)
                        .setDisplayName(msg.remoteDisplayName)
                        .setModTime(Date().time)
                        .setRemoteKey(buddyKey)
                        .setType(ImConversationData.IM_XMPP)
                        .setParticipants(buddyKey).createImConversationData()
                getImData().addConversation(conversation)
            }
            val messages = HashMap<Int, InstantMessageData>()
            incomingTransfers.add(FileTransferInfo(conversation, transfer, messages, PENDING, 0, msg.remoteAddress))
            msg.fileItemsList.forEach { item ->
                val file = File(getFilesFolder(), fixUppercaseExtension(item.remotefileName))
                val messageText =
                        context.resources.getString(com.bria.common.R.string.tFileNameSize, file.name, FileTransferUtils.readableFileSize(item.fileSizeBytes))
                //val messageText = "File: ${file.name} Size: ${FileTransferUtils.readableFileSize(item.fileSizeBytes)}"
                val message = InstantMessageData.InstantMessageDataBuilder()
                        .setMessage(messageText).setConversationId(if (hasConversation) conversation.id else ImConversationData.INVALID_CONVERSATION_ID)
                        .setFilePath(file.path).setTime(Date().time).setModTime(Date().time)
                        .setRemoteAddress(conversation.remoteKey)
                        .setStatus(ImStatusData.INCOMING_FT_WAITING_FOR_USER_PROMPT_UNREAD)
                        .setFileSize(item.fileSizeBytes)
                        .setExternalId(getFileTransferExternalId(transfer.handle(), item.fileTransferItemHandle))
                        .createInstantMessageData()
                messages.put(item.fileTransferItemHandle, message)
            }
            if (hasConversation) {
                getImData().addMessages(messages.values.toList())
            } else {

                getImData().addMessages(messages.values.toList(), buddyKey)
            }
        }
    }

    @Synchronized
    override fun onFileTransferEndedEvent(xmppFileTransfer: XmppFileTransfer, msg: Xmppfiletransfer.XmppFileTransferEvents.FileTransferEndedEvent) {
        val found = mutableListOf<FileTransferInfo>()
        arrayOf(outgoingTransfers, incomingTransfers).forEach { list ->
            list.filter { transfer -> transfer.fileTransfer.handle() == xmppFileTransfer.handle() }.forEach { transfer -> found.add(transfer) }
            found.forEach { list.remove(it) }
            found.clear()
        }
    }

    @Synchronized
    override fun onFileTransferItemEndedEvent(xmppFileTransfer: XmppFileTransfer, msg: Xmppfiletransfer.XmppFileTransferEvents.FileTransferItemEndedEvent) {
        Log.d(TAG, "onFileTransferItemEndedEvent: " + msg.fileTransferItemHandle + " Reason: " + msg.endReason)
        outgoingTransfers.firstOrNull { it.fileTransfer.handle() == xmppFileTransfer.handle() }?.apply {
            this.messages[msg.fileTransferItemHandle]?.let { message ->
                message.id?.let { getImData().getMessageById(message.id) }
            }?.let { m ->
                when (msg.endReason) {
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Complete -> {
                        Log.d(TAG, "Completed")
                        m.moveToSentStatus()
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_BadConnection -> {
                        m.moveToFailToSend()
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_BadFile -> {
                        Log.d(TAG, "Bad File")
                        m.moveToNotExists()
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Interrupted -> {
                        Log.d(TAG, "Interrupted")
                        m.moveToFailToSend()
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Failed -> Log.d(TAG, "Failed")
                    Xmppfiletransfer.XmppFileTransferItemEndReason_LocalCancel -> {
                        Log.d(TAG, "Local Cancel")
                        m.moveToFailToSend()
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_RemoteCancel -> {
                        Log.d(TAG, "Remote Cancel")
                        m.moveToRemoteCancel()
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Unknown -> {
                        Log.d(TAG, "Unknown")
                        m.moveToUnknownError()
                    }
                }
                this.messages[msg.fileTransferItemHandle]?.status = m.status
                getImData().updateMessage(m)
            }
        }
        incomingTransfers.firstOrNull { it.fileTransfer.handle() == xmppFileTransfer.handle() }?.apply {
            this.messages[msg.fileTransferItemHandle]?.let { message -> getImData().getMessageById(message.id) }?.let { m ->
                when (msg.endReason) {
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Complete -> {
                        Log.d(TAG, "Completed")
                        m.fileTransferState = 100
                        m.status = m.getFileReadFlag() or ImStatusData.INCOMING_FT_RECEIVED_UNREAD
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_BadConnection -> {
                        Log.d(TAG, "Bad Connection")
                        m.fileTransferState = 0
                        m.status = m.getFileReadFlag() or ImStatusData.INCOMING_FT_FAILED_TO_RECEIVE_UNREAD
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_BadFile -> {
                        Log.d(TAG, "Bad File for incomming file ??")
                        m.fileTransferState = 0
                        m.status = m.getFileReadFlag() or ImStatusData.INCOMING_FT_FAILED_TO_RECEIVE_UNREAD
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Interrupted -> {
                        Log.d(TAG, "Interrupted")
                        m.fileTransferState = 0
                        m.status = m.getFileReadFlag() or ImStatusData.INCOMING_FT_FAILED_TO_RECEIVE_UNREAD
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Failed -> {
                        Log.d(TAG, "Failed")
                        m.fileTransferState = 0
                        m.status = m.getFileReadFlag() or ImStatusData.INCOMING_FT_FAILED_TO_RECEIVE_UNREAD
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_LocalCancel -> {
                        Log.d(TAG, "Local Cancel")
                        m.fileTransferState = 0
                        m.status = m.getFileReadFlag() + ImStatusData.INCOMING_FT_LOCAL_CANCEL_UNREAD
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_RemoteCancel -> {
                        Log.d(TAG, "Remote Cancel")
                        m.fileTransferState = 0
                        m.status = m.getFileReadFlag() + ImStatusData.INCOMING_FT_REMOTE_CANCEL_UNREAD
                    }
                    Xmppfiletransfer.XmppFileTransferItemEndReason_Unknown -> {
                        Log.d(TAG, "Unknown")
                        m.fileTransferState = 0
                        m.status = m.getFileReadFlag() + ImStatusData.INCOMING_FT_UNKNOWN_ERROR_UNREAD
                    }
                }
                this.messages[msg.fileTransferItemHandle]?.status = m.status
                getImData().updateMessage(m)
            }
        }
    }

    val progressDisposable: Disposable = progressEvent.filter { it -> it.isNotEmpty() }.sample(300, TimeUnit.MILLISECONDS).subscribe(
            { mapOfMessages ->

                val changedMessages = mutableListOf<InstantMessageData>()
                mapOfMessages.forEach { (handle, percent) ->
                    arrayOf(incomingTransfers, outgoingTransfers).forEach { list ->
                        list.firstOrNull { info -> info.messages.containsKey(handle) }?.let { transfer ->
                            transfer.messages[handle]?.let { message ->
                                Log.d(TAG, "message ${message.id} progress: ${message.fileTransferState}")
                                if (message.isInProgress() || message.isWaitingForUserPrompt() || message.isOutgoingPending()) {
                                    message.moveToProgress()
                                    message.fileTransferState = percent
                                    changedMessages.add(message)
                                }
                            }
                        }
                    }
                }
                getImData().updateMessages(changedMessages)
            },
            { error -> Log.d(TAG, "OnProgress Problem ${error?.message}") })


    override fun onFileTransferItemProgressEvent(xmppFileTransfer: XmppFileTransfer,
                                                 msg: Xmppfiletransfer.XmppFileTransferEvents.FileTransferItemProgressEvent) {
        Log.d(TAG, "onFileTransferItemProgressEvent: " + msg.fileTransferItemHandle + " percent: " + msg.percent)
        //cancelIfOutgoingPending(msg.fileTransferItemHandle)
        if (msg.percent < 100) {
            activeProgresses[msg.fileTransferItemHandle] = msg.percent
        } else {
            activeProgresses.remove(msg.fileTransferItemHandle)
        }
        progressEvent.onNext(activeProgresses)

    }

    private fun cancelIfOutgoingPending(fileTransferItemHandle: Int) {
        outgoingTransfers.forEach { transfer ->
            transfer.messages[fileTransferItemHandle]?.let { message ->
                if (message.isLocalCancel()) {
                    transfer.fileTransfer.getXmppFileTransferItem(fileTransferItemHandle)?.cancelItem(null, 0)
                    return
                }
            }
        }
    }

    override fun onErrorEvent(xmppFileTransfer: XmppFileTransfer, msg: Xmppfiletransfer.XmppFileTransferEvents.ErrorEvent) {
        Log.d(TAG, "onErrorEvent: " + msg.errorText + " ")
//        arrayOf(incomingTransfers, outgoingTransfers).forEach { list ->
//            list.firstOrNull { it.fileTransfer.handle() == xmppFileTransfer.handle() }?.let { transfer ->
//                transfer.messages.values.forEach { message ->
//                    msg.
//                }
//            }
//        }
    }

    fun retryTransfer(message: InstantMessageData) {
        val conversation = getImData().getConversationById(message.conversationId)
        val messages = HashMap<Int, InstantMessageData>()
        val fileTransfer = prepareFileTransfer(accountId = conversation.accountId, remoteKey = conversation.remoteKey) ?: return
        var item: XmppFileTransferItem? = null
        val file = File(message.filePath)
        if (file.exists()) {
            item = fileTransfer.newXmppFileTransferItem(file)
        }
        //message.modTime = Date().time
        val status = (if (message.isOutgoingImage()) ImStatusData.OUTGOING_IMAGE_TRANSFER_PREFIX else ImStatusData.OUTGOING_FT_PENDING) or
                (if (file.exists()) ImStatusData.OUTGOING_FLAG_PENDING else ImStatusData.OUTGOING_FT_NOT_EXISTS)
        message.status = status
        getImData().updateMessage(message)
        item?.let { itemInfo -> messages.put(itemInfo.handle(), message) }
        outgoingTransfers.add(FileTransferInfo(conversation, fileTransfer, messages, PENDING, 0, null))
        fileTransfer.start()
    }

    @Synchronized fun acceptMessage(message: InstantMessageData) {
        if (!prepareFileTransferStorage()) {
            Log.e(TAG, "acceptMessage: " + "storage permisson not accepted")
            return
        }
        Log.d(TAG, "acceptMessage")
        incomingTransfers.forEach { transfer ->
            transfer.messages.filterValues { it.externalId == message.externalId }.forEach { pair ->
                val item = transfer.fileTransfer.getXmppFileTransferItem(pair.key)
                val uniqueFile = getUniqueFile(message.filePath)
                item.acceptItem(uniqueFile)
                pair.value.apply {
                    if (message.isWaitingForUserPrompt()) {
                        message.status = message.getFileReadFlag() + ImStatusData.INCOMING_FT_RECEIVING_UNREAD
                        Log.d(TAG, "update message; ")
                        message.filePath = uniqueFile.path
                        this.status = message.status
                        this.filePath = message.filePath
                        showInfoOnMessages(transfer.messages.values, message)
                    }
                }
            }
            if (shouldAcceptFileTransfer(transfer)) transfer.fileTransfer.accept()
        }
    }

    @Synchronized fun rejectMessage(message: InstantMessageData) {
        Log.d(TAG, "rejectMessage")
        arrayOf(incomingTransfers, outgoingTransfers).forEach { list ->
            list.forEach { transfer ->
                transfer.messages.filterValues { it.externalId == message.externalId }.forEach { pair ->
                    val item = transfer.fileTransfer.getXmppFileTransferItem(pair.key)
                    pair.value.apply {
                        if (transferNotStarted(transfer)) {
                            item.rejectItem()
                            message.status = message.getFileReadFlag() + ImStatusData.INCOMING_FT_REJECTED_UNREAD
                            this.status = message.status
                            Log.d(TAG, "update message; ")
                            showInfoOnMessages(transfer.messages.values, message)
                            if (shouldRejectFileTransfer(transfer)) transfer.fileTransfer.reject()
                            else if (shouldAcceptFileTransfer(transfer)) transfer.fileTransfer.accept()
                        } else if (this.isInProgress()) {
                            item.cancelItem(null, 0)
                        } else if (this.isOutgoingPending()) {
                            message.moveToLocalCancel()
                            this.status = message.status
                            if (transfer.messages.filterValues { it.isOutgoingPending() }.isEmpty()) {
                                transfer.fileTransfer.end()
                            }
                        }
                    }
                }

            }
        }
    }

    private fun showInfoOnMessages(values: MutableCollection<InstantMessageData>, except: InstantMessageData?) {
        except?.let {
            val listToUpdate = mutableListOf<InstantMessageData>()
            listToUpdate.add(except)
            values.filter { it.id != except.id }.forEach {
                getImData().getMessageById(it.id)?.takeIf { it.getStatusWithoutPrefix() == ImStatusData.INCOMING_FLAG_WAITING_FOR_USER_PROMPT }?.let { message ->
                    message.moveToShowInfo()
                    listToUpdate.add(message)
                }
            }
            if (listToUpdate.isNotEmpty()) {
                getImData().updateMessages(listToUpdate)
            }
        }
    }


    fun fixUnfinishedFileTransfers() {
        getImData().unfinishedFileTransfers.forEach { unfinishedMessage ->
            unfinishedMessage?.let {
                if (findMessageInTransfers(it) == null) {
                    if (it.isIncomingFT()) {
                        it.moveToFailToReceive()
                    } else {
                        it.moveToFailToSend()
                    }
                    getImData().updateMessage(it)
                }
            }
        }
    }

    private fun findMessageInTransfers(im: InstantMessageData): InstantMessageData? {
        arrayOf(incomingTransfers, outgoingTransfers).forEach { list ->
            list.forEach { info ->
                info.messages.values.firstOrNull { im.externalId == it.externalId }?.let {
                    return it
                }
            }
        }
        return null
    }

    private fun shouldAcceptFileTransfer(transfer: FileTransferInfo): Boolean {
        var hasAccepted = false
        transfer.messages.forEach { pair ->
            if (pair.value.isWaitingForUserPrompt()) {
                return false
            } else if (pair.value.status == ImStatusData.INCOMING_FT_RECEIVING_UNREAD
                    || pair.value.status == ImStatusData.INCOMING_FT_RECEIVING_READ) {
                hasAccepted = true
            }
        }
        return hasAccepted
    }

    private fun transferNotStarted(transfer: FileTransferInfo): Boolean {
        transfer.messages.forEach { (_, message) ->
            if (message.isWaitingForUserPrompt()) return true
        }
        return false
    }

    private fun shouldRejectFileTransfer(transfer: FileTransferInfo): Boolean {
        transfer.messages.forEach { pair ->
            if (pair.value.status != ImStatusData.INCOMING_FT_REJECTED_UNREAD
                    && pair.value.status != ImStatusData.INCOMING_FT_REJECTED_READ) {
                return false
            }
        }
        return true
    }

    private fun getUniqueFile(filename: String): File {
        var newFile = File(filename)
        var index = 1
        while (newFile.exists()) {
            newFile = File(newFile.parent, """${newFile.nameWithoutExtension}($index).${newFile.extension}""")
            index += 1
        }
        return newFile
    }

    private fun getFileTransferExternalId(transferHandle: Int?, transferItemHandle: Int?): String {
        if (transferHandle != null && transferItemHandle != null) {
            return """ID$transferHandle||$transferItemHandle||${UUID.randomUUID()}"""
        } else {
            return UUID.randomUUID().toString()
        }
    }

    // storage related stuff

    private fun getImagesFolder(): String = RecordingUtils.getRecordingFolder(context) + File.separatorChar + "Images"
    private fun getFilesFolder(): String = RecordingUtils.getRecordingFolder(context) + File.separatorChar + "Files"

    private fun prepareFileTransferStorage(): Boolean {
        var success = false
        if (RecordingUtils.isStorageAvailable()) {
            success = createImageAndFileTransferFolders()
        } else {
            Log.e(TAG, "External storage is not available, File Transfer should be disabled")
        }
        return success
    }

    private fun createImageAndFileTransferFolders(): Boolean {
        var success: Boolean
        success = RecordingUtils.createBriaFolder(context)
        if (!success) return false
        val imagesFolder = Factories.getIOFactory().newFile(getImagesFolder())

        try {
            success = imagesFolder.isDirectory
            if (!success) {
                Log.d(TAG, "Creating folder: " + imagesFolder.path)
                success = imagesFolder.mkdir()
                if (!success) return false
            }

        } catch (e: Exception) {
            Log.e(TAG, "Unable to create folder: ${imagesFolder.path}")
            return false
        }

        val filesFolder = Factories.getIOFactory().newFile(getFilesFolder())
        try {
            success = filesFolder.isDirectory
            if (!success) {
                Log.d(TAG, "Creating folder: ${filesFolder.path}")
                success = filesFolder.mkdir()
            }

        } catch (e: Exception) {
            Log.e(TAG, "Unable to create folder: ${filesFolder.path}")
            return false
        }

        return success
    }


    fun destroy() {
        getImData().detachObserver(imObserverAdapter)
        incomingTransfers.clear()
        outgoingTransfers.clear()
        progressDisposable.dispose()
    }

    private fun getImData(): IImData = Controllers.get().im.imData

    fun InstantMessageData.getFileReadFlag(): Int = this.status and FILE_READ_TAG

    fun incomingFileIsNotImage(message: InstantMessageData?) {
        message?.let {
            val cachedMessage = getMessageByIdFromIncomingTransfers(it.id)
            if (cachedMessage == null) {
                it.status = ImStatusData.INCOMING_FT_RECEIVED_CHECKED_READ
                getImData().updateMessage(it)
            }
        }

    }

    fun incomingFileIsImage(message: InstantMessageData?) {
        message?.let {
            val cachedMessage = getMessageByIdFromIncomingTransfers(it.id)
            if (cachedMessage == null) {
                it.status = ImStatusData.INCOMING_IMAGE_TRANSFER_RECEIVED_READ
                getImData().updateMessage(it)
            }
        }
    }

    fun clearCacheForAccount(account: XmppAccount) {
        val updateMessages = mutableListOf<InstantMessageData>()
        val userAtDomain = account.settings.username + "@" + account.settings.domain
        incomingTransfers.filter { it -> it.imConversationData.accountId == userAtDomain }.forEach { transfer ->
            transfer.messages.forEach { (handle, message) ->
                getImData().getMessageById(message.id)?.let { m ->
                    m.moveToFailToReceive()
                    updateMessages.add(m)
                }
                transfer.fileTransfer.getXmppFileTransferItem(handle).cancelItem(null, 0)
            }
        }
        outgoingTransfers.filter { it -> it.imConversationData.accountId == userAtDomain }.forEach { transfer ->
            transfer.messages.forEach { (handle, message) ->
                getImData().getMessageById(message.id)?.let { m ->
                    m.moveToFailToSend()
                    updateMessages.add(m)
                }
                transfer.fileTransfer.getXmppFileTransferItem(handle).cancelItem(null, 0)
            }
        }

        getImData().updateMessages(updateMessages)
    }

    //BRACE-11295
    private fun fixUppercaseExtension(filename : String) : String {
        val extension = filename.substringAfterLast('.', "")
        if (extension.isNotEmpty()) {
            return filename.substringBeforeLast(".") + "." + extension.toLowerCase()
        }
        return filename


    }

    private fun getMessageByIdFromIncomingTransfers(id: Long): InstantMessageData? {
        incomingTransfers.forEach { transfer -> transfer.messages.filterValues { it.id == id }.forEach { return it.value } }
        return null
    }
}
