package com.bria.common.controller.im.refactoring.push

import android.annotation.SuppressLint
import android.os.Bundle
import android.text.TextUtils
import com.bria.common.controller.Controllers
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.ChatApi
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.im.refactoring.db.entities.ImStatusData
import com.bria.common.controller.im.refactoring.db.entities.InstantMessageData
import com.bria.common.controller.im.util.ImHelper
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.util.Log
import com.bria.common.util.MimeTypes
import com.bria.common.util.Utils
import io.reactivex.Observable
import io.reactivex.Observer
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.schedulers.Schedulers
import java.util.*
import java.util.regex.Pattern

class PushChatApi(pushDispatcherObservable : Observable<Bundle>) {
    val disposables = CompositeDisposable()
    private val TAG = "PushChatApi"

    init {
        Log.d(TAG, "PushChatApi created")
        pushDispatcherObservable.observeOn(Schedulers.io()).subscribe(object : Observer<Bundle> {
            override fun onSubscribe(d: Disposable) {
                Log.d(TAG, "Subscribed")
             disposables.add(d)
            }

            override fun onComplete() {

            }

            override fun onError(e: Throwable) {

            }

            override fun onNext(bundle: Bundle) {
                if (bundle.containsKey("event") &&
                        (bundle.getString("event") == "tunnelMessage" || bundle.getString("event") == "xmppIm")) {
                    parseMessage(bundle);
                }

            }
        })
    }

    @SuppressLint("MissingPermission")
    private fun getAccountUUID(account: Account): String {
        val accountUUID = account.getStr(EAccountSetting.PushAccountUUID)
        if (TextUtils.isEmpty(accountUUID)) {
            val serialNo = Utils.System.getSystemSerialNumber()
            Log.d(TAG, "Acc " + account.getStr(EAccountSetting.Nickname) + " creating UUID, sn: " + serialNo)
            return UUID.nameUUIDFromBytes((serialNo + account.getStr(EAccountSetting.UserName) + "@" +
                    account.getStr(EAccountSetting.Domain)).toByteArray()).toString()
        }
        return accountUUID
    }

    fun parseMessage(bundle: Bundle) {
        Log.d(TAG, bundle.keySet().joinToString(separator = "\n") { "key $it : value ${bundle.get(it)}"})

        val contentType = bundle.getString("contentType")
        if (MimeTypes.TEXT_PLAIN != contentType && MimeTypes.TEXT_HTML != contentType) {
            Log.e(TAG, "Invalid content-type: $contentType")
            return
        }

        val pushAccountUUID = bundle.getString("accountUUID")
        val from = bundle.getString("caller")
        val displayName = getDisplayNameFromMessage(from)
        val tempFromAddress = ImpsUtils.unSipAddress(getUriFromMessage(from))
        val remoteAddress = if (tempFromAddress.isEmpty()) from else tempFromAddress
        Log.d(TAG, "Remote Address: $remoteAddress")

        val isXmppPushMessage = bundle.getString("event") == "xmppIm"
        val accounts = if (isXmppPushMessage)
            Controllers.get().accounts.getAccounts(EAccountType.Xmpp)
        else
            Controllers.get().accounts.getAccounts(EAccountType.Sip)

        accounts.firstOrNull { acc -> ((acc.getBool(EAccountSetting.IsIMPresence) || acc.getBool(EAccountSetting.IsSMS))
                        && pushAccountUUID.equals(getAccountUUID(acc)) ||
                (acc.type == EAccountType.Xmpp && acc.getBool(EAccountSetting.UsePushNotifications)))}?.let{ account ->

            val impsEnabled = (account.type == EAccountType.Xmpp && account.getBool(EAccountSetting.UsePushNotifications))
                    || account.getBool(EAccountSetting.IsIMPresence) || account.getBool(EAccountSetting.IsSMS)
            if (!impsEnabled) {
                Log.e(TAG, "IMPS and SMS not enabled for " + account.getStr(EAccountSetting.Nickname))
            } else {
                if (!isXmppPushMessage) {
                    val servingServerUrl = bundle.getString("servingClusterURL")
                    val savedServingClusterUrl = account.getStr(EAccountSetting.ServingClusterUrl)
                    if (!TextUtils.isEmpty(servingServerUrl) && !TextUtils.isEmpty(savedServingClusterUrl)) {
                        if (servingServerUrl != savedServingClusterUrl) {
                            Log.d(TAG, "Push im from non-serving server url, current serving serverurl = " + savedServingClusterUrl + ", msg from " +
                                    servingServerUrl)
                            return
                        }
                    }
                } else {
                    Log.d(TAG, "XMPP Push Msg")
                }

                val accountKey = account.getStr(EAccountSetting.UserName) + "@" + account.getStr(EAccountSetting.Domain)
                val buddyKey = BuddyKeyUtils.getNewBuddyKey(account.type, accountKey, remoteAddress)
                Log.d(TAG, "buddykey = " + buddyKey)

                val messageType = bundle.getString("messageType", "IM")
                val isSMS = (!account.getBool(EAccountSetting.IsIMPresence) || messageType == "SMS" || messageType == "MMS") && !isXmppPushMessage
                val conversationType = if (isSMS) ImConversationData.SMS else {
                    if (isXmppPushMessage) ImConversationData.IM_XMPP else ImConversationData.IM_SIP }

                var imConversationData: ImConversationData? = Controllers.get().im.imData.getConversationByRemoteKeyAndType(buddyKey, conversationType)

                if (imConversationData == null) {
                    imConversationData = ImConversationData.ImConversationDataBuilder().setAccountId(accountKey)
                            .setDisplayName(displayName)
                            .setModTime(Date().time)
                            .setType(conversationType)
                            .setParticipants(buddyKey)
                            .setRemoteKey(buddyKey)
                            .createImConversationData()
                    Log.d(TAG, "Adding conversation")
                    Controllers.get().im.imData.addConversation(imConversationData!!)
                }

                val extId = bundle.getString("sessionId")
                imConversationData = Controllers.get().im.imData.getConversationByRemoteKeyAndType(buddyKey, conversationType)

                val content = ImHelper.processIncomingMessage(bundle.getString("contents"), contentType)
                Log.d(TAG, "Got im from $remoteAddress msg : $content")

                val date = Date()
                val instantMessageData = InstantMessageData.InstantMessageDataBuilder()
                        .setConversationId(if (imConversationData != null) imConversationData.id else ChatApi.INVALID_CONVERSATION_ID)
                        .setRemoteAddress(remoteAddress)
                        .setStatus(ImStatusData.INCOMING_UNREAD)
                        .setMessage(content)
                        .setExternalId(extId)
                        .setTime(date.time)
                        .setModTime(date.time)
                        .createInstantMessageData()

                if (ImHelper.isDuplicate(instantMessageData, Controllers.get().im.imData.lastAddedMessage)) {
                    Log.d(TAG, "Skipping duplicate message ")
                    return
                }

                if (imConversationData != null) {
                    Controllers.get().im.imData.addMessage(instantMessageData)
                } else {
                    Controllers.get().im.imData.addMessage(instantMessageData, buddyKey)
                }
            }
        } ?: run {
            Log.d(TAG, "No Accounts!")
        }
    }

    private fun getDisplayNameFromMessage(from : String) : String {
        val pattern = Pattern.compile("\"([^']*)\"")
        val matcher = pattern.matcher(from)
        if (matcher.find()) {
            return matcher.group(1)
        }
        return ""
    }

    /*
     * example input: Huawei Vtr-l09 1031"<sip:1031@fs.cpcdev.com>;tag=2a79adf9
     */
    fun getUriFromMessage(from : String) : String {
        val pattern = Pattern.compile("<([^']*)>")
        val matcher = pattern.matcher(from)
        var uri = ""
        while (matcher.find()) {
            uri =  matcher.group(1)
        }
        return uri
    }

    fun destroy() {
        disposables.dispose()
    }
}
