package com.bria.common.controller.im.refactoring.roomdb

import android.arch.persistence.room.Room
import android.content.Context
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.uireusable.datatypes.IdentifiableDataItem
import com.bria.common.util.Log
import com.bria.common.util.ThreadExecutors
import io.reactivex.Maybe
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import io.reactivex.subjects.Subject

class ChatRoomRepoImpl(val context : Context) : ChatRoomRepo{

companion object {
    const val TAG = "ChatRoomRepoImpl"
}

    //nvtd should be done per account
    val serverRoomsList = mutableSetOf<LocalRoomInfo>()

    val localRoomJids = mutableListOf<String>()


    var imDb: ImProviderDb = Room.databaseBuilder(context.applicationContext, ImProviderDb::class.java, "im").build()

    init {
            val disp = getAllChatRooms().observeOn(Schedulers.io()).subscribeOn(Schedulers.io()).subscribe({ roomList ->
                localRoomJids.clear()
                localRoomJids.addAll(roomList.map { it.jid })
            },{
                Log.d(TAG, "Error in loading roomList")
            })

    }

    var observableChatRoomAdded : Subject<List<ChatRoom>> = PublishSubject.create()
    var observableChatRoomUpdated : Subject<List<ChatRoom>> = PublishSubject.create()
    var observableChatRoomDelete : Subject<List<ChatRoom>> = PublishSubject.create()
    var observableMessagesAdded : Subject<List<Message>> = PublishSubject.create()
    var observableMessagesUpdated: Subject<List<Message>> = PublishSubject.create()
    var observableMessagesRemoved : Subject<List<Message>> = PublishSubject.create()
    var observableMessagesRead : Subject<List<Message>> = PublishSubject.create()

    val observableChatRoomChanged : Observable<List<ChatRoom>> = observableChatRoomAdded
            .mergeWith(observableChatRoomUpdated)
            .mergeWith(observableChatRoomDelete)

    val observableMessagesChanged : Observable<List<Message>> = observableMessagesAdded
            .mergeWith(observableMessagesUpdated)
            .mergeWith(observableMessagesRemoved)
            .mergeWith(observableMessagesRead)

    //val dataChangedObservable

    var observableServerRoomListChanged : Subject<List<LocalRoomInfo>> = PublishSubject.create()

    override fun getOnRoomAddedObservable(): Subject<List<ChatRoom>> = observableChatRoomAdded

    override fun getOnRoomUpdatedObservable(): Subject<List<ChatRoom>> = observableChatRoomUpdated
    override fun getOnRoomDeletedObservable(): Subject<List<ChatRoom>> = observableChatRoomDelete

    override fun getOnMessagesUpdatedObservable() = observableMessagesUpdated

    override fun getOnMessagesAddedObservable(): Subject<List<Message>> = observableMessagesAdded

    override fun getOnMessagesRemovedObservable(): Subject<List<Message>> = observableMessagesRemoved

    override fun getOnMessagesReadObservable(): Subject<List<Message>> = observableMessagesRead

    override fun getOnServerRoomListChangedObservable() = observableServerRoomListChanged

    override fun getAllChatRooms() : Single<List<ChatRoom>> {
        return imDb.getChatRoomDao().getAllRooms()
    }

    override fun getRoomByRoomKey(key : String) : ChatRoom? {
        return imDb.getChatRoomDao().getRoomByRoomKey(key)
    }

    override fun getChatRoom(id: Long) : ChatRoom? {
        return imDb.getChatRoomDao().getRoom(id)
    }

    override fun addChatRoom(chatRoom: ChatRoom) {
        val id  = imDb.getChatRoomDao().insertRoom(chatRoom)
        serverRoomsList += LocalRoomInfo(name = chatRoom.name, accId =  chatRoom.accountId, cleanedJid = chatRoom.chatKey)
        if (id != null) {
            chatRoom.id = id
            observableChatRoomAdded.onNext(listOf(chatRoom))
        }

    }

    override fun updateChatRoom(chatRoom: ChatRoom) {
        imDb.getChatRoomDao().updateChatRoom(chatRoom)
        observableChatRoomUpdated.onNext(listOf(chatRoom))
    }

    override fun deleteChatRoom(chatRoom: ChatRoom) {
        imDb.getChatRoomDao().deleteRoom(chatRoom)
        imDb.getChatRoomDao().deleteMessagesForChatRoom(chatRoom.id)
        observableChatRoomDelete.onNext(listOf(chatRoom))
    }

    override fun getLastMessageForChatRoom(id: Long): Maybe<Message> {
        return imDb.getChatRoomDao().getLastMessageForChatRoom(id)
    }

    override fun getNumberOfUnreadMessagesForChatRoom(id: Long): Single<Int> {
        return imDb.getChatRoomDao().getNumberOfUnreadMessagesForChatRoom(id)
    }

    override fun changeOwner(owner: String) {
        imDb.close()
        imDb = Room.databaseBuilder(context.applicationContext, ImProviderDb::class.java, "im + $owner").build()
    }

    override fun addMessage(message: Message) {
        ThreadExecutors.scheduleLightWork {
            val id = imDb.getChatRoomDao().insertMessage(message)
            if (id != null || id == -1L) {
                message.id = id
                imDb.getChatRoomDao().updateChatTime(message.creationTime, message.chatId)
                observableMessagesAdded.onNext(listOf(message))
            }
        }
    }

    //must be of same conversation
    override fun addMessages(messages: List<Message>) {
        ThreadExecutors.scheduleLightWork {
            val messageIds = imDb.getChatRoomDao().insertMessages(messages)
            messageIds.forEachIndexed { index, id ->
                id?.let { messages[index].id = id }
            }
            val message = messages.sortedWith((compareBy(Message::creationTime))).last()
            imDb.getChatRoomDao().updateChatTime(message.creationTime, message.chatId)
            observableMessagesAdded.onNext(messages)
        }
    }

    override fun updateMessages(messages: List<Message>) {
        imDb.getChatRoomDao().updateMessages(messages)
        observableMessagesUpdated.onNext(messages)
    }

    override fun deleteMessages(messages: List<Message>) {
        ThreadExecutors.scheduleLightWork {
            imDb.getChatRoomDao().deleteMessages(messages)
            observableMessagesRemoved.onNext(messages)
        }
    }

    override fun markRead(messages: List<Message>) {
        ThreadExecutors.scheduleLightWork {
            imDb.getChatRoomDao().markRead(messages.map { it.id })
            observableMessagesRead.onNext(messages)
        }
    }

    override fun markReadChatRoom(chatRoomId: Long) {
        ThreadExecutors.scheduleLightWork {
            imDb.getChatRoomDao().markReadChatRoom(chatRoomId)
            observableChatRoomUpdated.onNext(listOf(imDb.getChatRoomDao().getRoom(chatRoomId) ?: ChatRoom()))
        }
    }

    override fun getNMessageWithOffsetForChat(id: Long, offset : Int, limit : Int): List<Message> {
        return imDb.getChatRoomDao().getNMessageWithOffsetForChat(id, offset, limit)
    }

    override fun getMessageById(id: Long): Message? {
        return imDb.getChatRoomDao().getMessageById(id)
    }

    override fun addRoomInfo(info: LocalRoomInfo) {
        serverRoomsList.add(info)
        observableServerRoomListChanged.onNext(serverRoomsList.toList())
    }

    override fun addRoomInfos(infos : List<LocalRoomInfo>) {
        val filteredRooms = infos.filter { !localRoomJids.contains(it.cleanedJid) }
        serverRoomsList.addAll(filteredRooms)
        observableServerRoomListChanged.onNext(serverRoomsList.toList())
    }


    override fun removeRoomInfosForAcc(accId : String) {
        val toRemove = serverRoomsList.filter { it.accId == accId }
        serverRoomsList.removeAll(toRemove)
    }

    override fun roomNameInTheList(name : String) : Boolean{
        return serverRoomsList.firstOrNull { it.name == name } != null
    }

    override fun getServerRoomList(query : String) : MutableList<LocalRoomInfo> {
        //ntd remove rooms already in database
        if (query.isEmpty()) {
            return serverRoomsList.toList().sortedBy { it.name }.toMutableList()
        } else {
            return serverRoomsList.filter { it.name.contains(query, true) }.toList().sortedBy { it.name }.toMutableList()
        }
    }

    override fun getAllPrivateChatRooms(): Single<List<ChatRoom>> {
        return imDb.getChatRoomDao().getAllPrivateChatRooms()
    }

    override fun destroy() {
        imDb.close()
    }
}

data class LocalRoomInfo(val name : String, val accId : String, val cleanedJid : String) : IdentifiableDataItem {
    override fun getUniqueIdentifier(): String {
        return cleanedJid
    }

}
