package com.bria.common.controller.im.refactoring.xmppgroup

import android.text.TextUtils
import com.bria.common.controller.Controllers
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.IAccountsFilter
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.ImpsUtils.removePortFromDomain
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.ChatApi
import com.bria.common.controller.im.refactoring.ChatApi.*
import com.bria.common.controller.im.refactoring.IImData
import com.bria.common.controller.im.refactoring.ImDataObserverAdapter
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData.GROUPCHAT_XMPP
import com.bria.common.controller.im.refactoring.db.entities.ImStatusData
import com.bria.common.controller.im.refactoring.db.entities.InstantMessageData
import com.bria.common.controller.im.refactoring.xmpp.XmppChatInfo
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.sdkwrapper.SipStackManager
import com.bria.common.util.Log
import com.bria.common.util.im.Participant
import com.bria.common.util.im.ParticipantsSet
import com.counterpath.sdk.XmppAccount
import com.counterpath.sdk.XmppApiMultiUserChat
import com.counterpath.sdk.XmppMultiUserChat
import com.counterpath.sdk.handler.XmppMultiUserChatHandler
import com.counterpath.sdk.pb.Xmppmultiuserchat
import java.util.*

class XmppGroupChatApi(private val mChatApi: ChatApi) : XmppMultiUserChatHandler {

    private val alreadySavedMessages = LinkedList<String>()

    private val cache: XmppGroupChatCache = XmppGroupChatCache()

    private val mServerForAccount = HashMap<String, String>() //(account, mucServer)

    private val stackManager: SipStackManager?
        get() = mChatApi.stackManager

    private val imData: IImData
        get() = Controllers.get().im.imData

    private val mImDataObserverAdapter = object : ImDataObserverAdapter() {

        override fun onConversationAdded(conversation: ImConversationData) {
            if (conversation.isGroupChat) {
                cache.updateConversationId(conversation.remoteKey, conversation.id)
                //update participants list - if changed
                cache.getXmppChatInfoByConversationId(conversation.id!!)?.let { info ->
                    val newParticipantsString = info.participantsString
                    if (newParticipantsString != conversation.participants) {
                        conversation.participants = newParticipantsString
                    }
                    conversation.remoteKey = info.roomId
                }
                imData.updateConversation(conversation)
            }
        }

        override fun onConversationRemoved(conversation: ImConversationData) {
            super.onConversationRemoved(conversation)
            cache.getMultiUserChat(conversation.id)?.let { chat ->
                chat.leave("")
                cache.remove(chat)
            }
        }
    }

    init {
        imData.attachObserver(mImDataObserverAdapter)
    }

    fun startNewConversation(participants: ParticipantsSet, xmppAccount: XmppAccount): XmppGroupChatInfo? {
        Log.d(TAG, "startNewConversation: ")
        stackManager ?: return null
        val chatApi = XmppApiMultiUserChat.get(xmppAccount)
        chatApi.newChat()?.let { xmppChat ->
            val tempHistorylist = ArrayList<Xmppmultiuserchat.XmppMultiUserChatHistoryItem>()
            Log.d(TAG, "Join MUCRoom")
            val roomId = participants.serializedParticipantKeys
            xmppChat.join("", xmppAccount.settings.username, "", "message:20", tempHistorylist)  //roomID is participantstring
            val info = cache.addXmppGroupChat(roomId, xmppAccount, participants, xmppChat) //set participants as pending
            putConversation(xmppChat, roomId, GROUPCHAT_XMPP) //i na new message
            info.waitingInvitation = ParticipantsSet()
            info.waitingInvitation.addAll(participants)
            return info
        }
        return null
    }

    fun destroy() {
        cache.clear()
        imData.detachObserver(mImDataObserverAdapter)
    }

    // use only if room is not yet created
    fun sendMessage(participantsSet: ParticipantsSet, remoteKey: String, message: String): Boolean {
        val manager = stackManager ?: return false
        Log.d(TAG, "sendMessage from participant set")
        var info = cache.getXmppGroupChatInfoByRoom(remoteKey)
        if (info == null) {
            var account: Account? = null
            val xmppAccount: XmppAccount?
            val xmppAccounts = Controllers.get().accounts.getAccounts(IAccountsFilter.ACTIVE_XMPP)
            if (xmppAccounts.isNotEmpty()) {
                account = xmppAccounts[0]
            }
            if (account != null) {
                xmppAccount = manager.registrationManager.getXmppAccount(account)
            } else {
                return false
            }
            info = startNewConversation(participantsSet, xmppAccount)
        }
        if (info == null) {
            Log.e(TAG, "Unable to create group chat room")
        }
        val time = Date().time
        val chat = info!!.multiUserChat
        val messageId = chat.sendMessage(message, "")
        val imConversationData = getConversationByRemoteKeyAndType(remoteKey, info.type)
        val instantMessageData = InstantMessageData.InstantMessageDataBuilder()
                .setMessage(message)
                .setConversationId(if (imConversationData != null) imConversationData.id else ImConversationData.INVALID_CONVERSATION_ID)
                .setStatus(ImStatusData.OUTGOING_PENDING)
                .setRemoteAddress(remoteKey)
                .setTime(time)
                .setExternalId(getTemporaryPendingId(messageId)) //temporary ID
                // (until we get real ID from the server
                .createInstantMessageData()
        imData.addMessage(instantMessageData, info.roomId)
        return true
    }

    fun sendMessage(conversationData: ImConversationData, message: String): Int {
        Log.d(TAG, "sendMessage: conversationData: " + conversationData.id)
        val info = cache.getXmppChatInfoByConversationId(conversationData.id!!)
        if (info == null) {
            Log.d(TAG, "sendMessage: conversation is not active")
            return -1
        }
        val time = Date().time
        val chat = info.multiUserChat
        
        val messageId = chat.sendMessage(message, "")
        val instantMessageData = InstantMessageData.InstantMessageDataBuilder()
                .setMessage(message)
                .setConversationId(conversationData.id)
                .setStatus(ImStatusData.OUTGOING_PENDING)
                .setRemoteAddress(conversationData.remoteKey)
                .setModTime(time)
                .setTime(time)
                .setExternalId(getTemporaryPendingId(messageId)) //temporary ID
                // (until we get real ID from the server
                .createInstantMessageData()
        imData.addMessage(instantMessageData, info.roomId)
        return 0
    }

    private fun putConversation(xmppChat: XmppMultiUserChat, roomId: String, conversationType: Int) {
        Log.d(TAG, "putConversation: roomId: $roomId")
        var imConversationData: ImConversationData?
        val xmppAccount = xmppChat.api.account
        val accountKey = xmppAccount.settings.username + "@" + xmppAccount.settings.domain
        //String buddyKey = BuddyKeyUtils.getNewBuddyKey(EAccountType.Xmpp, accountKey, remoteAddress);
        imConversationData = getConversationByRemoteKeyAndType(roomId, conversationType)
        val chatInfo = cache.getXmppChatInfo(xmppChat)
        if (imConversationData == null) {
            imConversationData = ImConversationData.ImConversationDataBuilder()
                    .setAccountId(accountKey)
                    .setDisplayName("")
                    .setModTime(Date().time)
                    .setType(conversationType)
                    .setParticipants(if (chatInfo != null) chatInfo.participantsString else roomId)
                    .setRemoteKey(roomId)
                    .createImConversationData()
            Log.d(TAG, "putConversation: addConversation: id: " + imConversationData!!.id)
            imData.addConversation(imConversationData)
        } else {
            cache.updateConversationId(roomId, imConversationData.id!!)
        }
    }

    fun addParticipants(conversationId: Long, newParticipants: ParticipantsSet) {
        Log.d(TAG, "addParticipants: convId: $conversationId")
        cache.getXmppChatInfoByConversationId(conversationId)?.let { info ->
            newParticipants.forEach { participant ->
                info.multiUserChat.invite(participant.remoteAddress, "")
            }
        }
        imData.getConversationById(conversationId)?.let { imConversationData ->
            val oldParticipants = ParticipantsSet(imConversationData.participants)
            oldParticipants.addAll(newParticipants)
            imConversationData.participants = oldParticipants.serializedParticipantKeys
            imData.updateConversation(imConversationData)
        }

    }

    override fun onRoomListRetrieved(xmppAccount: XmppAccount, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.RoomListRetrievedEvent) {
        val manager = stackManager ?: return
        val roomDetails = StringBuilder("\n")
        roomDetails.append("RoomListRetrieved:\n ")
        val xmppApiGroupChat = XmppApiMultiUserChat.get(xmppAccount)
        val account = manager.registrationManager.getAccount(xmppAccount)
        for (roomListItem in evt.roomsList) {
            val jid = roomListItem.jid
            var info = cache.getXmppGroupChatInfoByRoom(jid)
            val imConversationData = getConversationByRemoteKeyAndType(jid, ImConversationData.GROUPCHAT_XMPP)
            //            if (imData == null) {
            //                imData = getConversationByRemoteKeyAndType(jid, GROUPCHAT_COLLAB);
            //            }
            roomDetails.append("    Room  name").append(roomListItem.name).append(" JID: ").append(roomListItem.jid).append("\n")
            if (imConversationData != null && (info == null || info.multiUserChat == null)) {
                //join xmppChat
                val chat = xmppApiGroupChat.newChat()
                //IOS Way
                //NSTimeInterval ti = [self.lastMessage.time timeIntervalSince1970] + account.serverTimeShift;
                //historySince = [@"since:" stringByAppendingFormat:@"%ld", (long)ti];
                //[[CpcServicesImpl instance].cpcXmppMUCManager join:_xmppMUCHandle room:roomUri.aor nickname:CpcDecryptDataToString(account.usernameData,
                //        NSStringFromClass([CpcAccount class])) password:nil historyRequester:historySince historyToAdd:nil];
                var deltaTimeStr = ""
                val messageData = imData.getLastMessageForConversation(imConversationData.id!!)
                if (messageData != null) {
                    val deltaTime = (System.currentTimeMillis() - messageData.time) / 1000 - 1
                    deltaTimeStr = "seconds:" + deltaTime.toString()
                }
                val tempHistoryList = ArrayList<Xmppmultiuserchat.XmppMultiUserChatHistoryItem>()
                chat.join(jid, account!!.getStr(EAccountSetting.UserName), "", deltaTimeStr, tempHistoryList)
                if (info == null) {
                    val participantsSet = ParticipantsSet(imConversationData.participants)
                    info = cache.addXmppGroupChat(jid, xmppAccount, participantsSet, chat)
                    info!!.conversationId = imConversationData.id!!
                    info.isActive = true
                    info.updateParticipants()
                    mChatApi.fireOnRoomStatusChange(info.roomId, true)
                } else {
                    info.setChat(chat)
                }
            }
        }
        Log.d(TAG, roomDetails.toString())
        cache.print()
        rejoinGroupChats(account, xmppApiGroupChat)
    }

    private fun rejoinGroupChats(account: Account?, xmppApiGroupChat: XmppApiMultiUserChat) {
        var accountId = ImpsUtils.getUserAtDomainForAccount(account!!)
        accountId = ImpsUtils.removePortFromDomain(accountId)
        val groupChats = imData.getAllGroupChatConversations(accountId)
        for (imConversationData in groupChats) {
            //rejoin only if last message is in last 24h
            val MILIS_IN_DAY = (24 * 60 * 60 * 1000).toLong()
            if (System.currentTimeMillis() - imConversationData.modTime > MILIS_IN_DAY) {
                break
            }
            //if room was not previously created than should not rejoin
            if (TextUtils.isEmpty(imConversationData.participants) || TextUtils.equals(imConversationData.participants, imConversationData
                            .remoteKey)) {
                return
            }
            var info = cache.getXmppGroupChatInfoByRoom(imConversationData.remoteKey)
            val chat: XmppMultiUserChat
            if (info == null || info.multiUserChat == null) {
                chat = xmppApiGroupChat.newChat()
                Log.d(TAG, "rejoinGroupChats: chat added: " + chat.toString())
                if (info == null) {
                    val participantsSet = ParticipantsSet(imConversationData.participants)
                    info = cache.addXmppGroupChat(imConversationData.remoteKey, chat.api.account, participantsSet, chat)
                    info!!.conversationId = imConversationData.id!!
                    info.isActive = true
                }
                info.setChat(chat)
            } else {
                chat = info.multiUserChat
            }

            val tempHistorylist = ArrayList<Xmppmultiuserchat.XmppMultiUserChatHistoryItem>()
            val deltaTimeStr = "seconds:86400" //24* 60 * 60 give me all messages since yesterday
            Log.d(TAG, "rejoinGroupChats:")
            chat.join(imConversationData.remoteKey, account.getStr(EAccountSetting.UserName), "", deltaTimeStr, tempHistorylist)
        }
    }

    fun initCollabChat(roomId: String, username: String, account: XmppAccount): Long {
        cache.addAsConferenceRoom(roomId)
        val imConversationData = getConversationByRemoteKeyAndType(roomId, ImConversationData.GROUPCHAT_COLLAB)
        var info = cache.getXmppGroupChatInfoByRoom(roomId)
        if (imConversationData != null && info != null) {
            return imConversationData.id!!
        }
        val userAtDomain = ImpsUtils.removePortFromDomain(account.settings.username + "@" + account.settings.domain)
        if (!mServerForAccount.containsKey(userAtDomain)) {
            Log.d(TAG, "Service not available for $userAtDomain")
            cache.addToWaitingList(CollabChatInfo(roomId, username, account))
            return -1
        }
        var chat: XmppMultiUserChat? = null
        if (info != null) {
            chat = info.multiUserChat
        }
        if (chat == null) {
            val chatApi = XmppApiMultiUserChat.get(account)
            chat = chatApi.newChat()
            val tempHistorylist = ArrayList<Xmppmultiuserchat.XmppMultiUserChatHistoryItem>()
            val deltaTimeStr = "seconds:7200" //2* 60 * 60 give me all messages in last 2 hours

            chat!!.join(roomId, username, "", deltaTimeStr, tempHistorylist)
            Log.d(TAG, "initCollabChat: chat added: " + chat.toString())
        }
        if (info == null) {
            info = cache.addXmppGroupChat(roomId, account, ParticipantsSet(), chat)
            info!!.conversationId = if (imConversationData != null) imConversationData.id else -1
            info.isActive = true
            info.setChat(chat)
            mChatApi.fireOnRoomStatusChange(info.roomId, true)
        } else {
            info.setChat(chat)
        }
        return if (imConversationData == null) {
            putConversation(chat, roomId, ImConversationData.GROUPCHAT_COLLAB)
            -1
        } else {
            imConversationData.id!!
        }
    }

    //Added cause sometimes roomId has domain, and sometimes don't
    private fun getConversationByRemoteKeyAndType(remoteKey: String, type: Int): ImConversationData? {
        var data: ImConversationData? = imData.getConversationByRemoteKeyAndType(remoteKey, type)
        if (data == null) {
            val username = ImpsUtils.removeDomainFromAddress(remoteKey)
            if (!TextUtils.isEmpty(username)) {
                data = imData.getConversationByRemoteKeyAndType(username, type)
            }
        }
        return data
    }

    override fun onParticipantAdded(xmppMultiUserChat: XmppMultiUserChat?, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantAddedEvent) {

        Log.d(TAG, "onParticipantAdded: xmppMultiUserChat not null: " + (xmppMultiUserChat != null))
        val info = cache.getXmppChatInfo(xmppMultiUserChat)
        val account = xmppMultiUserChat!!.api.account
        val userName = account.settings.username
        val domain = account.settings.domain
        val accAddress = "$userName@$domain"
        var remoteAddress = evt.state.jid
        //fix if jid not received
        if (TextUtils.isEmpty(remoteAddress)) {
            remoteAddress = evt.nickname
        }
        val newParticipant = BuddyKeyUtils.getNewBuddyKey(EAccountType.Xmpp, accAddress, remoteAddress)
        if (info == null) {
            Log.d(TAG, "onParticipantAdded: info should not be null!")
            //nvtd try to recover from this,
            // for example create dummy roomId, and when roomId is matched with XmppChat update roomId and update unsaved messages
            // if conversation is created on disk before roomId is received, do not save message with dummy roomId until real roomId is received!
            return
        } else {
            info.setParticipantActive(newParticipant)
        }
        val messageText = PARTICIPANT_JOINED_EVENT_ID + LOCAL_MESSAGE_SEPARATOR + newParticipant
        val time = System.currentTimeMillis()
        val externalId = newParticipant + ":" + time.toString()
        val instantMessageData = InstantMessageData.InstantMessageDataBuilder()
                .setConversationId(info.conversationId)
                .setRemoteAddress("")
                .setStatus(ImStatusData.LOCAL)
                .setMessage(messageText)
                .setExternalId(externalId)
                .setTime(time)
                .setModTime(time)
                //.setXmppThreadId()
                .createInstantMessageData()
        val conversationData = getConversationByRemoteKeyAndType(info.roomId, info.type)
        if (conversationData != null) {
            instantMessageData.conversationId = conversationData.id
            if (!info.isCollab) { //do not log join event on collab
                imData.addMessage(instantMessageData, info.roomId)
            }
            conversationData.participants = info.participantsString
            imData.updateConversation(conversationData)
        }
    }

    override fun onParticipantRemoved(xmppMultiUserChat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantRemovedEvent) {

        Log.d(TAG, "onParticipantRemoved: " + evt.nickname)
        val info = cache.getXmppChatInfo(xmppMultiUserChat)
        val account = xmppMultiUserChat.api.account
        val userName = account.settings.username
        val domain = account.settings.domain
        var remoteAddress = evt.jid
        if (TextUtils.isEmpty(remoteAddress)) {
            remoteAddress = evt.nickname
        }
        val removedParticipant = BuddyKeyUtils.getNewBuddyKey(EAccountType.Xmpp, "$userName@$domain", remoteAddress)
        if (info == null) {
            return
        }
        val messageText = PARTICIPANT_LEFT_EVENT_ID + LOCAL_MESSAGE_SEPARATOR + removedParticipant
        val time = System.currentTimeMillis()
        val externalId = removedParticipant + ":" + time.toString()
        val instantMessageData = InstantMessageData.InstantMessageDataBuilder()
                .setConversationId(info.conversationId)
                .setRemoteAddress("")
                .setStatus(ImStatusData.LOCAL)
                .setMessage(messageText)
                .setExternalId(externalId)
                .setTime(time)
                .setModTime(time)
                //.setXmppThreadId()
                .createInstantMessageData()
        getConversationByRemoteKeyAndType(info.roomId, info.type)?.let { conversationData ->
            instantMessageData.conversationId = conversationData.id
            if (!info.isCollab) { //do not log left event on collab
                imData.addMessage(instantMessageData, info.roomId)
            }
            conversationData.participants = info.participantsString
            imData.updateConversation(conversationData)
        }
    }

    override fun onParticipantUpdated(xmppMultiUserChat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantUpdatedEvent) {
        Log.d(TAG, "onParticipantUpdated: " + evt.state.presence)

        // not implemented for now.
    }

    override fun onParticipantSelfUpdated(xmppMultiUserChat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantSelfUpdatedEvent) {
        Log.d(TAG, "onParticipantSelfUpdated: ")
    }

    override fun onMultiUserChatReady(chat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatReadyEvent) {
        Log.d(TAG, "onMultiUserChatReady: " + evt.room)
        val info = cache.getXmppChatInfo(chat)
        if (info != null) {
            Log.d(TAG, "onMultiUserChatReady: room found " + info.conversationId)
            info.roomId = evt.room
            info.activate()
            val imConversationData = imData.getConversationById(info.conversationId)
            if (imConversationData != null) {
                imConversationData.remoteKey = evt.room
                imData.updateConversation(imConversationData)
                mChatApi.fireOnRoomStatusChange(info.roomId, true)
            }
            info.waitingInvitation?.let { participants ->
                participants.forEach { participant ->
                    chat.invite(participant.remoteAddress, chat.api.account.settings.username + " invites you to Group Chat") //temporary hard coded
                }
                info.waitingInvitation.clear()
            }


        }

        chat.requestConfigurations()
    }

    override fun onMultiUserChatSubjectChanged(xmppMultiUserChat: XmppMultiUserChat,
                                               evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatSubjectChangedEvent) {
        Log.d(TAG, "onMultiUserChatSubjectChanged: " + evt.subject)
    }

    override fun onMultiUserChatNewMessage(chat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatNewMessageEvent) {
        Log.d(TAG, "onMultiUserChatNewMessage to chat : " + chat.toString())

        val info = cache.getXmppChatInfo(chat)
        //updating cache
        if (info == null) {
            Log.e(TAG, "Unknown state. We received a messsage, but didn't receive invitation")
            return
        } else if (info.conversationId == XmppChatInfo.INVALID_CONVERSATION_ID) {
            if (!TextUtils.isEmpty(info.roomId)) {
                val imConversationData = getConversationByRemoteKeyAndType(info.roomId, if (info.isCollab)
                    ImConversationData.GROUPCHAT_COLLAB
                else
                    ImConversationData.GROUPCHAT_XMPP)
                if (imConversationData != null) {
                    cache.updateConversationId(info.roomId, imConversationData.id!!)
                }
            }
        }
        //our message, ignore
        val xmppAccount = chat.api.account
        val userName = xmppAccount.settings.username
        val domain = xmppAccount.settings.domain
        var remoteAddress = evt.nickname
        if (!remoteAddress.contains("@")) {
            remoteAddress = "$remoteAddress@$domain"
        }
        val instantMessageData: InstantMessageData?
        if (TextUtils.equals(remoteAddress, userName) || TextUtils.equals(remoteAddress, "$userName@$domain")) {
            if (alreadySavedMessages.contains(evt.messageId)) {
                alreadySavedMessages.remove(evt.messageId)
                return
            }
            instantMessageData = InstantMessageData.InstantMessageDataBuilder()
                    .setMessage(evt.plain)
                    .setConversationId(info.conversationId)
                    .setStatus(ImStatusData.OUTGOING_SENT)
                    .setRemoteAddress(remoteAddress)
                    .setTime(evt.timestamp * 1000)
                    .setModTime(evt.timestamp * 1000)
                    .setExternalId(evt.messageId)
                    .createInstantMessageData()
        } else {
            instantMessageData = InstantMessageData.InstantMessageDataBuilder()
                    .setConversationId(info.conversationId)
                    .setRemoteAddress(remoteAddress)
                    .setStatus(if (info.isCollab) ImStatusData.INCOMING_COLLAB_UNREAD else ImStatusData.INCOMING_UNREAD)
                    .setMessage(evt.plain)
                    .setExternalId(evt.messageId)
                    .setTime(evt.timestamp * 1000)
                    .setModTime(evt.timestamp * 1000)
                    //.setXmppThreadId()
                    .createInstantMessageData()
        }
        if (instantMessageData == null) {
            return
        }
        if (info.conversationId == ImConversationData.INVALID_CONVERSATION_ID) {
            imData.addMessage(instantMessageData, info.roomId)
        } else {
            imData.addMessage(instantMessageData)
        }
    }

    override fun onParticipantChatStateReceived(xmppMultiUserChat: XmppMultiUserChat,
                                                participantChatStateEvent: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantChatStateEvent) {
        Log.d(TAG, "onParticipantChatStateReceived: ")
    }

    override fun onMultiUserChatInvitationReceived(chat: XmppMultiUserChat,
                                                   evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatInvitationReceivedEvent) {
        val manager = stackManager ?: return

        Log.d(TAG, "onMultiUserChatInvitationReceived() Jid: - " + evt.jid + "r: " + evt.reason + " r:" + evt.room + " delay:" + evt
                .isDelayedDelivery + " ms:" + evt
                .millisecond + " time:" + evt.timeStamp)
        val xmppAccount = chat.api.account
        val account = manager.registrationManager.getAccount(xmppAccount)
        val tempHistorylist = ArrayList<Xmppmultiuserchat.XmppMultiUserChatHistoryItem>()

        //chat.accept(account.getStr(EAccountSetting.UserName), "since: " + evt.getTimeStamp(), tempHistorylist); //copied from ImController
        chat.accept(account!!.getStr(EAccountSetting.UserName), "message:0", tempHistorylist) //copied from ImController
        // do not add creator in list of participants
        //String creatorKey = getNewBuddyKey(EAccountType.Xmpp, ImpsUtils.getUserAtDomainForAccount(account), evt.getJid());
        //participantSet.add(creatorKey);
        if (cache.getXmppChatInfo(chat) == null) {
            cache.addXmppGroupChat(evt.room, xmppAccount, ParticipantsSet(), chat)
            putConversation(chat, evt.room, GROUPCHAT_XMPP)
        }
    }

    override fun onMultiUserChatInvitationDeclined(xmppMultiUserChat: XmppMultiUserChat,
                                                   evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatInvitationDeclinedEvent) {

    }

    override fun onMultiUserChatError(xmppMultiUserChat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatErrorEvent) {
        val info = cache.getXmppChatInfo(xmppMultiUserChat)

        Log.d(TAG, "onMultiUserChatError: " + evt.error + " ;type= " + evt.type)
        if (info != null) {
            Log.d(TAG, "onMultiUserChatError: roomId: " + info.roomId + " Conv ID: " + info.conversationId)
        }
    }

    override fun onLocalUserLeft(xmppMultiUserChat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.LocalUserLeftEvent) {
        Log.d(TAG, "onLocalUserLeft: " + evt.reason)
    }

    override fun onMultiUserChatConfigurationRequested(xmppMultiUserChat: XmppMultiUserChat,
                                                       evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatConfigurationRequestedEvent) {
        Log.d(TAG, "onMultiUserChatConfigurationRequested: " + evt.configurations.instructions)
        evt.xmppDataForm.fields.firstOrNull { it.name == "muc#roomconfig_anonymity" }?.let { field ->
            if (field.values.isNotEmpty() && field.values[0] != "nonanonymous") {
                field.values[0] = "nonanonymous"
            }
        }
        xmppMultiUserChat.setConfigurations(evt.xmppDataForm)
    }

    override fun onMultiUserChatRoomStateChanged(chat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatRoomStateChangedEvent) {
        Log.d(TAG, "onMultiUserChatRoomStateChanged() event came in to APP")
        //for room details changed so for now print to console.
        var roomDetails = "\n"
        roomDetails += "Room Name: " + evt.state.name
        roomDetails += "\nCreation Info: " + evt.state.creation
        roomDetails += "\nDescription: " + evt.state.description
        roomDetails += "\nSubject: " + evt.state.subject
        roomDetails += "\nIs Created By Self: " + evt.state.isCreatedBySelf
        roomDetails += "\nIs Moderated: " + evt.state.isModerated
        roomDetails += "\nIs Open: " + evt.state.isOpen
        roomDetails += "\nIs Password Proteced: " + evt.state.isPasswordProteced
        roomDetails += "\nIs Persistent: " + evt.state.isPersistent
        roomDetails += "\nIs Public: " + evt.state.isPublic
        roomDetails += "\nIs Ready: " + evt.state.isReady
        roomDetails += "\nIs Recorded: " + evt.state.isRecorded
        roomDetails += "\nNumber of Participants: " + evt.state.numberOfParticipants + "\n"
        Log.d(TAG, "onMultiUserChatRoomStateChanged() gave the following status update: $roomDetails")
    }

    override fun onSendMessageSuccess(xmppMultiUserChat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.SendMessageSuccessEvent) {
        updateMessageIdAndStatus(evt.message, evt.messageId, ImStatusData.OUTGOING_SENT)
        Log.d(TAG, "onSendMessageSuccess message: " + evt.message + " message id: " + evt.messageId)
    }

    override fun onSendMessageFailure(xmppMultiUserChat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.SendMessageFailureEvent) {
        updateMessageIdAndStatus(evt.message, Date().time.toString(), ImStatusData.OUTGOING_SENDING_FAILED)
        Log.d(TAG, "onSendMessageFailure message: " + evt.message)
    }

    private fun updateMessageIdAndStatus(tempId: Int, externalId: String, status: Int) {
        alreadySavedMessages.add(externalId)
        val temporary = getTemporaryPendingId(tempId)
        val instantMessageData = imData.getMessageByExternalId(temporary)
        if (instantMessageData != null) {
            instantMessageData.status = status
            instantMessageData.externalId = externalId
            imData.updateMessage(instantMessageData)
        }
    }

    override fun onServiceAvailability(chat: XmppMultiUserChat?, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.ServiceAvailabilityEvent?) {
        val manager = stackManager ?: return
        if (evt != null) {
            Log.d(TAG, "onServiceAvailability() - " + evt.available + " " + evt.service)
        } else {
            Log.d(TAG, "onServiceAvailability() - evt null")
        }
        if (evt == null || !evt.available || TextUtils.isEmpty(evt.service)) {
            return
        }
        if (chat == null || chat.api == null || chat.api.account == null) {
            return
        }

        val xmppAccount = chat.api.account
        val account = manager.registrationManager.getAccount(xmppAccount) ?: return
        val strippedDomain = removePortFromDomain(account.getStr(EAccountSetting.Domain))
        val userAtDomain = account.getStr(EAccountSetting.UserName) + "@" + strippedDomain
        mServerForAccount[userAtDomain] = evt.service
        val xmppApiGroupChat = XmppApiMultiUserChat.get(xmppAccount)
        xmppApiGroupChat.getRoomList()

        val infos = cache.waitingCollabInfos
        for ((roomId, username, account1) in infos) {
            initCollabChat(roomId, username, account1)
        }
        infos.clear()
    }

    override fun onMultiUserChatListRequested(xmppMultiUserChat: XmppMultiUserChat,
                                              evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatListRequestedEvent) {
        Log.d(TAG, "onMultiUserChatListRequested$evt")
    }

    fun isTyping(conversationId: Long): Boolean {
        val info = cache.getXmppChatInfoByConversationId(conversationId)
        return info != null && info.isTyping
    }

    private fun getTemporaryPendingId(messageId: Int): String = String.format("PENDING_GROUP-%s", messageId)

    fun isParticipantActive(conversationId: Long, participant: Participant): Boolean {
        val info = cache.getXmppChatInfoByConversationId(conversationId)
        return info?.isParticipantActive(participant) ?: false
    }

    fun isGroupChatActive(conversationId: Long): Boolean {
        val info = cache.getXmppChatInfoByConversationId(conversationId) ?: return false
        return info.isActive
    }

    fun isGroupChatCapable(buddyKey: String): Boolean {
        val accountKey = BuddyKeyUtils.getAccountFromNewBuddyKey(buddyKey)
        val userAtDomain = ImpsUtils.removePortFromDomain(accountKey)
        return mServerForAccount.containsKey(userAtDomain)
    }

    fun leaveGroupChat(roomId: String) {
        val info = cache.getXmppGroupChatInfoByRoom(roomId) ?: return
        if (info.isCollab && info.conversationId != INVALID_CONVERSATION_ID) {
            imData.getConversationById(info.conversationId)?.let { imConversationData ->
                imData.removeConversation(imConversationData)
            }
        }
        info.multiUserChat?.let {chat ->
            cache.remove(chat)
            chat.leave("")
        }
    }

    fun clearCacheForAccount(xmppAccount: XmppAccount?) {
        xmppAccount ?: return
        mServerForAccount.remove(ImpsUtils.removePortFromDomain("""${xmppAccount.settings.username}@${xmppAccount.settings.domain}"""))
        cache.clear(xmppAccount)
    }

    override fun onRoomBookmarksReceived(chat: XmppMultiUserChat?, args: Xmppmultiuserchat.XmppMultiUserChatEvents.RoomBookmarksReceivedEvent?) {

    }


    companion object {
        private val TAG = "XmppGroupChatApi"
    }
}

