package com.bria.common.controller.phone

import android.content.Context
import com.bria.common.controller.Controllers
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.calllog.db.CallLogEntity
import com.bria.common.controller.calllog.db.CallType
import com.bria.common.controller.contacts.local.Contact
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.ESetting
import com.bria.common.sdkwrapper.CallData
import com.bria.common.sdkwrapper.CallManager
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import com.bria.common.util.RecordingUtils
import com.bria.common.util.Utils
import java.util.*

private const val TAG = "CallLogHelper"

class CallLogHelper(private val context: Context, private val controllers: Controllers) {

    fun generateCallLogEntity(callData: CallData) : CallLogEntity? {
        if (!controllers.settings.getBool(ESetting.CallLogV2)) {
            Log.d(TAG, "Call Log v2 disabled, old API is used")
            return null
        }

        if (callData.isSuppressLog) {
            Log.d(TAG, "Call Log suppressed")
            return null
        }

        val remoteExtension = getRemoteExtension(callData)
        val account = getAccount(callData)
        val contact = getContact(callData)

        return CallLogEntity(
                number = remoteExtension,
                numberType = getNumberType(contact, remoteExtension),
                callType = getCallType(callData),
                date = callData.callStartTimestamp,
                durationSeconds = getDurationSeconds(callData),
                contactId = getContactId(callData, contact),
                displayName = getDisplayName(callData, contact),
                provisioningUser = getProvisioningUser(),
                accountNickname = callData.accountNickname,
                accountUsername = account?.getStr(EAccountSetting.UserName) ?: "",
                accountDomain = account?.getStr(EAccountSetting.Domain) ?: "",
                remoteDomain = getRemoteDomain(callData, remoteExtension),
                remoteName = callData.remoteSipDisplayName,
                recordPath = getRecordPath(callData),
                forwardedTo = callData.forwardTo ?: "",
                hostedCollab = callData.isHostedCollab,
                new = true
        )
    }

    private fun isVccsCall(callData: CallData) = controllers.collaboration.monitoredConference != null &&
            (!callData.collabUrl.isNullOrEmpty() || callData.isHostedCollab)

    private fun isVoiceMail(callData: CallData) = callData.isVoiceMail(context) // todo: replace this deprecated method

    private fun getDurationSeconds(callData: CallData): Int {
        return if (callData.callAnswerTimestamp > 0 && callData.callDisconnectTimestamp > callData.callAnswerTimestamp) {
            (callData.callDisconnectTimestamp - callData.callAnswerTimestamp) / 1000
        } else {
            0
        }.toInt()
    }

    private fun getContactId(callData: CallData, contact: Contact?) : String {
        if (contact != null) {
            if (contact.type == Contact.Type.LDAP || contact.type == Contact.Type.GENBAND_DIRECTORY) {
                return GlobalConstants.CALL_LOG_LDAP_CONTACT
            }
            if (contact.type == Contact.Type.GENBAND_FRIEND) {
                return GlobalConstants.CALL_LOG_GENBAND_FRIEND_ID
            }
            if (isVccsCall(callData)) {
                return GlobalConstants.CALL_LOG_VCCS_CONFERENCE
            }
        }

        return contact?.id ?: ""
    }

    private fun getDisplayName(callData: CallData, contact: Contact?) : String {
        return controllers.collaboration.monitoredConference?.details?.title ?: (contact?.displayName ?: callData.remoteDisplayName)
    }

    private fun getNumberType(contact: Contact?, number: String) : String {
        return contact?.allPhoneNumbers?.find { it -> it.number == number }?.getSubTypeString(context) ?: ""
    }

    private fun getCallType(callData: CallData) : CallType {
        if (callData.direction == CallData.DIRECTION_OUTGOING) {
            return when {
                isVccsCall(callData) -> CallType.CONFERENCE
                isVoiceMail(callData) -> CallType.VOICE_MAIL
                else -> CallType.OUTGOING
            }
        } else {
            val duration = getDurationSeconds(callData)
            var callType = CallType.INCOMING

            if (duration <= 0 && !callData.callRedirected) {
                callType = CallType.MISSED
            }

            when (callData.statusCode) {
                CallManager.STATUS_CODE_BUSY_HERE -> {
                    if (callData.callCancelled) {
                        callType = CallType.INCOMING
                    }
                }

                CallManager.STATUS_CODE_REQUEST_TERMINATED or
                CallManager.STATUS_CODE_MOVED_TEMPORARILY -> {
                    if (duration <= 0) {
                        callType = CallType.MISSED
                    }
                }
            }

            if (callData.isSuppressMissedCall) {
                callType = CallType.INCOMING
            }

            return callType
        }
    }

    private fun getRemoteExtension(callData: CallData) : String {
        if (callData.isHostedCollab && !callData.collabUrl.isNullOrEmpty()) {
            return callData.collabUrl
        }

        if (controllers.settings.getBool(ESetting.FeatureGenband) && controllers.settings.getBool(ESetting.ShowUriDomain)) {
            return callData.remoteUserWithDomain
        }

        // If it is an outgoing call, use the string with dial plan settings applied
        // else use the callData to get the remote extension
        if (callData.direction == CallData.DIRECTION_OUTGOING) {
            return callData.remoteUserNoDialPlanApplied
        }

        if (callData.remoteUri.contains("@")) {
            // Remote uri contains username i.e. CLID enabled
            val tokenizer = StringTokenizer(callData.remoteUri, "@")
            tokenizer.nextToken()
            var remoteDomain = tokenizer.nextToken()

            if (remoteDomain.endsWith(">")) {
                remoteDomain = remoteDomain.substring(0, remoteDomain.length - 1)
            }

            val account = controllers.accounts.primaryAccount
            if (account != null && account.getStr(EAccountSetting.Domain).compareTo(remoteDomain) != 0) {
                if (!Utils.Brands.isBrand(context, "Phytter")) {
                    return callData.remoteUser + "@" + remoteDomain
                }
            }
        }

        return callData.remoteUser
    }

    private fun getRemoteDomain(callData: CallData, remoteExtension: String) : String =
            if (!callData.remoteDomain.isNullOrEmpty()) callData.remoteDomain else ImpsUtils.getDomainFromAddress(remoteExtension)

    private fun getAccount(callData: CallData) : Account? = controllers.accounts.getAccount(callData.accountNickname)

    private fun getContact(callData: CallData) : Contact? {
        val searchDisplayName = if (callData.direction == CallData.DIRECTION_INCOMING) "" else callData.remoteDisplayName
        var searchNumber = callData.remoteUserNoDialPlanApplied
        if (!callData.remoteDomain.isNullOrEmpty()) {
            val domain = ImpsUtils.removePortFromDomain(callData.remoteDomain)
            searchNumber = ImpsUtils.getAddressWithDomain(callData.remoteUserNoDialPlanApplied, domain)
        }

        return controllers.contacts
                .contactsApi
                .contact
                .byNumber(searchNumber)
                .andDisplayName(searchDisplayName)
                .asMaybe()
                .onErrorComplete()
                .blockingGet()
    }

    private fun getProvisioningUser() : String {
        return when {
            controllers.settings.getBool(ESetting.FeatureProvisioning) -> controllers.settings.getStr(ESetting.ProvisioningUsername)
            else -> ""
        }
    }

    private fun getRecordPath(callData: CallData) : String {
        if (controllers.phone.callRecordingEnabled()) {
            // auto-stop call recording (end of call)
            val callId = callData.callId
            var callRecordingStopped = false
            if (controllers.base.sipStackManager.callManager.isPartOfActiveCallRecording(callId)) {
                callRecordingStopped = controllers.base.sipStackManager.callManager.stopCallRecording(callId)
            }
            // always check if a temporary .wav file exists, matching this callId
            if (callRecordingStopped) {
                val lastId = controllers.callLog.api.getAll().blockingFirst(mutableListOf())
                        .maxBy { it.id }?.id?.toInt() ?: 0
                val recordPath = RecordingUtils.getFileNameFromCallLogId(lastId + 1)
                if (RecordingUtils.renameTempFile(context, callId, recordPath)) {
                    // successful rename (it means that we have recorded this call)
                    return recordPath
                }
            }
        }
        return ""
    }
}
