package com.bria.common.customelements.internal.views.buttons

import android.content.Context
import android.graphics.*
import android.graphics.drawable.Drawable
import android.support.annotation.ColorInt
import android.text.TextPaint
import android.util.AttributeSet
import android.util.Log
import android.view.View
import com.bria.common.R
import com.bria.common.R.string.*
import com.bria.common.controller.ClientConfig
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.modules.BriaGraph
import com.bria.common.uiframework.coloring.Coloring

/**
 * A custom view used for Dialer screen buttons to optimize rendering times.
 */
class DialerButton : View {

    private companion object {
        private val TAG = "DialerButton"!!
        private val ALPHA_FULL = 255
        private val ALPHA_DIM = 150
        private val NUMBER_SIZE_RATIO = 0.45f
        private val TEXT_SIZE_RATIO = 0.2f
        private val THIN_FONT_PATH = "fonts/Roboto-Thin.ttf"

        val TEXT_RESOURCES_MAP = mapOf(
                tKeypad1 to tKeypad1text,
                tKeypad2 to tKeypad2text,
                tKeypad3 to tKeypad3text,
                tKeypad4 to tKeypad4text,
                tKeypad5 to tKeypad5text,
                tKeypad6 to tKeypad6text,
                tKeypad7 to tKeypad7text,
                tKeypad8 to tKeypad8text,
                tKeypad9 to tKeypad9text,
                tKeypad0 to tKeypad0text,
                tKeypadAstx to tKeypadAstxText,
                tKeypadHash to tKeypadHashText
        )
    }

    private lateinit var numberPaint: TextPaint
    private lateinit var numberBounds: RectF
    private lateinit var number: String
    private lateinit var textPaint: TextPaint
    private lateinit var textBounds: RectF
    private lateinit var text: String
    private var hideText: Boolean = false
    private var contentColor: Int = Color.RED
    private var icon: Drawable? = null

    private var coloring: Coloring? = null
    private var storage: Settings? = null
    private lateinit var textMap: Map<String, String>

    // <editor-fold desc="Constructors">
    constructor(context: Context) : super(context) {
        init(null, 0, 0)
    }

    constructor(context: Context, attrs: AttributeSet) : super(context, attrs) {
        init(attrs, 0, 0)
    }

    constructor(context: Context, attrs: AttributeSet, defStyle: Int) : super(context, attrs, defStyle) {
        init(attrs, defStyle, 0)
    }

    @Suppress("unused")
    constructor(context: Context, attrs: AttributeSet, defStyle: Int, defStyleRes: Int) : super(context, attrs, defStyle, defStyleRes) {
        init(attrs, defStyle, defStyleRes)
    }
    // </editor-fold>

    private fun init(attrs: AttributeSet?, defStyle: Int, defStyleRes: Int) {
        // load attributes
        val attributes = context.obtainStyledAttributes(attrs, R.styleable.DialerButton, defStyle, defStyleRes)
        icon = attributes.getDrawable(R.styleable.DialerButton_icon)
        number = attributes.getString(R.styleable.DialerButton_number)
        checkNotNull(number, { "Number cannot be null" })
        hideText = attributes.getBoolean(R.styleable.DialerButton_hide_text, false)
        attributes.recycle()

        // convert resources to strings
        textMap = mutableMapOf()
        for ((key, value) in TEXT_RESOURCES_MAP) {
            textMap += context.getString(key) to context.getString(value)
        }

        coloring = if (!isInEditMode) Coloring.get() else null
        storage = if (!isInEditMode) BriaGraph.settings else null
        setupText()
        setupIcon()
        setupPaints()

        if (ClientConfig.get().isDebugMode) {
            Log.d(TAG, "View created") // too many logs - debug only
        }
    }

    override fun onSizeChanged(w: Int, h: Int, oldW: Int, oldH: Int) {
        super.onSizeChanged(w, h, oldW, oldH)
        setupSizes()
        invalidate()
        if (ClientConfig.get().isDebugMode) {
            Log.d(TAG, "View size changed (${oldW}x$oldH -> ${w}x$h)")
        }
    }

    override fun onMeasure(widthMeasureSpec: Int, heightMeasureSpec: Int) {
        super.onMeasure(widthMeasureSpec, heightMeasureSpec)
        if (ClientConfig.get().isDebugMode) {
            Log.d(TAG, "View measured (${MeasureSpec.getSize(widthMeasureSpec)}x${MeasureSpec.getSize(heightMeasureSpec)})")
        }
    }

    override fun onLayout(changed: Boolean, left: Int, top: Int, right: Int, bottom: Int) {
        super.onLayout(changed, left, top, right, bottom)
        if (ClientConfig.get().isDebugMode) {
            Log.d(TAG, "View laid out${if (changed) ", changed" else ""}: ${(right - left)}x${bottom - top}")
        }
    }

    private fun setupText() {
        text = checkNotNull(textMap[number], { "Sub-text not found for $number" })
    }

    private fun setupIcon() {
        if (!isInEditMode) {
            val backgroundActiveColor = if (isInEditMode) Color.GRAY else coloring!!.decodeColor(storage?.getStr(ESetting.ColorPhoneSelection))
            icon = coloring?.createContrastStateDrawable(contentColor, backgroundActiveColor, true, icon)
        }
        icon?.callback = this
        icon?.alpha = ALPHA_DIM
    }

    private fun setupSizes() {
        checkNotNull(numberPaint, { "Number paint cannot be null here" })
        checkNotNull(textPaint, { "Text paint cannot be null here" })

        if (width == 0 || height == 0) {
            numberBounds = RectF()
            textBounds = RectF()
        } else {
            // set text sizes based on view size
            numberPaint.textSize = height.toFloat() * NUMBER_SIZE_RATIO
            textPaint.textSize = height.toFloat() * TEXT_SIZE_RATIO

            // find exact string sizes
            val numberWidth = numberPaint.measureText(number)
            val numberFont = numberPaint.fontMetrics
            val numberHeight = numberFont.descent - numberFont.ascent

            val textWidth = textPaint.measureText(text)
            val textFont = textPaint.fontMetrics
            val textHeight = textFont.descent - textFont.ascent

            // number is centered
            numberBounds = RectF(width / 2f - numberWidth / 2f, height / 2f - numberHeight / 2f,
                    width / 2f + numberWidth / 2f, height / 2f + numberHeight / 2f)
            // text's position below the number, offset it by its own height
            textBounds = RectF(width / 2f - textWidth / 2f, numberBounds.centerY() + textHeight / 2f,
                    width / 2f + textWidth / 2f, numberBounds.centerY() + textHeight / 2f + textHeight)

            icon?.let {
                // offset a bit to top to accommodate for misaligned icons due to text baseline
                val offset = textBounds.height().toInt() / 4
                it.bounds = Rect(textBounds.left.toInt(), textBounds.top.toInt() - offset,
                        textBounds.right.toInt(), textBounds.bottom.toInt() - offset)
            }
        }
    }

    private fun setupPaints() {
        contentColor = if (isInEditMode) Color.BLUE else coloring!!.decodeColor(storage?.getStr(ESetting.ColorPhoneText))

        numberPaint = TextPaint(Paint.ANTI_ALIAS_FLAG or Paint.LINEAR_TEXT_FLAG)
        with(numberPaint) {
            textAlign = Paint.Align.CENTER
            typeface = Typeface.createFromAsset(context.assets, THIN_FONT_PATH)
            color = contentColor
            alpha = ALPHA_FULL
        }

        textPaint = TextPaint(Paint.ANTI_ALIAS_FLAG or Paint.LINEAR_TEXT_FLAG)
        with(textPaint) {
            textAlign = Paint.Align.CENTER
            color = contentColor
            alpha = ALPHA_DIM
        }

        setupSizes()
    }

    fun setContentColor(@ColorInt color: Int) {
        contentColor = color
        setupPaints()
        setupIcon()
        invalidate()
    }

    fun setHideText(hide: Boolean) {
        hideText = hide
        invalidate()
    }

    fun setIcon(drawable: Drawable?) {
        icon = drawable
        setupIcon()
        invalidate()
    }

    override fun onDraw(canvas: Canvas) {
        super.onDraw(canvas)
        canvas.drawText(number, numberBounds.centerX(), numberBounds.centerY(), numberPaint)
        if (!hideText) {
            canvas.drawText(text, textBounds.centerX(), textBounds.centerY(), textPaint)
        }
        icon?.draw(canvas)
    }

}
