package com.bria.common.customelements.internal.views.buttons

import android.annotation.SuppressLint
import android.content.Context
import android.content.res.TypedArray
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.Rect
import android.graphics.drawable.ColorDrawable
import android.graphics.drawable.Drawable
import android.support.annotation.ColorInt
import android.support.annotation.StyleableRes
import android.support.v4.content.ContextCompat
import android.util.AttributeSet
import android.view.View
import android.view.View.OnClickListener
import android.widget.CheckBox
import com.bria.common.R
import com.bria.common.controller.settings.ESetting
import com.bria.common.modules.BriaGraph
import com.bria.common.uiframework.coloring.Coloring
import com.bria.common.util.Log
import java.util.*

class TwoStateImageView : View {

    private companion object {
        private val TAG = "TwoStateImageView"
    }

    private var clickListener: View.OnClickListener? = null
    private var mainIconIdle: Drawable? = null
    private var mainIconActive: Drawable? = null
    private var currentIcon: Drawable? = null
    private var overlayIcon: Drawable? = null
    private var drawableBounds: Rect = Rect()
    private var innerPadding: Int = 0
    private var shouldColor: Boolean = false
    private var shouldContrastColor: Boolean = false
    private var shouldOverlayIdleIcon: Boolean = false

    @ColorInt
    private var backColor: Int = Color.TRANSPARENT
    @ColorInt
    private var iconColorIdle: Int = Color.LTGRAY
    @ColorInt
    private var iconColorActive: Int = Color.GREEN

    private val mClickDelegate = OnClickListener {
        isActive = !isActive
        clickListener?.onClick(this@TwoStateImageView)
    }

    /**
     * Activates or deactivates the View. This doesn't set the [CheckBox.isChecked] attribute, so test this View's state with [isActive].
     */
    private var _isActive: Boolean = false
    var isActive: Boolean
        get() = _isActive
        set(active) {
            _isActive = active
            updateCurrentIcon()
        }


    // <editor-fold desc="Constructors">
    private constructor(context: Context) : super(context) {
        init(context, null, 0, 0)
    }

    constructor(context: Context, attrs: AttributeSet?) : super(context, attrs) {
        init(context, attrs, 0, 0)
    }

    constructor(context: Context, attrs: AttributeSet?, defStyleAttr: Int) : super(context, attrs, defStyleAttr) {
        init(context, attrs, defStyleAttr, 0)
    }
    // </editor-fold>

    /**
     * Initializes the View (all settings, colors, visibilities, etc).
     */
    private fun init(context: Context, attrs: AttributeSet?, defStyleAttr: Int, defStyleRes: Int) {
        val attributes: TypedArray = context.obtainStyledAttributes(attrs, R.styleable.TwoStateImageView, defStyleAttr, defStyleRes)

        mainIconIdle = attributes.getDrawable(R.styleable.TwoStateImageView_main_icon_idle)
        mainIconActive = attributes.getDrawable(R.styleable.TwoStateImageView_main_icon_active)
        overlayIcon = attributes.getDrawable(R.styleable.TwoStateImageView_overlay_icon)
        shouldColor = attributes.getBoolean(R.styleable.TwoStateImageView_coloring, true)
        shouldContrastColor = attributes.getBoolean(R.styleable.TwoStateImageView_contrast_coloring, false)
        shouldOverlayIdleIcon = attributes.getBoolean(R.styleable.TwoStateImageView_overlay_idle_icon, false)
        innerPadding = attributes.getDimensionPixelSize(R.styleable.TwoStateImageView_inner_padding, 0)

        if (!isInEditMode && (mainIconIdle == null || mainIconActive == null)) {
            throw RuntimeException("Cannot use null-values for main icon states!")
        }

        // setup colors and drawables
        if (isInEditMode) {
            if (background == null) background = ColorDrawable(dimColor(backColor))
        } else {
            updateIconColors(attributes)
        }
        attributes.recycle()

        // setup click delegate
        isClickable = true
        super.setOnClickListener(mClickDelegate)

        // finally update the icon
        currentIcon = (if (isActive) mainIconActive else mainIconIdle)
    }

    override fun setOnClickListener(l: View.OnClickListener?) {
        clickListener = l
    }

    private fun updateIconColors(attributes: TypedArray? = null) {
        val coloring = Coloring.get()
        val settings = BriaGraph.settings

        backColor = coloring.decodeColor(settings.getStr(ESetting.ColorBrandDefault))
        val idleSetting = attributes?.let { findSetting(it, R.styleable.TwoStateImageView_icon_color_name_idle, ESetting.ColorWhite) }
        iconColorIdle = idleSetting?.let { coloring.decodeColor(settings.getStr(it)) } ?: Color.LTGRAY
        val activeSetting = attributes?.let { findSetting(it, R.styleable.TwoStateImageView_icon_color_name_active, ESetting.ColorSelection) }
        iconColorActive = activeSetting?.let { coloring.decodeColor(settings.getStr(it)) } ?: Color.GREEN

        if (background == null) background = coloring.createBackgroundDrawable(Color.TRANSPARENT, backColor, Color.TRANSPARENT, true)
        if (shouldColor) {
            if (shouldContrastColor) {
                iconColorIdle = coloring.getContrastColor(iconColorIdle)
            }
            mainIconIdle = coloring.colorDrawable(mainIconIdle, iconColorIdle)
            mainIconIdle!!.callback = this
            mainIconActive = coloring.colorDrawable(mainIconActive, iconColorActive)
            mainIconActive!!.callback = this
        }
        if (overlayIcon != null) {
            overlayIcon!!.callback = this
        }
    }

    fun setActiveIcon(icon: Int) {
        mainIconIdle =  ContextCompat.getDrawable(context, icon)
        updateIconColors()
        updateCurrentIcon()
    }

    fun setIdleIcon(icon: Int) {
        mainIconActive = ContextCompat.getDrawable(context, icon)
        updateIconColors()
        updateCurrentIcon()
    }

    fun setOverlayIcon(icon: Drawable?) {
        overlayIcon = icon
        if (overlayIcon != null) {
            overlayIcon!!.callback = this
        }
        invalidate()
    }

    private fun updateCurrentIcon() {
        currentIcon = if (isActive) mainIconActive else mainIconIdle
        currentIcon?.bounds = Rect(drawableBounds)
        invalidate()
    }

    /**
     * Reads and matches the setting from the given [StyleableRes] attribute.
     * @param attributes Which attributes instance to use
     * @param id Which styleable attribute to read from
     * @param defaultValue If something fails, method returns this value
     * @return A color [ESetting] matched by the given `id`
     */
    private fun findSetting(attributes: TypedArray, @StyleableRes id: Int, defaultValue: ESetting): ESetting {
        try {
            val defValueIndex = Arrays.asList(*AllowedColors.SETTINGS).indexOf(defaultValue)
            val enumConstant = attributes.getInteger(id, defValueIndex)
            if (enumConstant >= 0 && enumConstant < AllowedColors.SETTINGS.size) {
                return AllowedColors.SETTINGS[enumConstant]
            }
        } catch (e: Exception) {
            Log.w(TAG, "Failed to load color: ${e.message}", e)
            throw RuntimeException("Color name must be a valid color index from AllowedColors array!", e)
        }
        return ESetting.ColorLtGray
    }

    /**
     * Dims the given color by reducing the alpha value on it.
     * @param color Which color to dim
     * @return A new, dimmed Color value
     */
    private fun dimColor(color: Int): Int {
        if (color == Color.TRANSPARENT) return color

        val dimLevel = 0x90
        val r = Color.red(color)
        val g = Color.green(color)
        val b = Color.blue(color)
        return Color.argb(dimLevel, r, g, b)
    }

    override fun onSizeChanged(w: Int, h: Int, oldW: Int, oldH: Int) {
        super.onSizeChanged(w, h, oldW, oldH)
        drawableBounds.left = innerPadding + paddingStart
        drawableBounds.top = innerPadding + paddingTop
        drawableBounds.right = w - innerPadding - paddingEnd
        drawableBounds.bottom = h - innerPadding - paddingBottom
        overlayIcon?.bounds = Rect(drawableBounds)
        currentIcon?.bounds = Rect(drawableBounds)
    }

    override fun onDraw(canvas: Canvas) {
        super.onDraw(canvas) // this is only for the background
        currentIcon?.draw(canvas)
        if ((!isActive && shouldOverlayIdleIcon) || (isActive && !shouldOverlayIdleIcon)) {
            overlayIcon?.draw(canvas)
        }
    }

    @SuppressLint("GetContentDescriptionOverride")
    override fun getContentDescription(): CharSequence? = super.getContentDescription().toString() + if (isActive) ":on" else ":off"

    // This is causing crash - BRACE-10722
//    public override fun onSaveInstanceState(): Parcelable? {
//        val originalState = super.onSaveInstanceState()
//        val state = TwoStateImageViewState(originalState)
//        state.isActive = isActive
//        return state
//    }
//
//    public override fun onRestoreInstanceState(savedState: Parcelable?) {
//        if (savedState !is TwoStateImageViewState) {
//            super.onRestoreInstanceState(savedState)
//            return
//        }
//        super.onRestoreInstanceState(savedState.superState)
//        isActive = savedState.isActive
//    }

}
