package com.bria.common.modules

import android.annotation.SuppressLint
import android.app.Application
import android.support.v4.app.NotificationManagerCompat
import com.bria.common.ProcessLifecycleTracker
import com.bria.common.controller.ClientConfig
import com.bria.common.controller.ControllerObservable
import com.bria.common.controller.accounts.IAccountsCtrlActions
import com.bria.common.controller.accounts.IAccountsCtrlObserver
import com.bria.common.controller.commlog.ICommLogCtrlEvents
import com.bria.common.controller.commlog.ICommLogCtrlObserver
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents
import com.bria.common.controller.contacts.buddy.IBuddyCtrlObserver
import com.bria.common.controller.contacts.local.IContactsCtrlEvents
import com.bria.common.controller.license.ILicenseCtrlActions
import com.bria.common.controller.license.LicenseUtil
import com.bria.common.controller.phone.GoodCallQualityProvenObservable
import com.bria.common.controller.presence.IPresenceCtrlEvents
import com.bria.common.controller.presence.PresenceDataProvider
import com.bria.common.controller.presence.PresenceStatuses
import com.bria.common.controller.presence.refactoring.EPresenceStatus
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.crashlytics.CrashlyticsImpl
import com.bria.common.crashlytics.CrashlyticsStub
import com.bria.common.crashlytics.ICrashlytics
import com.bria.common.modules.android.AndroidSoundInfo
import com.bria.common.modules.android.AppLifetimeReceivers
import com.bria.common.modules.android.BackgroundForegroundScreenOffReceiver
import com.bria.common.modules.store.GoodReviewDriver
import com.bria.common.network.ConnectivityObservable
import com.bria.common.network.NetworkModule
import com.bria.common.notification.NotificationsHandler
import com.bria.common.teams.TeamsModule
import com.bria.common.uiframework.activities.Orion
import com.bria.common.uiframework.helpers.BriaShortcutManager
import com.bria.common.uireusable.dataprovider.ISimpleDataProvider
import com.bria.common.util.CpcCallingSchemesPatternFactory
import com.bria.common.util.IObservable
import com.bria.common.util.Utils
import java.util.*
import java.util.regex.Pattern

@SuppressLint("NewApi")
/*
* Why not Koin?
* Due to KoinContext.resolveInstance method being synchronized, creation of services is serialized.
* At this point in time this would work against us since we initialize on several threads during startup.
*
* Why not Dagger?
* Due to code generation, regular compile errors are obfuscated. Also, the build is prolonged.
*
* Use fun to express transient objects and val by lazy to define singletons.
* */
object BriaGraph {
    private val LOG_TAG = "BriaGraph"

    lateinit var application: Application

    fun initialize(application: Application) {
        this.application = application
    }

    private val stack: Deque<() -> Unit> = ArrayDeque()

    private fun <T> T.registerForShutdown(callback: (obj: T) -> Unit): T {

        val noArgCallback: () -> Unit = {
            callback(this)
        }

        stack.push(noArgCallback)

        return this
    }

    fun shutdown() {
        while (stack.isNotEmpty()) {
            val callback = stack.pop()
            try {
                callback()
            } catch (e: Throwable) {
                // shutting down, don't block the process
                Utils.crashInDebug(LOG_TAG, e)
            }
        }
    }

    val settings: Settings by lazy {
        Settings.get(application)
    }

    val androidSoundInfo: AndroidSoundInfo by lazy {
        AndroidSoundInfo(application)
    }

    val appLifetimeReceivers: AppLifetimeReceivers by lazy {
        AppLifetimeReceivers(application, BackgroundForegroundScreenOffReceiver())
    }

    val processLifecycleTracker: ProcessLifecycleTracker by lazy {
        ProcessLifecycleTracker(application)
    }

    val networkModule: NetworkModule by lazy {
        NetworkModule(application, settings, ConnectivityObservable.getObservable())
                .registerForShutdown { it.destroy() }
    }

    val shortcutManager: BriaShortcutManager by lazy {
        BriaShortcutManager(application, callLogObservable, callLogActions, settings, contactCtrl)
                .registerForShutdown { it.destroy() }
    }

    val orion: Orion by lazy {
        Orion(application)
                .registerForShutdown { it.destroy() }
    }

    val goodReviewDriver: GoodReviewDriver by lazy {
        GoodReviewDriver(
                settings,
                Utils.Build.getFullVersion(application),
                GoodCallQualityProvenObservable.getObservable(),
                ClientConfig.get().isDebugMode(),
                LicenseUtil.getAppBaseLicenseType(application))
                .registerForShutdown { it.shutdown() }
    }

    val cpcCallingSchemesPattern: Pattern by lazy {
        CpcCallingSchemesPatternFactory.compileSchemesPattern(application)
    }

    val notificationManager: NotificationManagerCompat by lazy {
        NotificationManagerCompat.from(application)
    }

    val notificationsHandler: NotificationsHandler by lazy {
        NotificationsHandler(application, ControllerObservable.getObservable(), notificationManager)
                .registerForShutdown { it.destroy() }
    }

    val presenceStatuses: PresenceStatuses by lazy {
        PresenceStatuses(settings, accountsController, accountsObservable)
                .registerForShutdown { it.shutdown() }
    }

    val presenceDataProvider: ISimpleDataProvider<EPresenceStatus> by lazy {
        PresenceDataProvider(presenceStatuses)
    }

    val accountsController: IAccountsCtrlActions by lazy {
        ControllerObservable.getObservable().blockingFirst().accountsCtrl.events
    }

    val accountsObservable: IObservable<IAccountsCtrlObserver> by lazy {
        ControllerObservable.getObservable().blockingFirst().accountsCtrl.observable
    }

    val licenseController: ILicenseCtrlActions by lazy {
        ControllerObservable.getObservable().blockingFirst().licenseCtrl.events
    }

    val presenceController: IPresenceCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().presenceCtrl.events
    }

    val callLogActions: ICommLogCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().localCommLogCtrl.events
    }

    val contactCtrl: IContactsCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().contactsCtrl.events
    }

    val buddyCtrl: IBuddyCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().buddyCtrl.events
    }

    val buddyObservable: IObservable<IBuddyCtrlObserver> by lazy {
        ControllerObservable.getObservable().blockingFirst().buddyCtrl.observable
    }

    val callLogObservable: IObservable<ICommLogCtrlObserver> by lazy {
        if (settings.getBool(ESetting.BroadWorksEnabled) && settings.getBool(ESetting.FeatureBroadWorks)) {
            ControllerObservable.getObservable().blockingFirst().getBroadWorksCommLogCtrl().getObservable();
        } else {
            ControllerObservable.getObservable().blockingFirst().localCommLogCtrl.observable
        }
    }

    val crashlytics: ICrashlytics by lazy {
        // factory
        if (settings.getBool(ESetting.FeatureCrashlytics)) {
            if (!ClientConfig.get().isDebugMode) {
                return@lazy CrashlyticsImpl(application)
            }
        }

        return@lazy CrashlyticsStub()
    }

    val teams: TeamsModule by lazy {
        TeamsModule(accountsController)
    }

    // testing

    class A {
        var shutdownCallback: (() -> Unit)? = null

        fun shutdown() {
            shutdownCallback?.invoke()
        }
    }

    class BDependsOnA(val a: A) {
        var shutdownCallback: (() -> Unit)? = null

        fun shutdown() {
            shutdownCallback?.invoke()
        }
    }

    val a: A by lazy {
        A().registerForShutdown { it.shutdown() }
    }

    val bDependsOnA: BDependsOnA by lazy {
        BDependsOnA(a).registerForShutdown { it.shutdown() }
    }
}
