package com.bria.common.modules.android

import android.app.NotificationManager
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.os.Build
import com.bria.common.controller.Controller
import com.bria.common.network.ConnectivityReceiver
import com.bria.common.util.Log
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable

/**
 * A simple holder for all Android-bound components.
 */
class AndroidIntegrationComponents(private var context: Context?, controller: Observable<Controller>?) {

    private companion object {
        private val TAG = "AndroidIntegrationComponents"
    }

    private var callActionReceiver: BrandableCallActionReceiver? = BrandableCallActionReceiver()
    private var doNotDisturbReceiver: DoNotDisturbReceiver? = null
    private var screenOnOffReceiver: ScreenOnOffReceiver? = ScreenOnOffReceiver()
    var connectivityReceiver: ConnectivityReceiver? = ConnectivityReceiver()


    private var controllerDisposable: Disposable? = null

    init {
        // register call action receiver, no additional arguments are required
        context?.registerReceiver(callActionReceiver, IntentFilter("${context?.packageName}.action.CALL"))

        // register screen off receiver
        val screenOnOffFilter = IntentFilter()
        screenOnOffFilter.addAction(Intent.ACTION_SCREEN_ON)
        screenOnOffFilter.addAction(Intent.ACTION_SCREEN_OFF)
        context?.registerReceiver(screenOnOffReceiver, screenOnOffFilter)

        context?.registerReceiver(connectivityReceiver, ConnectivityReceiver.getIntentFilter())

        // wait for controllers
        controllerDisposable = controller
            ?.observeOn(AndroidSchedulers.mainThread())
            ?.subscribe({
                // create DND receiver only on M
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                    doNotDisturbReceiver = DoNotDisturbReceiver(it)
                    context?.registerReceiver(doNotDisturbReceiver, IntentFilter(NotificationManager.ACTION_INTERRUPTION_FILTER_CHANGED))
                }
                controllerDisposable?.dispose()
            }, {
                Log.e(TAG, "Failed to load Controllers", it)
                controllerDisposable?.dispose()
            })
    }

    fun destroy() {
        callActionReceiver?.let { context?.unregisterReceiver(it) }
        callActionReceiver = null

        doNotDisturbReceiver?.let { context?.unregisterReceiver(it) }
        doNotDisturbReceiver = null

        screenOnOffReceiver?.let { context?.unregisterReceiver(it) }
        screenOnOffReceiver = null

        connectivityReceiver?.let {
            context?.unregisterReceiver(it)
        }
        connectivityReceiver = null

        controllerDisposable?.dispose()
        controllerDisposable = null
        context = null
    }

}
