package com.bria.common.modules.framework

import android.support.annotation.IntRange
import com.bria.common.modules.annotations.LifecycleState
import com.bria.common.modules.annotations.LifecycleStep
import com.bria.common.modules.framework.ComponentLifecycle.publishComponentState
import com.bria.common.modules.framework.ComponentLifecycle.publishComponentStep
import com.bria.common.modules.framework.ComponentLifecycle.track
import com.bria.common.util.Preconditions
import com.bria.common.util.reactive.CustomSchedulers
import io.reactivex.subjects.BehaviorSubject

/**
 * A reactive state tracker for any component. Add new components to the tracker and notify state changes by calling [publishComponentState].
 * For incremental creation and destruction, call [publishComponentStep]. To listen for changes, call [track].
 *
 * The full graph looks like this:
 * ```
 *                              +-----------------------------------------+
 *                              |                 +-------+               |
 *                              |                 |       |               |
 *                              |                 |       |               |
 *                              |                 |       |               |
 * +------------+       +---------------+     +---+-------v----+   +------v--------+
 * |            |       |               |     |                |   |               |
 * |    NONE    +------->   CREATING    +----->  CREATE STEP   +--->    CREATED    |
 * |            |       |               |     |                |   |               |
 * +------------+       +---------------+     +----------------+   +---------------+
 *                                                                         |
 *                                                                         |
 *                      +---------------+     +----------------+   +-------v-------+
 *                      |               |     |                |   |               |
 *                      |   DESTROYED   <-----+  DESTROY STEP  <---+  DESTROYING   |
 *                      |               |     |                |   |               |
 *                      +-------^-------+     +---^-------+----+   +---------------+
 *                              |                 |       |                |
 *                              |                 |       |                |
 *                              |                 |       |                |
 *                              |                 +-------+                |
 *                              +------------------------------------------+
 * ```
 */
object ComponentLifecycle {

    private val TAG = "ComponentLifecycle"

    const val STATE_NONE = 0
    const val STATE_CREATING = 1
    const val STATE_CREATE_STEP = 2
    const val STATE_CREATED = 3
    const val STATE_DESTROYING = 4
    const val STATE_DESTROY_STEP = 5
    const val STATE_DESTROYED = 6

    @LifecycleState
    private val NORMAL_STATES = arrayOf(STATE_NONE, STATE_CREATING, STATE_CREATED, STATE_DESTROYING, STATE_DESTROYED)
    @LifecycleStep
    private val STEP_STATES = arrayOf(STATE_CREATE_STEP, STATE_DESTROY_STEP)

    private val TRANSITIONS = mapOf(
            STATE_NONE to arrayOf(STATE_CREATING),
            STATE_CREATING to arrayOf(STATE_CREATE_STEP, STATE_CREATED),
            STATE_CREATE_STEP to arrayOf(STATE_CREATE_STEP, STATE_CREATED),
            STATE_CREATED to arrayOf(STATE_DESTROYING),
            STATE_DESTROYING to arrayOf(STATE_DESTROY_STEP, STATE_DESTROYED),
            STATE_DESTROY_STEP to arrayOf(STATE_DESTROY_STEP, STATE_DESTROYED),
            STATE_DESTROYED to arrayOf()
    )

    /**
     * Provides basic information about the lifecycle of a component you are tracking.
     * @param component ID of the component on which the event happened. IDs are usually added by calling [publishComponentState]
     * @param state Which state is currently active. This will be one of the [LifecycleState] constants
     */
    open class LifecycleStateEvent(open val component: Int, @LifecycleState open val state: Int) {
        override fun toString(): String = "'$component' Lifecycle: ${stateName(state).toLowerCase()}"
    }

    /**
     * Provides information about the incremental initialization or destruction step within a component you are tracking.
     * @param component ID of the component on which the event happened. IDs are usually added by calling [publishComponentState]
     * @param state Which state is currently active. This will be one of the [LifecycleStep] constants
     * @param step Which incremental step is the component currently in
     * @param totalSteps Component-specific maximum number of steps
     */
    class LifecycleStepEvent(override val component: Int, @LifecycleStep override val state: Int,
                             @IntRange(from = 1) val step: Int, @IntRange(from = 1) val totalSteps: Int) : LifecycleStateEvent(component, state) {
        override fun toString(): String = "'$component' Lifecycle: ${stateName(state).toLowerCase()}, step $step/$totalSteps"
    }

    /**
     * A map of all components and their states.
     */
    private val componentStateMap: MutableMap<Int, BehaviorSubject<LifecycleStateEvent>> = mutableMapOf()

    /**
     * Looks for (and possibly creates a new) Subject for the [componentStateMap] mapped to the given component ID.
     * @param component ID of the component to look for
     * @param throwIfNotFound Whether to throw a new Throwable if the Subject is not found
     * @return A non-`null` Subject instance containing either the last published state, or [STATE_NONE] by default
     */
    @Synchronized
    private fun findSubject(component: Int, throwIfNotFound: Boolean): BehaviorSubject<LifecycleStateEvent> {
        var subject = componentStateMap[component]
        if (subject == null) {
            if (throwIfNotFound) {
                throw IllegalStateException("Could not find component with ID '$component'. Make sure you published at least one state first")
            }

            val noneStateEvent = LifecycleStateEvent(component, STATE_NONE)
            subject = BehaviorSubject.createDefault(noneStateEvent)
            componentStateMap.put(component, subject)
        }

        return subject!!
    }

    /**
     * Returns a new synthetic event for the given component and state.
     * This event will structurally be identical to a real event, but does not come from a published space.
     * @param component ID of the component which needs a synthetic event
     * @param state Which state does the synthetic event need to hold
     */
    private fun synthetic(component: Int, @LifecycleState state: Int) = LifecycleStateEvent(component, state)

    // ------------------- PUBLIC API ------------------- //

    /**
     * Publishes a new normal state for the given component ID. States must be published in order, starting from [STATE_CREATING]
     * @param component ID of the component for which you are publishing the current state
     * @param state The new state. Make sure you never publish [STATE_NONE] as it's handled internally
     */
    @Synchronized
    fun publishComponentState(component: Int, @LifecycleState state: Int) {
        // validate state argument
        var error = "State '${stateName(state)}' is not valid. Check possible states in $TAG, or use 'publish component step' function for incremental steps"
        Preconditions.checkArgument(NORMAL_STATES.contains(state), error)

        // find the current state
        val componentStates: BehaviorSubject<LifecycleStateEvent> = findSubject(component, false)
        val currentState = componentStates.value!!.state

        // check if next state is valid
        error = "Component '$component' cannot transition from state '${stateName(currentState)}' to '${stateName(state)}'"
        Preconditions.checkState(TRANSITIONS[currentState]!!.contains(state), error)

        // everything is valid now, publish the new state
        componentStates.onNext(LifecycleStateEvent(component, state))
    }

    /**
     * Publishes a new step state. You can only publish step states using this function.
     * @param component ID of the component for which you are publishing the current state
     * @param state The new state. Make sure you never publish [STATE_NONE] as it's handled internally
     * @param step The current **incremental** step in the current creation or destruction routine of the given component
     * @param totalSteps The **total** number of steps in the current creation or destruction routine of the given component
     */
    @Synchronized
    fun publishComponentStep(component: Int, @LifecycleStep state: Int, @IntRange(from = 1) step: Int, @IntRange(from = 1) totalSteps: Int) {
        // validate state argument
        var error = "Step '${stateName(state)}' is not valid. Check possible step states in $TAG, or use 'publish component state' function for normal states"
        Preconditions.checkArgument(STEP_STATES.contains(state), error)

        // find the current state
        val componentStates: BehaviorSubject<LifecycleStateEvent> = findSubject(component, true)
        val currentStateInfo = componentStates.value
        val currentState = currentStateInfo!!.state

        // check if next state is valid
        error = "Component '$component' cannot transition from state '${stateName(currentState)}' to '${stateName(state)}'"
        Preconditions.checkState(TRANSITIONS[currentState]!!.contains(state), error)

        // check if next step is valid (it is valid by default if previous state was not a step state)
        Preconditions.checkArgument(totalSteps > 0, "'Total steps' must be a positive integer'")
        Preconditions.checkArgument(step > 0, "'Step' must be a positive integer")
        Preconditions.checkArgument(step <= totalSteps, "'Step' must be less than or equal to 'total steps'")
        if (STEP_STATES.contains(currentState)) {
            val currentStep = (currentStateInfo as LifecycleStepEvent).step
            Preconditions.checkArgument(currentStep < step, "Step '$step' is not valid after step '$currentStep'")
            val currentTotalSteps = currentStateInfo.totalSteps
            Preconditions.checkArgument(totalSteps == currentTotalSteps, "'Total steps' should not change between states")
        }

        // everything is valid now, publish the new state
        componentStates.onNext(LifecycleStepEvent(component, state, step, totalSteps))
    }

    /**
     * Gets the tracker observable for the requested component.
     * Note that the returned observable is automatically being observed on the [CustomSchedulers.lifecycles] scheduler.
     * @param component ID of the component for which you wish to track state
     * @return An observable that is being observed on the [CustomSchedulers.lifecycles] scheduler
     */
    @Synchronized
    @Suppress("unused")
    fun trackInBackground(component: Int) = findSubject(component, false).observeOn(CustomSchedulers.lifecycles())!!

    /**
     * Gets the tracker observable for the requested component.
     * Note that the returned observable doesn't have a default scheduler to be observed on.
     * @param component ID of the component for which you wish to track state
     * @return An observable that you can immediately subscribe to
     */
    @Synchronized
    fun track(component: Int) = findSubject(component, false)

    /**
     * Gets the last published state for the requested component.
     * Note that this call might be blocking if no state was emitted previously.
     * @param component ID of the component for which you wish to read the last state
     * @return The last published state identifier, one of [LifecycleState] constants
     */
    @Synchronized
    @LifecycleState
    fun getLastState(component: Int) = track(component).blockingLatest().iterator().next().state

    /**
     * Clears the tracker observable for the given component. If there was a tracker in place, its observable will get an 'on complete' event.
     * @param component ID of the component for which you wish to clear state
     * @return `True` if state tracking has stopped, `false` if there was no tracker for the given component
     */
    @Synchronized
    fun clear(component: Int): Boolean = componentStateMap.remove(component)?.let { it.onComplete(); it } != null

    /**
     * Clears all trackers. Each cleared tracker observable will get an 'on complete' event.
     */
    @Synchronized
    fun clearAll() {
        componentStateMap.values.forEach { it.onComplete() }
        componentStateMap.clear()
    }

    /**
     * Gets the state name based on the given state ID.
     * @param state Which state to convert to string
     * @return Either a textual representation of the state's name, or a fatal exception if invalid state is given
     */
    @JvmStatic
    @Synchronized
    fun stateName(@LifecycleState state: Int): String = when (state) {
        ComponentLifecycle.STATE_NONE -> "NONE"
        ComponentLifecycle.STATE_CREATING -> "CREATING"
        ComponentLifecycle.STATE_CREATE_STEP -> "CREATE_STEP"
        ComponentLifecycle.STATE_CREATED -> "CREATED"
        ComponentLifecycle.STATE_DESTROYING -> "DESTROYING"
        ComponentLifecycle.STATE_DESTROY_STEP -> "DESTROY_STEP"
        ComponentLifecycle.STATE_DESTROYED -> "DESTROYED"
        else -> throw IllegalArgumentException("Unknown state '$state'")
    }

}
