package com.bria.common.modules.store

import com.bria.common.controller.license.EBaseLicenseType
import com.bria.common.controller.phone.GoodCallQualityProvenObservable
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.ISettings
import com.bria.common.util.Log
import com.bria.common.util.Utils
import io.reactivex.Observable
import io.reactivex.disposables.Disposable
import io.reactivex.subjects.PublishSubject

/*
* Drives good reviews on play store.
* */
class GoodReviewDriver(
        private val settings: ISettings<ESetting>,
        private val appVersion: String,
        private val callQualityObservable: Observable<GoodCallQualityProvenObservable.CallQuality>,
        private val isDebugMode: Boolean,
        private val appBaseLicenseType: EBaseLicenseType) {

    private val LOG_TAG = "GoodReviewDriver"

    private var disposable: Disposable? = null

    fun initialize() {
        val playStoreBuild = isDebugMode || appBaseLicenseType == EBaseLicenseType.eAndroidMarket
        if (!playStoreBuild) {
            Log.d(LOG_TAG, "Not a play store build, no need to subscribe.")

            return
        }

        disposable = callQualityObservable.subscribe(
                { callQuality -> handleCallQualityEmission(callQuality) },
                { throwable -> Utils.crashInDebug(LOG_TAG, throwable) }
        )
    }

    fun shutdown() {
        disposable?.dispose()
    }

    private var currentNumberOfQualityProvenCalls: Int
        get() = settings.getInt(ESetting.AppRatingCallCounter)
        set(value) = settings.set(ESetting.AppRatingCallCounter, value)

    private val GOOD_CALLS_THRESHOLD = 5

    private fun handleCallQualityEmission(callQuality: GoodCallQualityProvenObservable.CallQuality) {
        Log.d(LOG_TAG, "Call quality is $callQuality")

        if (wasAppRatedByUser()) {
            Log.d(LOG_TAG, "Current version of the application is already rated.")

            return
        }

        Log.d(LOG_TAG, "Current number of quality proven calls is: $currentNumberOfQualityProvenCalls")

        if (callQuality != GoodCallQualityProvenObservable.CallQuality.Proven) {
            currentNumberOfQualityProvenCalls = 0
        } else {
            currentNumberOfQualityProvenCalls++
        }

        if (currentNumberOfQualityProvenCalls >= GOOD_CALLS_THRESHOLD) {
            currentNumberOfQualityProvenCalls = 0
            mPublishSubject.onNext(Any())
        }
    }

    fun wasAppRatedByUser(): Boolean {
        val ratedVersion = settings.getStr(ESetting.AppVersionRated)

        val result = appVersion == ratedVersion

        return result
    }

    private fun markAsRated() {
        settings.set(ESetting.AppVersionRated, appVersion)
    }

    private val mPublishSubject = PublishSubject.create<Any>()

    fun getShowAppRatingDialogObservable(): Observable<Any> {
        return mPublishSubject
    }

    fun handleUserAction(userAction: UserAction) {
        Log.d(LOG_TAG, "User action: $userAction")

        when (userAction) {
            GoodReviewDriver.UserAction.NotEnjoying -> {
                // don't ask user again in this version
                markAsRated()
            }
            GoodReviewDriver.UserAction.Rated -> {
                // don't ask user again in this version
                markAsRated()

                Log.i(LOG_TAG, "User decided to rate this application (version ${appVersion})")
            }
            GoodReviewDriver.UserAction.DoesNotWantToRate -> {
                // don't ask user again in this version
                markAsRated()
            }
            GoodReviewDriver.UserAction.WantsToBeReminded -> {
                // no markAsRated() call here, ask user again at later time
            }
        }
    }

    enum class UserAction {
        NotEnjoying,
        Rated,
        DoesNotWantToRate,
        WantsToBeReminded
    }
}
