package com.bria.common.permission

import android.Manifest
import android.app.Dialog
import android.content.Context
import android.os.Bundle
import android.support.v4.app.ActivityCompat
import android.support.v7.app.AlertDialog
import com.bria.common.R
import com.bria.common.analytics.Analytics
import com.bria.common.analytics.Constants
import com.bria.common.uiframework.activities.AbstractActivity
import com.bria.common.util.Log
import com.bria.common.util.Utils

object PermissionExplanationDialogBuilder {

    const val TAG = "PermissionExplanationDialogBuilder"

    const val PERMISSIONS_TO_EXPLAIN = "PERMISSIONS_TO_EXPLAIN"
    const val PERMISSIONS_TO_REQUEST = "PERMISSIONS_TO_REQUEST"
    const val MESSAGE = "MESSAGE"
    const val REQUEST_CODE = "REQUEST_CODE"

    fun build(activity: AbstractActivity, bundle: Bundle): Dialog? {
        val permissionsToExplain = bundle.getStringArray(PERMISSIONS_TO_EXPLAIN)
        if (permissionsToExplain == null || permissionsToExplain.isEmpty()) {
            Utils.crashInDebug(TAG, "No permissions to explain.")
            return null
        }

        val permissionsToRequest = bundle.getStringArray(PERMISSIONS_TO_REQUEST)
        if (permissionsToRequest == null || permissionsToRequest.isEmpty()) {
            Utils.crashInDebug(TAG, "No permissions to request.")
            return null
        }

        val requestCode = bundle.getInt(REQUEST_CODE)

        val appName = activity.getString(R.string.app_name)
        val wantsTo = activity.getString(R.string.tWantsTo)
        val permissionToExplainString = permissionsToExplain.joinToString(", ") { getFromDictionary(activity, it) }
        val title = "$appName $wantsTo $permissionToExplainString"

        var message = bundle.getString(MESSAGE, "")
        val reconsiderPermission = activity.getString(R.string.tReconsiderPermission)
        val showNever = activity.getString(R.string.tShowNever)
        message = "$message $reconsiderPermission $showNever"

        return AlertDialog.Builder(activity)
            .setTitle(title)
            .setMessage(message)
            .setCancelable(false)
            .setPositiveButton(R.string.tOk) { _, _ ->
                Log.d(TAG, "Explanation accepted for $permissionToExplainString")
                sendAnalyticsForExplanationDialog(true, permissionsToExplain)

                ActivityCompat.requestPermissions(activity, permissionsToRequest, requestCode)
            }
            .setNegativeButton(R.string.tCancel) { _, _ ->
                Log.d(TAG, "Explanation not accepted for $permissionToExplainString")
                sendAnalyticsForExplanationDialog(false, permissionsToExplain)

                activity.explanationDialogCancelled(requestCode)
            }
            .create()
    }

    private fun getFromDictionary(context: Context, androidPermission: String): String =
            when (androidPermission) {
                Manifest.permission.WRITE_CONTACTS -> context.getString(R.string.tPermWriteContact)
                Manifest.permission.READ_CONTACTS -> context.getString(R.string.tPermReadContact)
                Manifest.permission.CAMERA -> context.getString(R.string.tPermCamera)
                Manifest.permission.WRITE_EXTERNAL_STORAGE -> context.getString(R.string.tPermUseExternal)
                Manifest.permission.READ_PHONE_STATE -> context.getString(R.string.tPermReadPhoneState)
                Manifest.permission.RECORD_AUDIO -> context.getString(R.string.tPermMicrophone)
                Manifest.permission.PROCESS_OUTGOING_CALLS -> context.getString(R.string.tPermProcessOutgoing)
                Manifest.permission.READ_EXTERNAL_STORAGE -> context.getString(R.string.tPermReadExternalStorage)
                else -> {
                    val prefix = "android.permission."
                    if (androidPermission.contains(prefix)) {
                        androidPermission.replace(prefix, "")
                    } else {
                        androidPermission
                    }
                }
            }

    private fun sendAnalyticsForExplanationDialog(accepted: Boolean, permissions: Array<String>) {
        try {
            val bundle = Bundle()
            bundle.putBoolean("accepted", accepted)
            bundle.putString("permissions", permissions.joinToString(","))
            Analytics.send(Constants.Events.PERMISSION_EXPLANATION_OUTCOME, bundle)
        } catch (ignored: Throwable) {
        }
    }
}
