package com.bria.common.uiframework.helpers

import android.annotation.SuppressLint
import android.annotation.TargetApi
import android.content.Context
import android.content.Intent
import android.content.pm.ShortcutInfo
import android.content.pm.ShortcutManager
import android.graphics.Bitmap
import android.graphics.drawable.BitmapDrawable
import android.graphics.drawable.Icon
import android.net.Uri
import android.os.Build
import android.support.annotation.RequiresApi
import com.bria.common.R
import com.bria.common.controller.Controllers
import com.bria.common.controller.commlog.CallLogDataObject
import com.bria.common.controller.commlog.ICommLogCtrlEvents
import com.bria.common.controller.commlog.ICommLogCtrlObserver
import com.bria.common.controller.contacts.local.Contact
import com.bria.common.controller.contacts.local.IContactsCtrlEvents
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.modules.ModuleClassFinder
import com.bria.common.modules.glide.GlideApp
import com.bria.common.util.AndroidUtils
import com.bria.common.util.GlobalConstants
import com.bria.common.util.IObservable
import com.bria.common.util.Utils
import io.reactivex.Maybe
import io.reactivex.Single
import io.reactivex.schedulers.Schedulers


@RequiresApi(Build.VERSION_CODES.N_MR1)
@TargetApi(Build.VERSION_CODES.M)
@SuppressLint("RxSubscribeOnError")
class BriaShortcutManager(val context: Context, val mCallLogCtrlObserver: IObservable<ICommLogCtrlObserver>, val mCallLogCtrlEvents: ICommLogCtrlEvents, val settings: Settings, val contactCtrlEvents: IContactsCtrlEvents) : ICommLogCtrlObserver, IBriaShortcutManager {

    private val TAG: String = "BriaShortcutManager"

    private var shortcutManager: ShortcutManager
    private var shortcutInfoMessage: ShortcutInfo

    init {
        shortcutManager = context.getSystemService(ShortcutManager::class.java)
        val mComposeNewImMessage = Intent(context, ModuleClassFinder.instance.mainActivityClass)
        mComposeNewImMessage.setAction(GlobalConstants.FAST_SHORTCUT_ACTION_MESSAGE)
        shortcutInfoMessage =
                ShortcutInfo.Builder(context, "composeIM")
                        .setShortLabel(context.getString(R.string.tComposeNewIM))
                        .setLongLabel(context.getString(R.string.tComposeNewIM))
                        .setIcon(Icon.createWithResource(context, R.drawable.icon_message_active))
                        .setIntent(mComposeNewImMessage)
                        .build()

        mCallLogCtrlObserver.attachObserver(this)


    }


    @SuppressLint("RxLeakedSubscription")
    override fun update() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O && settings.getBool(ESetting.FeatureShortcutManager)) {
            val callIntent = Intent(context, ModuleClassFinder.instance.mainActivityClass)
            callIntent.setAction(GlobalConstants.FAST_SHORTCUT_ACTION_CALL)


            Single.fromCallable {
                Controllers.get().callLog.allCallLogs.take(3)
            }
                    .subscribeOn(Schedulers.io())
                    .subscribe({ callLogs ->
                        shortcutManager.removeAllDynamicShortcuts()

                        shortcutManager.addDynamicShortcuts(listOf(shortcutInfoMessage))

                        shortcutManager.addDynamicShortcuts(
                                callLogs.map {
                                    val contact = getContact(it)
                                    val labelString = if (it.remoteName.isNullOrEmpty()) it.number else it.remoteName
                                    callIntent.setData(Uri.fromParts("sip", it.number, null))
                                    val icon = Icon.createWithBitmap(getBitmapFromContact(contact))
                                    ShortcutInfo.Builder(context, it.id.toString())
                                            .setShortLabel(labelString)
                                            .setLongLabel(labelString)
                                            .setIcon(icon)
                                            .setIntent(callIntent)
                                            .build()
                                }
                        )

                    }, { Utils.crashInDebug(TAG, it) });


        }
    }

    override fun onCallLogCursorRefreshed() {
        update()
    }

    override fun onLogListUpdated() {
    }


    protected fun getContact(callLogDataObject: CallLogDataObject): Contact? {
        val name = callLogDataObject.name
        val contactApi = contactCtrlEvents.getContactsApi()
        try {
            return contactApi.getContact().byId(callLogDataObject.contactId).withAllData().asObservable()
                    .onErrorResumeNext(contactApi.getContact().byNumber(getNumberWithoutDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable())
                    .onErrorResumeNext(contactApi.getContact().byNumber(getSanitizedNumberWithDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable())
                    .onErrorResumeNext(contactApi.getContact().byNumber(getNumberWithDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable())
                    .onErrorResumeNext(contactApi.getContact().byNumber(getSanitizedNumberWithoutDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable()).blockingFirst()
        } catch (excaption: Exception) {
            return null
        }
    }


    override fun onMissedCallListUpdated(missedCallCount: Int) {
    }

    override fun destroy() {
        mCallLogCtrlObserver.detachObserver(this)
    }

    protected fun getNumberWithoutDomain(callLogDataObject: CallLogDataObject): String {
        return mCallLogCtrlEvents.getNumberWithoutDomain(callLogDataObject)
    }

    protected fun getSanitizedNumberWithDomain(callLogDataObject: CallLogDataObject): String {
        return mCallLogCtrlEvents.getSanitizedNumberWithDomain(callLogDataObject)
    }

    protected fun getNumberWithDomain(callLogDataObject: CallLogDataObject): String {
        return mCallLogCtrlEvents.getNumberWithDomain(callLogDataObject)
    }

    protected fun getSanitizedNumberWithoutDomain(callLogDataObject: CallLogDataObject): String {
        return mCallLogCtrlEvents.getSanitizedNumberWithoutDomain(callLogDataObject)
    }


    fun getBitmapFromContact(contact: Contact?): Bitmap {
        val photoMaybe = Maybe.fromCallable {
            (AndroidUtils.getDrawable(context, R.drawable.default_avatar) as BitmapDrawable).bitmap
        }
        if (contact == null) {
            return photoMaybe.blockingGet();
        } else {
            return contact.getPhotoAsMaybe(context)
                    .subscribeOn(Schedulers.io())
                    .observeOn(Schedulers.computation())
                    .map({ bitmap -> GlideApp.with(context).asBitmap().load(bitmap).circleCrop().into(100, 100).get() })
                    .observeOn(Schedulers.io())
                    .switchIfEmpty(photoMaybe)
                    .blockingGet()

        }
    }


}
