package com.bria.voip.ui.main.calllog.refactoring

import android.annotation.SuppressLint
import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.media.MediaMetadataRetriever
import android.net.Uri
import android.os.Bundle
import android.provider.ContactsContract
import android.text.TextUtils
import android.text.format.DateUtils
import com.bria.common.R
import com.bria.common.analytics.BIAnalytics
import com.bria.common.controller.ClientConfig
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.DialPlanHelper
import com.bria.common.controller.calllog.CallLogApi
import com.bria.common.controller.calllog.db.CallLogEntity
import com.bria.common.controller.calllog.db.CallType
import com.bria.common.controller.collaboration.rx.impl.ParticipantManager
import com.bria.common.controller.commlog.local.LocalCallLogController
import com.bria.common.controller.contacts.genband.GenbandContactDataConversion
import com.bria.common.controller.contacts.local.ContactsApi
import com.bria.common.controller.contacts.local.data.PhoneNumber
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.license.features.EFeature
import com.bria.common.controller.phone.IPhoneCtrlEvents
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.controller.settings.branding.EGuiVisibility
import com.bria.common.controller.settings.gui.EGuiElement
import com.bria.common.sdkwrapper.CallData
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.uireusable.dataprovider.ISimpleDataProvider
import com.bria.common.util.*
import com.bria.voip.service.BriaVoipService
import com.bria.voip.ui.main.contacts.ContactEditPresenter
import com.bria.voip.ui.main.contacts.PhoneNumberActionsScreen
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import java.lang.Exception
import java.util.*
import java.util.concurrent.TimeUnit


private const val TAG = "CallLogListPresenter"
private const val MAX_CALL_TYPE_ICONS = 4

class CallLogListPresenter: AbstractPresenter(), ISimpleDataProvider<CallLogItemModel> {

    val title = getString(R.string.tCallLog)

    private lateinit var mCallLogApi: CallLogApi
    private lateinit var mContactsApi: ContactsApi

    private var mIsMetaswitchBrand : Boolean = false

    private val mGroupedList: MutableList<CallLogItemModel> = Collections.synchronizedList(mutableListOf())

    private var mExpandedPosition = -1

    private val mCompositeDisposable = CompositeDisposable()
    private val mUpdateContactInfoDisposable = CompositeDisposable()

    // Time listener is required to keep relative date time updated, see CallLogGroupItem#getFormattedNumberTypeAndData
    private val mTimeTickReceiver = object : BroadcastReceiver() {
        override fun onReceive(context: Context, intent: Intent) {
            firePresenterEvent(Events.DATA_SET_CHANGED)
        }
    }


    enum class Events : IPresenterEventTypeEnum {
        DATA_SET_CHANGED,
        DATA_SET_UPDATED,
        DATA_SET_COLLAPSED,
        DATA_SET_EXPANDED,
        SHOW_MESSAGE_LONG,
        SHOW_CALL_REMINDER,
        SHOW_DELETE_CALL_RECORD_CONFIRMATION,
        SHOW_CONFERENCE_CALL_CONFIRMATION,
        SHOW_HOSTED_CONFERENCE_CALL_CONFIRMATION,
        SHOW_WEB_VIEW,
        SHOW_PHONE_NUMBER_ACTION_PICKER,
        SHOW_CONTACT_PICKER,
        SHOW_CONTACT_EDIT,
        SHOW_CONTACT_DETAILS,
        SHOW_BUDDY_DETAILS,
        SHOW_CONVERSATION,
        // define new events here
        ;
        override fun getType(): IPresenterEventTypeEnum = this
    }

    private enum class CallAction {
        AudioCall,
        PrefixCall,
        VideoCall
    }

    // Life-cycle

    override fun onCreate() {
        super.onCreate()

        mIsMetaswitchBrand = Utils.Brands.isBrand(context, "Metaswitch")

        mCallLogApi = mControllers.callLog.api
        mCompositeDisposable.add(mCallLogApi.getAll()
                // already working on IO thread, since we do not switch threads here observation also will be on IO thread
                .subscribe(this::onCallLogDataChanged) { Log.e(TAG, "Call log error", it) })

        mContactsApi = mControllers.contacts.contactsApi
        mCompositeDisposable.add(mContactsApi.observable
                .subscribeOn(Schedulers.io())
                .throttleFirst(1, TimeUnit.SECONDS, Schedulers.io())
                .subscribe({ onContactsChanged() }, { Log.e(TAG, "Contacts error", it) }))
    }

    override fun onDestroy() {
        super.onDestroy()

        stopUpdatingContactInfo()
        clear(mCompositeDisposable)
    }

    override fun onSubscribe() {
        super.onSubscribe()
        firePresenterEvent(Events.DATA_SET_CHANGED)
        context.applicationContext.registerReceiver(mTimeTickReceiver, IntentFilter(Intent.ACTION_TIME_TICK))
    }

    override fun onUnsubscribe() {
        super.onUnsubscribe()
        context.applicationContext.unregisterReceiver(mTimeTickReceiver)
    }

    fun onNewConfig(bundle: Bundle?) {
        handleCallReminder(bundle)
    }

    // Data Provider

    override fun getItemsCount() = mGroupedList.size

    override fun getItemAt(position: Int) = mGroupedList.getOrNull(position)

    override fun clean() { }

    // PUBLIC section

    // Menu settings

    fun isMenuDeleteAllEnabled() = itemsCount != 0

    fun isMenuDeleteMissedEnabled() = itemsCount != 0

    fun isMenuDeleteConferenceEnabled() = itemsCount != 0 && mControllers.collaboration.isCollabAvailable

    fun isMenuDeleteEnabled() = itemsCount != 0

    fun isMenuCallEnabled(position: Int) = !isAnonymous(getGroupItemAt(position))

    fun isMenuAddContactEnabled(position: Int): Boolean {
        if (mControllers.broadworks.isBroadworksEnabled) {
            return false
        }

        val group = getGroupItemAt(position) ?: return false

        // todo: move genbandEnabled check to GenbandCallLogPresenter
        if (isAnonymous(group) && !mControllers.settings.getBool(ESetting.FeatureGenband)) {
            return false
        }

        return getNumberForContacts(group).isNotEmpty() && !isContactExists(group)
    }

    fun isMenuViewContactEnabled(position: Int) = getGroupItemAt(position)?.let { isContactExists(it) } ?: false

    fun isMenuUpdateContactEnabled(position: Int) = isMenuAddContactEnabled(position) /*todo && contacts.isNotEmpty*/

    fun isMenuDeleteCallRecordEnabled(position: Int) = false
            //!mControllers.broadworks.isBroadworksEnabled && //mControllers.settings.callRecordingEnabled &&
            //getGroupItemAt(position)?.let { RecordingUtils.fileExists(context, it.recordPath) } ?: false

    fun isMenuSendSmsEnabled(position: Int): Boolean {
        val group = getGroupItemAt(position) ?: return false

        if (isAnonymous(group)) {
            return false
        }

        val account = mControllers.accounts.getAccount(group.accountNickname) ?: return false

        val featureImps = mControllers.license.checkFeature(EFeature.IMPS)
        val sms = mControllers.settings.getBool(ESetting.Sms) && account.getBool(EAccountSetting.IsSMS)
        val listOfSmsApiAccounts = mControllers.accounts.getAccounts { it.type == EAccountType.SmsApi && it.getBool(EAccountSetting.IsSMS) }
        val smsApi = mControllers.settings.getBool(ESetting.FeatureSMSApi) && listOfSmsApiAccounts.size > 0

        return featureImps && (sms || smsApi)
    }

    fun isMenuSendImEnabled(position: Int): Boolean {
        if (mControllers.broadworks.isBroadworksEnabled) {
            return false
        }

        val group = getGroupItemAt(position) ?: return false

        if (isAnonymous(group)) {
            return false
        }

        if (!isLocalContactId(group.contactId)/* || getBuddy(callLogDataObject) == null*/) {
            return false
        }

        val account = mControllers.accounts.getAccount(group.accountNickname) ?: return false

        // todo: move meetMe check to GenbandCallLogPresenter
        val meetMe = group.number.equals(account.getStr(EAccountSetting.GenbandAccMeetMeConferenceSipUri), ignoreCase = true)
        val voiceMail = group.number == account.getStr(EAccountSetting.VMNumber)

        val featureImps = mControllers.license.checkFeature(EFeature.IMPS)
        val im = mControllers.settings.getBool(ESetting.ImPresence) &&
                (account.type == EAccountType.Xmpp || account.getBool(EAccountSetting.IsIMPresence))

        return featureImps && im && !meetMe && !voiceMail
    }

    //fun isMenuFileShareEnabled(position: Int) = false // todo: Do we no longer need this menu option?

    fun isMenuVideoCallEnabled(position: Int): Boolean {
        if (!isMenuCallEnabled(position)) {
            return false
        }

        val group = getGroupItemAt(position) ?: return false

        val accountNickname = group.accountNickname

        val account = (if (!accountNickname.isEmpty()) {
            mControllers.accounts.getAccount(accountNickname)
        } else {
            mControllers.accounts.primaryAccount
        }) ?: return false

        // todo: move meetMe check to GenbandCallLogPresenter
        val meetMe = group.number.equals(account.getStr(EAccountSetting.GenbandAccMeetMeConferenceSipUri), ignoreCase = true)
        val voiceMail = group.number == account.getStr(EAccountSetting.VMNumber)

        return mControllers.accounts.isVideoEnabled(account) && !account.getBool(EAccountSetting.AlwaysOfferVideo) && !meetMe && !voiceMail
    }

    fun isMenuTransferCallEnabled(position: Int): Boolean {
        val group = getGroupItemAt(position) ?: return false

        if (mControllers.collaboration.hasConferenceCall() || isConference(group)) {
            return false
        }

        if (mControllers.broadworks.isBroadworksEnabled) {
            return false
        }

        if (isAnonymous(group)) {
            return false
        }

        if (mControllers.phone.callCount != 1) {
            return false
        }

        return if (mControllers.phone.phoneState != IPhoneCtrlEvents.EPhoneState.eInCall) {
            false
        } else !mControllers.settings.getBool(ESetting.FeatureGenband) || // todo: move this check to GenbandCallLogDisplayPresenter
                !mControllers.accounts.primaryAccount.getBool(EAccountSetting.GenbandAccDisableCallTransfer) && // todo: move this check to GenbandCallLogDisplayPresenter
                ClientConfig.get().getGuiVisibility(EGuiElement.CallControlTransfer) != EGuiVisibility.Hidden
    }

    fun isMenuPrefixCallEnabled(position: Int) = mControllers.phone.isPrefixCallingEnabled && isMenuCallEnabled(position)

    fun isMenuConferenceCallEnabled(position: Int): Boolean {
        return mControllers.collaboration.isCollabAvailable // todo
    }

    fun isMenuHostedConferenceCallEnabled(position: Int): Boolean {
        return mControllers.collaboration.isCollabAvailable // todo
    }

    fun isMenuOnlyConferenceRelatedActionsAllowed(position: Int) = isConference(getGroupItemAt(position))

    // Menu actions

    @SuppressLint("RxSubscribeOnError")
    fun onMenuDeleteAll() {
        mExpandedPosition = -1
        mCompositeDisposable.add(mCallLogApi.deleteAll().subscribe())
    }

    @SuppressLint("RxSubscribeOnError")
    fun onMenuDeleteMissed() {
        mExpandedPosition = -1
        mCompositeDisposable.add(mCallLogApi.delete(CallType.MISSED).subscribe())
    }

    @SuppressLint("RxSubscribeOnError")
    fun onMenuDeleteConference() {
        mExpandedPosition = -1
        mCompositeDisposable.add(mCallLogApi.delete(CallType.CONFERENCE).subscribe())
    }

    fun getContextMenuTitle(position: Int) : String? {
        val group = getGroupItemAt(position)
        return if (group?.displayName.isNullOrEmpty()) getNumberFormattedForScreen(group) else group?.displayName
    }

    @SuppressLint("RxSubscribeOnError")
    fun onMenuDelete(position: Int) {
        if (position == mExpandedPosition) {
            mExpandedPosition = -1;
        }

        val group = getGroupItemAt(position) ?: return
        val entities = group.items.map { item -> CallLogEntity().apply { id = item.id.toLong() } }
        mCompositeDisposable.add(mCallLogApi.delete(*entities.toTypedArray()).subscribe())
    }

    fun onMenuCall(position: Int) {
        callNumber(getGroupItemAt(position), CallAction.AudioCall)
        BIAnalytics.get().reportUIClick(BIAnalytics.DIAL_FROM_CALL_LOG_LISTING)
    }

    fun onMenuAddContact(position: Int) {
        val group = getGroupItemAt(position) ?: return

        val phoneNumber = getNumberForContacts(group)
        val phoneType = getPhoneTypeForContacts(phoneNumber)
        val firstName = if (!group.displayName.isNotEmpty()) group.displayName else group.remoteName

        val bundle = Bundle(5)
        bundle.putSerializable(ContactEditPresenter.KEY_SCREEN_ACTION, ContactEditPresenter.ACTION_ADD)
        bundle.putSerializable(GlobalConstants.KEY_CONTACT_TYPE, GlobalConstants.EContactsFilterType.CONTACTS)
        bundle.putString(ContactEditPresenter.KEY_CONTACT_PHONE_NUMBER, phoneNumber)
        bundle.putInt(ContactEditPresenter.KEY_PHONE_TYPE, phoneType)
        bundle.putString(ContactEditPresenter.KEY_FIRST_NAME, firstName)

        firePresenterEvent(Events.SHOW_CONTACT_EDIT, bundle)
    }

    fun onMenuViewContact(position: Int) {
        val group = getGroupItemAt(position) ?: return

        var contactId = group.contactId
        if (contactId == GlobalConstants.CALL_LOG_XMPP_BUDDY_ID) {
            val xmppBuddy = mControllers.buddy.getXmppBuddyByPhoneNumber(group.number) ?: return
            val bundle = Bundle(1)
            bundle.putString(GlobalConstants.KEY_BUDDY_ID, ImpsUtils.getBuddyKey(xmppBuddy))
            firePresenterEvent(Events.SHOW_BUDDY_DETAILS, bundle)
        } else {
            // todo: move this Genband code to GenbandCallLogDisplayPresenter
            if (mControllers.settings.getBool(ESetting.FeatureGenband)) {
                if (!isLocalContactId(group.contactId)) {
                    val genbandKey = group.externalId
                    if (!TextUtils.isEmpty(genbandKey)) {
                        val friend = mControllers.genbandContact.getContactByNickname(genbandKey, group.accountNickname)
                        if (friend != null) {
                            contactId = GenbandContactDataConversion.getFriendsContact(friend).id
                        }
                    }
                }
            }

            val bundle = Bundle(2)
            bundle.putSerializable(GlobalConstants.KEY_CONTACT_TYPE, GlobalConstants.EContactsFilterType.CONTACTS)
            bundle.putString(GlobalConstants.KEY_CONTACT_ID, contactId)

            firePresenterEvent(Events.SHOW_CONTACT_DETAILS, bundle)
        }
    }

    fun onMenuUpdateContact(position: Int) =
           // Explicit call of method with three arguments to avoid interpret passed int value as delay
            firePresenterEvent(Events.SHOW_CONTACT_PICKER, position, 0)

    fun onMenuTransferCall(position: Int) {
        val group = getGroupItemAt(position) ?: return

        if (mControllers.phone.callCount == 1) {
            val call = mControllers.phone.callListCopy[0]
            if (call.isTransferPossible) {
                mControllers.phone.transfer(call.callId, getNumberWithDomain(group), group.accountNickname, false)
            } else {
                firePresenterEvent(Events.SHOW_MESSAGE_LONG, getString(R.string.tPhoneTabTransferNotPossible))
            }
        }
    }

    fun onMenuVideoCall(position: Int) = callNumber(getGroupItemAt(position), CallAction.VideoCall)

    fun onMenuDeleteCallRecord(position: Int) =
           // Explicit call of method with three arguments to avoid interpret passed int value as delay
            firePresenterEvent(Events.SHOW_DELETE_CALL_RECORD_CONFIRMATION, position, 0)

    fun onMenuSendSms(position: Int) {
        val group = getGroupItemAt(position) ?: return

        var account: Account? = mControllers.accounts.primaryAccount

        if (isLocalContactId(group.contactId)) {
            account = mControllers.accounts.getAccount(group.accountNickname)
            if (account?.getBool(EAccountSetting.IsSMS) != true) {
                account = mControllers.accounts.primaryAccount
            }
        }

        val listOfSmsApiAccounts = mControllers.accounts.getAccounts { it.isEnabled && it.type == EAccountType.SmsApi && it.getBool(EAccountSetting.IsSMS) }
        if (mControllers.settings.getBool(ESetting.FeatureSMSApi) && listOfSmsApiAccounts.isNotEmpty()) {
            account = listOfSmsApiAccounts[0]
        }

        if (account == null || !account.isRegistered || !account.getBool(EAccountSetting.IsSMS)) {
            firePresenterEvent(Events.SHOW_MESSAGE_LONG, getString(R.string.tNoBuddiesToastAccountsSMS))
            return
        }

        val accountId = account.getStr(EAccountSetting.Nickname)
        var number = group.number
        if (account.type == EAccountType.SmsApi) {
            number = DialPlanHelper.applyDialPlan(number, account)
            if (!number.contains("@")) {
                number = "$number@${account.getStr(EAccountSetting.Domain)}"
            }
        }

        val numbers = ArrayList<String>().apply { add(number) }
        val conversationType = if (account.type == EAccountType.SmsApi) ImConversationData.SMS_API else ImConversationData.SMS

        val bundle = Bundle(3)
        bundle.putString(GlobalConstants.KEY_CHAT_ACCOUNT, accountId)
        bundle.putInt(GlobalConstants.KEY_CONVERSATION_TYPE, conversationType)
        bundle.putStringArrayList(GlobalConstants.KEY_CHAT_SMS_NUMBERS, numbers)

        firePresenterEvent(Events.SHOW_CONVERSATION, bundle)
    }

    fun onMenuSendIm(position: Int) {
        val group = getGroupItemAt(position) ?: return

        if (!isLocalContactId(group.contactId)) {
            return;
        }

        var account: Account? = mControllers.accounts.getAccount(group.accountNickname)

        // If buddy's corresponding account is deleted or unregistered try to get other registered accounts
        // with enabled IM on same domain (Bria iPhone default behavior)
        // If there are no appropriate accounts error message will be shown (account will remain null or unregistered)
        if (account == null || !account.isRegistered) {
            val accountsForDomain = mControllers.accounts.getActiveImAccountsForDomain(group.accountDomain)
            if (!accountsForDomain.isEmpty()) {
                account = accountsForDomain[0]
            }
        }

        if (account == null || !account.isRegistered || !account.getBool(EAccountSetting.IsIMPresence)) {
            firePresenterEvent(Events.SHOW_MESSAGE_LONG, getString(R.string.tNoBuddiesToastAccountsIM))
            return
        }

        var remoteAddress = group.number
        if (!remoteAddress.contains("@")) {
            remoteAddress = "$remoteAddress@${account.getStr(EAccountSetting.Domain)}"
        }
        val buddyKey = BuddyKeyUtils.getNewBuddyKey(account.type, ImpsUtils.getUserAtDomainForAccount(account), remoteAddress)
        val buddyKeyList = ArrayList<String>().apply { add(buddyKey) }
        val conversationType = if (account.type == EAccountType.Xmpp) ImConversationData.IM_XMPP else ImConversationData.IM_SIP

        val bundle = Bundle(2)
        bundle.putStringArrayList(GlobalConstants.KEY_BUDDY_IDS, buddyKeyList)
        bundle.putSerializable(GlobalConstants.KEY_CONVERSATION_TYPE, conversationType)

        firePresenterEvent(Events.SHOW_CONVERSATION, bundle)
    }

//    fun onMenuFileShare(position: Int) {
//    }

    fun onMenuPrefixCall(position: Int) = callNumber(getGroupItemAt(position), CallAction.PrefixCall)

    fun onMenuConferenceCall(position: Int) {
        val group = getGroupItemAt(position) ?: return

        val vccsLink = group.number

        if (mControllers.collaboration.hasConferenceCall() &&  mControllers.collaboration.equalsVccsLink(vccsLink)) {
            BriaVoipService.get()!!.showCallUi(GlobalConstants.COLLABORATION_CALL_ORIGIN)
        } else {
            firePresenterEvent(Events.SHOW_WEB_VIEW, vccsLink)
        }
    }

    fun onMenuHostedConferenceCall(position: Int) {
        if (ParticipantManager.isCurrentUserModerator()) {
            BriaVoipService.get()!!.showCallUi(GlobalConstants.COLLABORATION_CALL_ORIGIN)
        } else {
            mControllers.collaboration.startCollaboration(null)
        }
    }

    fun onHostedConferenceCallConfirmed(position: Int) = mControllers.collaboration.startCollaboration(null)

    fun onConferenceCallConfirmed(position: Int) = mControllers.collaboration.startCollaboration(Uri.parse(getGroupItemAt(position)?.number))

    fun onContactForUpdateSelected(contactId: String, position: Int) {
        val group = getGroupItemAt(position) ?: return

        val phoneNumber = getNumberForContacts(group)
        val phoneType = getPhoneTypeForContacts(phoneNumber)

        val bundle = Bundle(4)
        bundle.putSerializable(ContactEditPresenter.KEY_SCREEN_ACTION, ContactEditPresenter.ACTION_EDIT)
        bundle.putString(GlobalConstants.KEY_CONTACT_ID, contactId)
        bundle.putString(ContactEditPresenter.KEY_CONTACT_PHONE_NUMBER, phoneNumber)
        bundle.putInt(ContactEditPresenter.KEY_PHONE_TYPE, phoneType)

        firePresenterEvent(Events.SHOW_CONTACT_EDIT, bundle)
    }

    @SuppressLint("RxLeakedSubscription")
    fun onDeleteCallRecordConfirmed(position: Int) {
        val item = getItemAt(position) ?: return
        if (item is CallLogContentItem) {
            // Remove file from storage
            if (!RecordingUtils.deleteFile(context, item.recordPath)) {
                firePresenterEvent(Events.SHOW_MESSAGE_LONG, getString(R.string.tFailedToDeleteRecording))
                return
            }

            // Reset recordPath value in DB
            // todo: avoid blocking call here
            val callLogEntity = mCallLogApi.get(item.id.toLong()).blockingFirst().apply { recordPath = "" }
            mCallLogApi.put(callLogEntity).subscribe({}, { error -> Log.w(TAG, "PUT entity to DB error", error) })
        }
    }

    // Item actions

    fun onItemClicked(position : Int) {
        if (isGroup(position)) {
            toggleGroup(position)
        }
    }

    fun onItemCallButtonClicked(position: Int) {
        val group = getGroupItemAt(position) ?: return

        if (isConference(group)) {
            if (group.hostedCollab) {
                if (mControllers.collaboration.hasConferenceCall()) {
                    firePresenterEvent(Events.SHOW_HOSTED_CONFERENCE_CALL_CONFIRMATION)
                } else {
                    mControllers.collaboration.startCollaboration(null)
                }
            } else {
                if (mControllers.collaboration.hasConferenceCall()) {
                    // Explicit call of method with three arguments to avoid interpret passed int value as delay
                    firePresenterEvent(Events.SHOW_CONFERENCE_CALL_CONFIRMATION, position, 0)
                } else {
                    firePresenterEvent(Events.SHOW_WEB_VIEW, group.number)
                }
            }
        } else {
            if (mControllers.settings.getBool(ESetting.SingleTouchtoCall /*todo: rename to SingleTouchToCall*/) &&
                    !mControllers.phone.isCallbackEnabled &&
                    !isMenuVideoCallEnabled(position) &&
                    !isMenuPrefixCallEnabled(position)) {
                callNumber(group, CallAction.AudioCall)
            } else {
                val account = mControllers.accounts.primaryAccount
                val contact = mContactsApi.contact.byId(group.contactId).withAllData().asMaybe().blockingGet()
                val phoneNumber = contact?.allPhoneNumbers?.find { equalNumbers(it.number, group.number) } ?:
                        PhoneNumber(PhoneNumber.MainType.PHONE_NUMBER, -1, group.number)

                val bundle = Bundle(4)
                bundle.putInt(PhoneNumberActionsScreen.KEY_SELECTED_ACCOUNT, account?.id ?: -1)
                bundle.putSerializable(PhoneNumberActionsScreen.KEY_CONTACT_TYPE, contact?.type)
                bundle.putString(PhoneNumberActionsScreen.KEY_CONTACT_NAME, contact?.displayName)
                bundle.putParcelable(PhoneNumberActionsScreen.KEY_PHONE_NUMBER, phoneNumber)

                firePresenterEvent(Events.SHOW_PHONE_NUMBER_ACTION_PICKER, bundle)
            }
        }
    }

    fun onItemContactButtonClicked(position: Int) {
        val group = getGroupItemAt(position) ?: return
        if (group.contactId.isNotEmpty()) {
            onMenuViewContact(position)
        } else {
            onMenuAddContact(position)
        }
    }

    // = PRIVATE section

    private fun onCallLogDataChanged(newEntityList : List<CallLogEntity>) {
        mGroupedList.clear()
        mGroupedList.addAll(buildGroupedList(newEntityList))

        firePresenterEvent(Events.DATA_SET_CHANGED)
        restartUpdatingContactInfo()
    }

    private fun onContactsChanged() {
        restartUpdatingContactInfo()
    }

    private fun handleCallReminder(reminderBundle: Bundle?) {
        val nameKey = LocalCallLogController.KEY_CALL_REMINDER_NAME
        val dateKey = LocalCallLogController.KEY_CALL_REMINDER_DATE
        if (reminderBundle != null && reminderBundle.containsKey(nameKey) && reminderBundle.containsKey(dateKey)) {
            val message = getString(com.bria.voip.R.string.tCallReminderDisplayText, reminderBundle.getString(nameKey), reminderBundle.getString(dateKey))
            firePresenterEvent(Events.SHOW_CALL_REMINDER, message)
        }
    }

    // Grouped List management

    private fun buildGroupedList(entityList : List<CallLogEntity>) : MutableList<CallLogItemModel> {
        val newGroupedList = mutableListOf<CallLogItemModel>()
        var groupEntity: CallLogEntity? = null
        var group: CallLogGroupItem? = null

        // Build groups
        var groupCount = 0
        for (entity in entityList) {
            if (group != null && groupEntity != null && equalNumbers(groupEntity.number, entity.number)) {
                // Fill the current group with content items
                group.items.add(buildContent(context, entity))

                if (group.callTypes.size < MAX_CALL_TYPE_ICONS) {
                    group.callTypes.add(entity.callType)
                }
            } else {
                // Create a new group
                groupEntity = entity
                group = buildGroup(groupCount++, groupEntity)
                newGroupedList.add(group)
            }
        }

        // Update expanded group
        val oldExpandedGroup = mGroupedList.getOrNull(mExpandedPosition)
        if (oldExpandedGroup != null && newGroupedList.isNotEmpty()) {
            val newExpandedPosition = newGroupedList.indexOf(oldExpandedGroup)
            if (newExpandedPosition == -1) {
                // Not found.
                if (mExpandedPosition == 0) {
                    // The last call was updated (Expanded group extended by new entry)
                    val newExpandedGroup = newGroupedList.first() as CallLogGroupItem
                    newExpandedGroup.isExpanded = true
                    newGroupedList.addAll(1, newExpandedGroup.items)
                } else {
                    // Expanded group was removed
                    mExpandedPosition = -1
                }
            } else {
                // Found. Update expanded position by new value
                mExpandedPosition = newExpandedPosition
                newGroupedList.addAll(mExpandedPosition + 1, (oldExpandedGroup as CallLogGroupItem).items)
            }
        }

        return newGroupedList
    }

    private fun buildGroup(id : Int, callLogEntity: CallLogEntity) : CallLogGroupItem {
        return CallLogGroupItem(
                id = id.toString(),
                contactId = callLogEntity.contactId,
                number = callLogEntity.number,
                numberType = callLogEntity.numberType,
                displayName = getGroupDisplayName(callLogEntity),
                date = callLogEntity.date,
                // Below uses for options only
                remoteName = callLogEntity.remoteName,
                remoteDomain = callLogEntity.remoteDomain,
                accountNickname = callLogEntity.accountNickname,
                hostedCollab = callLogEntity.hostedCollab
        ).apply {
            callTypes.add(callLogEntity.callType)
            items.add(buildContent(context, callLogEntity))
        }
    }

    private fun buildContent(context: Context, callLogEntity: CallLogEntity) : CallLogContentItem {
        return CallLogContentItem(
                id = callLogEntity.id.toString(),
                callType = callLogEntity.callType,
                callInfo = formatCallInfo(callLogEntity),
                callDate = formatCallDate(callLogEntity),
                callDuration = formatCallDuration(context, callLogEntity.durationSeconds),
                isCallDurationVisible = callLogEntity.callType != CallType.MISSED,
                isRecordInfoVisible = callLogEntity.recordPath.isNotEmpty(),
                recordInfo = formatRecordInfo(callLogEntity),
                recordPath = callLogEntity.recordPath
        )
    }

    private fun toggleGroup(position: Int) {
        var currentPosition = position
        val expandedGroup = getItemAt(mExpandedPosition)
        synchronized(mGroupedList) {
            if (expandedGroup is CallLogGroupItem) {
                // Expanded. Collapse it first
                expandedGroup.isExpanded = false
                mGroupedList.removeAll(expandedGroup.items)
                firePresenterEvent(Events.DATA_SET_COLLAPSED, Pair(mExpandedPosition + 1, expandedGroup.items.size))

                // Take care of current position after list was collapsed
                if (currentPosition > mExpandedPosition) {
                    currentPosition -= expandedGroup.items.size
                }
            }

            val currentGroup = getItemAt(currentPosition)
            if (mExpandedPosition == currentPosition) {
                // Already collapsed by code above, just clear state
                mExpandedPosition = -1
            } else if (currentGroup is CallLogGroupItem && !currentGroup.isExpanded) {
                // Collapsed. Expand it now
                mExpandedPosition = currentPosition
                currentGroup.isExpanded = true
                mGroupedList.addAll(mExpandedPosition + 1, currentGroup.items)
                firePresenterEvent(Events.DATA_SET_EXPANDED, Pair(mExpandedPosition + 1, currentGroup.items.size))
            }
        }
    }

    // Update info with contacts data

    private fun restartUpdatingContactInfo() {
        stopUpdatingContactInfo()

        synchronized(mGroupedList) {
            for ((index, group) in mGroupedList.withIndex()) {
                if (group is CallLogGroupItem) {
                    val number = getNumberForContacts(group)
                    // We already on IO scheduler here
                    val groupUpdateDisposable = mContactsApi.contact.byId(group.contactId).withAllData().asMaybe()
                            .switchIfEmpty(mContactsApi.contact.byNumber(number).withAllData().asMaybe())
                            .subscribe({ contact ->
                                group.contactId = contact.id
                                group.displayName = contact.displayName ?: group.displayName
                                group.contactPhoto = contact.getPhotoAsMaybe(context).blockingGet()
                                group.numberType = contact.allPhoneNumbers.
                                        find { equalNumbers(it.number, number) }?.
                                        getSubTypeString(context) ?: group.numberType
                                group.contactPresence = contact.presence?.status?.getIcon(context)
                                group.isContactPresenceVisible = group.contactPresence != null
                                group.externalId = contact.buddyKey ?: ""

                                // Explicit call of method with three arguments to avoid interpret passed index as delay
                                firePresenterEvent(Events.DATA_SET_UPDATED, index, 0)
                            }, {
                                Log.e(TAG, "Contacts error", it)
                            }, {
                                // Contact not found
                                if (group.contactId.isEmpty()) {
                                    // Not found at all, ignore
                                } else {
                                    // Contact was exist before, but now looks like removed. Update with original data
                                    val callLogEntity = mCallLogApi.get(group.getEntityId()).blockingFirst()
                                    group.contactId = ""
                                    group.displayName = getGroupDisplayName(callLogEntity)
                                    group.contactPhoto = null
                                    group.numberType = ""
                                    group.contactPresence = null
                                    group.isContactPresenceVisible = group.contactPresence != null
                                    group.externalId = ""

                                    // Explicit call of method with three arguments to avoid interpret passed index as delay
                                    firePresenterEvent(Events.DATA_SET_UPDATED, index, 0)
                                }
                            })

                    mUpdateContactInfoDisposable.add(groupUpdateDisposable)
                }
            }
        }
    }

    private fun stopUpdatingContactInfo() {
        clear(mUpdateContactInfoDisposable)
    }

    // Helper methods

    private fun getGroupDisplayName(callLogEntity: CallLogEntity) =
            if (callLogEntity.displayName.isNotEmpty()) callLogEntity.displayName else callLogEntity.number

    private fun equalNumbers(number1: String, number2: String) : Boolean {
        // todo: implement matching properly
        return TextUtils.equals(number1, number2)
    }

    private fun formatCallDate(callLogEntity: CallLogEntity) = DateUtils.formatDateTime(context, callLogEntity.date,
            DateUtils.FORMAT_SHOW_TIME or DateUtils.FORMAT_SHOW_DATE or DateUtils.FORMAT_SHOW_YEAR or DateUtils.FORMAT_SHOW_WEEKDAY)

    private fun formatCallDuration(context: Context, elapsedSeconds: Int): String {
        val minutes = elapsedSeconds / 60
        val seconds = elapsedSeconds % 60
        val minutesString = context.getString(if (minutes <= 1) R.string.logMinuteSingle else R.string.logMinutePlural, minutes)
        val secondsString = context.getString(if (seconds <= 1) R.string.logSecondSingle else R.string.logSecondPlural, seconds)
        return if (minutes > 0) "$minutesString $secondsString" else secondsString
    }

    private fun formatCallInfo(callLogEntity : CallLogEntity) = when (callLogEntity.callType) {
        CallType.INCOMING -> getString(R.string.comm_log_incoming_call)
        CallType.OUTGOING -> getString(R.string.comm_log_outgoing_call)
        CallType.MISSED -> getString(R.string.comm_log_missed_call)
        CallType.VOICE_MAIL -> getString(R.string.tVoiceMail)
        CallType.CONFERENCE -> getString(R.string.comm_log_vccs_call)
    }

    private fun formatRecordInfo(callLogEntity: CallLogEntity): String {
        val durationString = formatRecordDuration(callLogEntity)
        return if (durationString.isEmpty()) callLogEntity.recordPath else "${callLogEntity.recordPath} ($durationString)"
    }

    private fun formatRecordDuration(callLogEntity: CallLogEntity): String {
        return if (callLogEntity.recordPath.isNotEmpty()) {
            try {
                val absolutePath = RecordingUtils.getAbsolutePath(context, callLogEntity.recordPath)
                val mediaMetadataRetriever = MediaMetadataRetriever()
                mediaMetadataRetriever.setDataSource(absolutePath)
                val duration = mediaMetadataRetriever.extractMetadata(MediaMetadataRetriever.METADATA_KEY_DURATION)?.toInt() ?: 0
                mediaMetadataRetriever.release()
                val durationSeconds = duration / 1000
                val minutes = durationSeconds / 60
                val seconds = durationSeconds % 60
                "$minutes:$seconds"
            } catch (e: Exception) {
                ""
            }
        } else {
            ""
        }
    }

    private fun getGroupItemAt(position: Int) = getItemAt(position) as? CallLogGroupItem

    private fun isGroup(position: Int): Boolean = getItemAt(position)?.isGroup ?: false

    private fun isAnonymous(group : CallLogGroupItem?) = Validator.isAnonymousAddress(group?.number) || "anonymous".equals(group?.number, ignoreCase = true)

    private fun isConference(group : CallLogGroupItem?) = group?.callTypes?.get(0) == CallType.CONFERENCE

    private fun isLocalContactId(contactId: String?) = !contactId.isNullOrEmpty() &&
            contactId != GlobalConstants.CALL_LOG_GENBAND_FRIEND_ID &&
            contactId != GlobalConstants.CALL_LOG_LDAP_CONTACT &&
            contactId != GlobalConstants.CALL_LOG_XMPP_BUDDY_ID &&
            contactId != GlobalConstants.CALL_LOG_VCCS_CONFERENCE

    private fun isContactExists(group : CallLogGroupItem) = isLocalContactId(group.contactId) // todo: possibly need to get contact here

    private fun callNumber(group : CallLogGroupItem?, callAction: CallAction) =
            group?.let { callNumber(it.number, it.displayName, it.accountNickname, callAction) }

    private fun callNumber(number: String, displayName : String, accountNickname: String, callAction: CallAction) {
        if (!Validator.isValidUserName(number) && !Validator.isValidPhoneNumber(number)) {
            firePresenterEvent(Events.SHOW_MESSAGE_LONG, getString(R.string.tDialNumberIncorrect1))
            return
        }

        when (callAction) {
            CallAction.AudioCall -> mControllers.phone.call(number, accountNickname, displayName, CallData.ECallType.Generic)
            CallAction.PrefixCall -> mControllers.phone.prefixCall(number, accountNickname, displayName)
            CallAction.VideoCall -> mControllers.phone.callVideo(number, accountNickname, displayName)
        }.also { success ->
            if (!success) {
                mControllers.phone.lastError?.let { firePresenterEvent(Events.SHOW_MESSAGE_LONG, it.description) }
            }
        }
    }

    private fun getNumberWithDomain(group: CallLogGroupItem) =
            ImpsUtils.getAddressWithDomain(group.number, if (group.remoteDomain.isNotEmpty()) group.remoteDomain else group.accountDomain)

    private fun getNumberWithoutDomain(group: CallLogGroupItem) = ImpsUtils.removeDomainFromAddress(group.number)

    private fun getNumberFormattedForScreen(group : CallLogGroupItem?) : String {
        group ?: return ""

        val number = if (mControllers.settings.getBool(ESetting.ShowUriDomain)) {
            getNumberWithDomain(group)
        } else {
            mControllers.phone.getFormattedNumber(getNumberWithoutDomain(group), false)
        }

        return if (mIsMetaswitchBrand) Validator.getMetaswitchFormattedNumber(number) else number
    }

    private fun getNumberForContacts(group: CallLogGroupItem): String {
        return if (mControllers.settings.getBool(ESetting.RemoveFullyQualifiedNameFromContacts)) {
            getNumberWithoutDomain(group)
        } else {
            getNumberWithDomain(group)
        }
    }

    private fun getPhoneTypeForContacts(phoneNumber: String): Int {
        return if (mControllers.settings.getBool(ESetting.HaveSoftphoneType) && !Validator.isNumeric(phoneNumber)) {
            PhoneNumber.SOFTPHONE_CODE
        } else {
            ContactsContract.CommonDataKinds.Phone.TYPE_MOBILE
        }
    }

}
