package com.bria.voip.ui.main.calllog.refactoring

import android.Manifest
import android.app.Dialog
import android.content.Intent
import android.content.pm.PackageManager
import android.net.Uri
import android.os.Bundle
import android.support.annotation.MainThread
import android.support.v7.app.AlertDialog
import android.support.v7.widget.LinearLayoutManager
import android.support.v7.widget.RecyclerView
import android.view.ContextMenu
import android.view.MenuItem
import android.view.View
import android.widget.TextView
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.annotations.Menu
import com.bria.common.uiframework.dialogs.ScreenHolderDialog
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.common.uiframework.screens.IScreenEnum
import com.bria.common.util.AndroidUtils
import com.bria.common.util.GlobalConstants
import com.bria.common.util.kotlin.setGone
import com.bria.common.util.kotlin.setVisible
import com.bria.voip.R
import com.bria.voip.ui.main.misc.EScreenList
import com.bria.voip.ui.shared.pickers.ContactPickerScreen

// Context menu IDs

private const val CONTEXT_MENU_DELETE = 1
private const val CONTEXT_MENU_CALL = 2
private const val CONTEXT_MENU_ADD_CONTACT = 3
private const val CONTEXT_MENU_VIEW_CONTACT = 4
private const val CONTEXT_MENU_UPDATE_CONTACT = 5
private const val CONTEXT_MENU_DELETE_CALL_RECORD = 6
private const val CONTEXT_MENU_SEND_SMS = 7
private const val CONTEXT_MENU_SEND_IM = 8
//private const val CONTEXT_MENU_FILE_SHARE = 9
private const val CONTEXT_MENU_VIDEO_CALL = 10
private const val CONTEXT_MENU_TRANSFER_CALL = 11
private const val CONTEXT_MENU_PREFIX_CALL = 12
private const val CONTEXT_MENU_CONFERENCE_CALL = 13
private const val CONTEXT_MENU_HOSTED_CONFERENCE_CALL = 14

// Dialog IDs

private const val CALL_REMINDER_DIALOG_ID = 3300
// private const val CALL_IN_PROGRESS_DIALOG_ID = 3301
private const val DELETE_CALL_RECORD_CONFIRM_DIALOG_ID = 3302
private const val HOSTED_CONFERENCE_CALL_CONFIRM_DIALOG_ID = 3303
private const val CONFERENCE_CALL_CONFIRM_DIALOG_ID = 3304
private const val CONTACT_DISPLAY_SCREEN_HOLDER_DIALOG_ID = 3305
private const val BUDDY_DISPLAY_SCREEN_HOLDER_DIALOG_ID = 3306
private const val CONTACT_EDIT_SCREEN_HOLDER_DIALOG_ID = 3307
private const val PHONE_NUMBER_ACTION_PICKER_ID = 3308

// Bundle Keys

private const val KEY_CALL_REMINDER_MESSAGE = "KEY_CALL_REMINDER_MESSAGE"
private const val KEY_ITEM_POSITION = "KEY_ITEM_POSITION"

// Screen Tag
private const val TAG = "CallLogListScreen"

@Layout(R.layout.calllog_list_screen)
@Menu(R.menu.call_log_list_menu)
class CallLogListScreen : AbstractScreen<CallLogListPresenter>() {

    @Inject(R.id.calllog_list_screen_recycler_view)
    private lateinit var mRecyclerView: RecyclerView

    @Inject(R.id.calllog_list_screen_empty_view)
    private lateinit var mEmptyView: TextView

    private lateinit var mAdapter: CallLogListAdapter

    private var mLastClickedItemPosition: Int = RecyclerView.NO_POSITION

    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        when (event.type) {
            CallLogListPresenter.Events.DATA_SET_CHANGED -> handleDataSetChanged()
            CallLogListPresenter.Events.DATA_SET_EXPANDED -> (event.data as Pair<Int, Int>).let { mAdapter.notifyDataInserted(it.first, it.second) }
            CallLogListPresenter.Events.DATA_SET_COLLAPSED -> (event.data as Pair<Int, Int>).let { mAdapter.notifyDataRemoved(it.first, it.second) }
            CallLogListPresenter.Events.DATA_SET_UPDATED -> mAdapter.notifyDataChanged(event.data as Int)

            CallLogListPresenter.Events.SHOW_MESSAGE_LONG -> toastLong(event.data as String)
            CallLogListPresenter.Events.SHOW_CALL_REMINDER -> showCallReminderDialog(event.data as String)
            CallLogListPresenter.Events.SHOW_DELETE_CALL_RECORD_CONFIRMATION -> showDeleteCallRecordConfirmationDialog(event.data as Int)
            CallLogListPresenter.Events.SHOW_CONFERENCE_CALL_CONFIRMATION ->  showConferenceCallConfirmationDialog(event.data as Int)
            CallLogListPresenter.Events.SHOW_HOSTED_CONFERENCE_CALL_CONFIRMATION ->  showHostedConferenceCallConfirmationDialog(event.data as Int)
            CallLogListPresenter.Events.SHOW_WEB_VIEW -> showWebViewActivity(event.data as String)
            CallLogListPresenter.Events.SHOW_PHONE_NUMBER_ACTION_PICKER -> showPhoneNumberActionPicker(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_CONTACT_PICKER ->  showContactPickerScreen(event.data as Int)
            CallLogListPresenter.Events.SHOW_CONTACT_EDIT -> showContactEditScreen(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_CONTACT_DETAILS -> showContactDetailsScreen(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_BUDDY_DETAILS -> showBuddyDetailsScreen(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_CONVERSATION -> showConversationScreen(event.data as Bundle)
        }
    }

    override fun getPresenterClass() = CallLogListPresenter::class.java

    override fun getTitle() = presenter.title

    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)
        mAdapter = CallLogListAdapter(mRecyclerView, presenter)
        mAdapter.onItemClickListener = { _, position -> presenter.onItemClicked(position) }
        mAdapter.onCreateContextMenuListener = mOnCreateContextMenuListener
        mAdapter.onItemCallButtonClickListener = { _, position -> presenter.onItemCallButtonClicked(position) }
        mAdapter.onItemContactButtonClickListener = { _, position -> presenter.onItemContactButtonClicked(position) }

        mRecyclerView.adapter = mAdapter
        mRecyclerView.layoutManager = LinearLayoutManager(activity)
    }

    override fun onStart(bundle: Bundle?) {
        super.onStart(bundle)
        presenter.subscribe(this)
    }

    @MainThread
    override fun onNewConfig(bundle: Bundle?) {
        super.onNewConfig(bundle)
        presenter.onNewConfig(bundle)
    }

    override fun onStop(finishing: Boolean) {
        super.onStop(finishing)
        presenter.unsubscribe()
    }

    override fun updateMenuItems(menu: android.view.Menu, updateKey: String?) {
        super.updateMenuItems(menu, updateKey)

        if (presenter.isMenuDeleteAllEnabled().not()) menu.removeItem(R.id.call_log_list_delete_all)
        if (presenter.isMenuDeleteMissedEnabled().not()) menu.removeItem(R.id.call_log_list_delete_missed)
        if (presenter.isMenuDeleteConferenceEnabled().not()) menu.removeItem(R.id.call_log_list_delete_conference)
    }

    @MainThread
    override fun onMenuItemClick(menuItem: MenuItem): Boolean {
        return when (menuItem.itemId) {
            R.id.call_log_list_delete_all -> { presenter.onMenuDeleteAll(); true }
            R.id.call_log_list_delete_missed -> { presenter.onMenuDeleteMissed(); true }
            R.id.call_log_list_delete_conference -> { presenter.onMenuDeleteConference(); true }
            else -> false
        } || super.onMenuItemClick(menuItem)
    }

    @MainThread
    override fun onNewMessage(message: Bundle, sender: IScreenEnum) {
        super.onNewMessage(message, sender)
        if (sender == EScreenList.CONTACT_PICKER) {
            val position = message.getInt(KEY_ITEM_POSITION, -1)
            val contactId = message.getStringArrayList(ContactPickerScreen.SELECTED_CONTACT_IDS)?.get(0) ?: ""
            if (position != -1 && contactId.isNotEmpty()) {
                presenter.onContactForUpdateSelected(contactId, position)
            }
        }
    }

    override fun createDialog(which : Int, data : Bundle?) : Dialog? {
        when(which) {
            CALL_REMINDER_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setTitle(R.string.tCallReminder)
                        .setMessage(data?.getString(KEY_CALL_REMINDER_MESSAGE))
                        .setNeutralButton(R.string.tOk, null)
                        .setCancelable(true)
                        .create()
//            CALL_IN_PROGRESS_DIALOG_ID ->
//                return AlertDialog.Builder(activity)
//                        .setMessage(R.string.tCallInProgressWait)
//                        .setPositiveButton(R.string.tOk, null)
//                        .setCancelable(true)
//                        .create()
            DELETE_CALL_RECORD_CONFIRM_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setMessage(R.string.tAreYouSureToDeleteCallRecording)
                        .setPositiveButton(R.string.tYes) { _, _ -> presenter.onDeleteCallRecordConfirmed(data?.getInt(KEY_ITEM_POSITION) ?: -1) }
                        .setNegativeButton(R.string.tNo, null)
                        .setCancelable(true)
                        .create()
            CONFERENCE_CALL_CONFIRM_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setTitle(R.string.tCollab_AnotherConferenceTitle)
                        .setMessage(R.string.tCollab_AnotherConferenceText)
                        .setPositiveButton(R.string.tCollab_AnotherConferencePositive) { _, _ -> presenter.onConferenceCallConfirmed(data?.getInt(KEY_ITEM_POSITION) ?: -1) }
                        .setNegativeButton(R.string.tCollab_AnotherConferenceNegative, null)
                        .setCancelable(true)
                        .create()
            HOSTED_CONFERENCE_CALL_CONFIRM_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setTitle(R.string.tCollab_AnotherConferenceTitle)
                        .setMessage(R.string.tCollab_AnotherConferenceText)
                        .setPositiveButton(R.string.tCollab_AnotherConferencePositive) { _, _ -> presenter.onHostedConferenceCallConfirmed(data?.getInt(KEY_ITEM_POSITION) ?: -1) }
                        .setNegativeButton(R.string.tCollab_AnotherConferenceNegative, null)
                        .setCancelable(true)
                        .create()
            CONTACT_DISPLAY_SCREEN_HOLDER_DIALOG_ID ->
                return createScreenHolderDialog(EScreenList.CONTACT_DISPLAY, ScreenHolderDialog.DIALOG_FULL_SCREEN, data)
            BUDDY_DISPLAY_SCREEN_HOLDER_DIALOG_ID ->
                return createScreenHolderDialog(EScreenList.BUDDY_DISPLAY, ScreenHolderDialog.DIALOG_FULL_SCREEN, data)
            CONTACT_EDIT_SCREEN_HOLDER_DIALOG_ID ->
                return createScreenHolderDialog(EScreenList.CONTACT_EDIT, ScreenHolderDialog.DIALOG_MAX_SIZE_NO_MARGIN, data)
            PHONE_NUMBER_ACTION_PICKER_ID ->
                return ScreenHolderDialog
                        .builder(activity)
                        .screen(EScreenList.PHONE_ACTION_SELECT)
                        .bottomSheet()
                        .bundle(data)
                        .build()
        }
        return super.createDialog(which, data)
    }

    private fun createScreenHolderDialog(screen: EScreenList, @ScreenHolderDialog.DialogStyle style: Int, bundle: Bundle?): Dialog {
        return ScreenHolderDialog
                .builder(activity)
                .screen(screen)
                .style(style)
                .showTitle(false)
                .bundle(bundle)
                .build()
    }

    private fun showCallReminderDialog(message : String) {
        showDialog(CALL_REMINDER_DIALOG_ID, Bundle(1).apply { putString(KEY_CALL_REMINDER_MESSAGE, message) })
    }

    private fun showDeleteCallRecordConfirmationDialog(position : Int) {
        showDialog(DELETE_CALL_RECORD_CONFIRM_DIALOG_ID, Bundle(1).apply { putInt(KEY_ITEM_POSITION, position) })
    }

    private fun showConferenceCallConfirmationDialog(position: Int) {
        showDialog(CONFERENCE_CALL_CONFIRM_DIALOG_ID, Bundle(1).apply { putInt(KEY_ITEM_POSITION, position) })
    }

    private fun showHostedConferenceCallConfirmationDialog(position: Int) {
        showDialog(HOSTED_CONFERENCE_CALL_CONFIRM_DIALOG_ID, Bundle(1).apply { putInt(KEY_ITEM_POSITION, position) })
    }

    private fun showWebViewActivity(link : String) {
        val webIntent = Intent(Intent.ACTION_VIEW, Uri.parse(link))
        if (AndroidUtils.canHandleIntent(webIntent, activity)) {
            activity.startActivity(webIntent)
        }
    }

    private fun showPhoneNumberActionPicker(bundle: Bundle) {
        showDialog(PHONE_NUMBER_ACTION_PICKER_ID, bundle)
    }

    private fun showContactPickerScreen(position : Int) {
        val bundle = Bundle(2).apply {
            putInt(KEY_ITEM_POSITION, position);
            putInt(ContactPickerScreen.MAX_NUMBER_OF_CHOICES, 1)
        }
        showScreenForResult(EScreenList.CONTACT_PICKER, ScreenHolderDialog.DIALOG_MAX_SIZE_NO_MARGIN, bundle)
    }

    private fun showContactEditScreen(bundle: Bundle) {
        goTo(EScreenList.CONTACT_EDIT, CONTACT_EDIT_SCREEN_HOLDER_DIALOG_ID, bundle)
    }

    private fun showContactDetailsScreen(bundle: Bundle) {
        goTo(EScreenList.CONTACT_DISPLAY, CONTACT_DISPLAY_SCREEN_HOLDER_DIALOG_ID, bundle)
    }

    private fun showBuddyDetailsScreen(bundle: Bundle) {
        goTo(EScreenList.BUDDY_DISPLAY, BUDDY_DISPLAY_SCREEN_HOLDER_DIALOG_ID, bundle)
    }

    private fun showConversationScreen(bundle: Bundle) {
        mCoordinator.flow(bundle).goTo(EScreenList.CONVERSATION)
    }

    private fun goTo(screen: EScreenList, which: Int, bundle: Bundle) {
        if (isInTabletMode) {
            showDialog(which, bundle)
        } else {
            mCoordinator.flow(bundle).goTo(screen)
        }
    }

    private fun handleDataSetChanged() {
        mAdapter.notifyDataChanged()

        if (presenter.itemsCount == 0) {
            mEmptyView.setVisible()
            mRecyclerView.setGone()
        } else {
            mEmptyView.setGone()
            mRecyclerView.setVisible()
        }

        invalidateMenu()
    }

    private fun checkOrRequestContactsPermission(permissionCode : Int) : Boolean {
        if (checkPermission(Manifest.permission.WRITE_CONTACTS)) {
            return true
        }
        requestPermission(Manifest.permission.WRITE_CONTACTS, permissionCode, getString(R.string.tPermissionContacts))
        return false
    }

    @MainThread
    override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<String>, grantResults: IntArray) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults)

        if (grantResults.isEmpty()) {
            return
        }

        if (mLastClickedItemPosition == -1) {
            return
        }

        if (grantResults[0] == PackageManager.PERMISSION_GRANTED) {
            when (requestCode) {
                GlobalConstants.CP_PERMISSION_ADD_CONTACT -> presenter.onMenuAddContact(mLastClickedItemPosition)
                GlobalConstants.CP_PERMISSION_VIEW_CONTACT -> presenter.onMenuViewContact(mLastClickedItemPosition)
                GlobalConstants.CP_PERMISSION_UPDATE_CONTACT -> showContactPickerScreen(mLastClickedItemPosition)
            }
        }
    }

    private var mOnCreateContextMenuListener = { menu: ContextMenu, _: View, _: ContextMenu.ContextMenuInfo?, position: Int ->

        mLastClickedItemPosition = position

        menu.setHeaderTitle(presenter.getContextMenuTitle(position))

        if (presenter.isMenuOnlyConferenceRelatedActionsAllowed(position)) {
            if (presenter.isMenuConferenceCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_CONFERENCE_CALL, ContextMenu.NONE, R.string.comm_log_vccs_call /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuHostedConferenceCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_HOSTED_CONFERENCE_CALL, ContextMenu.NONE, R.string.tCollab_CreateNewConferenceAgain)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }

            if (presenter.isMenuDeleteEnabled()) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_DELETE, ContextMenu.NONE, R.string.tDelete)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }

            if (presenter.isMenuDeleteCallRecordEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_DELETE_CALL_RECORD, ContextMenu.NONE, R.string.tDeleteCallRecording)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
        } else {
            if (presenter.isMenuDeleteEnabled()) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_DELETE, ContextMenu.NONE, R.string.tDelete)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_CALL, ContextMenu.NONE, R.string.callLog /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuAddContactEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_ADD_CONTACT, ContextMenu.NONE, R.string.addLogToContacts)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuViewContactEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_VIEW_CONTACT, ContextMenu.NONE, R.string.viewLogContactDetails)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuUpdateContactEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_UPDATE_CONTACT, ContextMenu.NONE, R.string.updateExistingLogContactDetails)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuDeleteCallRecordEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_DELETE_CALL_RECORD, ContextMenu.NONE, R.string.tDeleteCallRecording)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuSendSmsEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_SEND_SMS, ContextMenu.NONE, R.string.tSendSms)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuSendImEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_SEND_IM, ContextMenu.NONE, R.string.tBuddySendIM)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
//        if (presenter.isMenuFileShareEnabled(position)) {
//            menu.add(ContextMenu.NONE, CONTEXT_MENU_FILE_SHARE, ContextMenu.NONE, R.string.tFileShare)
//                    .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
//        }
            if (presenter.isMenuVideoCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_VIDEO_CALL, ContextMenu.NONE, R.string.videoCallLog /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuTransferCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_TRANSFER_CALL, ContextMenu.NONE, R.string.callTransferLog /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuPrefixCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_PREFIX_CALL, ContextMenu.NONE, R.string.tPrefixCall /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
        }
    }

    private var mOnContextMenuItemClickListener = MenuItem.OnMenuItemClickListener { menuItem ->
        var handled = true
        when (menuItem.itemId) {
            CONTEXT_MENU_DELETE -> presenter.onMenuDelete(mLastClickedItemPosition)
            CONTEXT_MENU_CALL -> presenter.onMenuCall(mLastClickedItemPosition)
            CONTEXT_MENU_ADD_CONTACT -> {
                if (checkOrRequestContactsPermission(GlobalConstants.CP_PERMISSION_ADD_CONTACT)) {
                    presenter.onMenuAddContact(mLastClickedItemPosition)
                }
            }
            CONTEXT_MENU_VIEW_CONTACT -> {
                if (checkOrRequestContactsPermission(GlobalConstants.CP_PERMISSION_VIEW_CONTACT)) {
                    presenter.onMenuViewContact(mLastClickedItemPosition)
                }
            }
            CONTEXT_MENU_UPDATE_CONTACT -> {
                if (checkOrRequestContactsPermission(GlobalConstants.CP_PERMISSION_UPDATE_CONTACT)) {
                    presenter.onMenuUpdateContact(mLastClickedItemPosition)
                }
            }
            CONTEXT_MENU_DELETE_CALL_RECORD ->  presenter.onMenuDeleteCallRecord(mLastClickedItemPosition)
            CONTEXT_MENU_SEND_SMS -> presenter.onMenuSendSms(mLastClickedItemPosition)
            CONTEXT_MENU_SEND_IM -> presenter.onMenuSendIm(mLastClickedItemPosition)
            //CONTEXT_MENU_FILE_SHARE -> presenter.onMenuFileShare(mLastClickedItemPosition)
            CONTEXT_MENU_VIDEO_CALL -> presenter.onMenuVideoCall(mLastClickedItemPosition)
            CONTEXT_MENU_TRANSFER_CALL -> presenter.onMenuTransferCall(mLastClickedItemPosition)
            CONTEXT_MENU_PREFIX_CALL -> presenter.onMenuPrefixCall(mLastClickedItemPosition)
            CONTEXT_MENU_CONFERENCE_CALL -> presenter.onMenuConferenceCall(mLastClickedItemPosition)
            CONTEXT_MENU_HOSTED_CONFERENCE_CALL -> presenter.onMenuHostedConferenceCall(mLastClickedItemPosition)

            else -> handled = false
        }
        handled
    }
}
