package com.bria.voip.ui.main.im

import android.content.Context
import android.support.annotation.IntDef
import android.text.TextUtils
import com.bria.common.controller.Controllers
import com.bria.common.controller.collaboration.utils.matcher.MatcherKey
import com.bria.common.controller.collaboration.utils.matcher.OnParticipantMatchedListener
import com.bria.common.controller.collaboration.utils.matcher.ParticipantMatch
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.IImData
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.uireusable.dataprovider.AbstractFilterableListDataProvider
import com.bria.common.uireusable.datatypes.ConversationListItemData
import com.bria.common.util.Log
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import java.util.*

class ConversationListDataProvider(val imData: IImData, val context: Context) : AbstractFilterableListDataProvider<ConversationListItemData>(), OnParticipantMatchedListener {

    companion object {
        const val TAG = "ConversationListDataProvider"
        const val FILTER_TYPE_IM = 0
        const val FILTER_TYPE_SMS = 1
        const val FILTER_TYPE_CHAT_ROOM = 2
    }

    @IntDef(FILTER_TYPE_IM, FILTER_TYPE_SMS, FILTER_TYPE_CHAT_ROOM)
    @Retention(AnnotationRetention.SOURCE)
    internal annotation class ImFilterType

    @ImFilterType
    var filterType = FILTER_TYPE_IM
    set(value) {
        if (field != value) {
            field = value
            refilter()
        }
    }

    private val disposables = CompositeDisposable()

    private val mPendingMatch = HashMap<MatcherKey, ConversationListItemData>()

    override fun provideFullDataSet(): List<ConversationListItemData> {
        val ret = ArrayList<ConversationListItemData>()
        if (filterType == FILTER_TYPE_CHAT_ROOM) {
            val chatRooms = imData.chatRoomRepo.getAllChatRooms().blockingGet()
            chatRooms.forEach {
                val itemData = ConversationListItemData(context, it)
                itemData.noOfUnreadMsgs = imData.chatRoomRepo.getNumberOfUnreadMessagesForChatRoom(itemData.id).blockingGet()
                //insert even if it has no messages
                ret.add(itemData)
            }

        } else {
            val conversations: Collection<ImConversationData>
            if (filterType == FILTER_TYPE_IM) {
                conversations = imData.allImConversationsByModTime
            } else {
                conversations = imData.allSmsConversationsByModTime
            }
            for (conversationData in conversations) {
                val itemData = ConversationListItemData(context, conversationData)
                if (conversationData.isSMSType) {
                    val imAddress = BuddyKeyUtils.getImAdressFromNewBuddyKey(conversationData.participants)
                    findContactByNumberAsync(itemData, imAddress)
                }

                itemData.noOfUnreadMsgs = imData.getNumberOfUnreadMessagesForConversation(conversationData.id!!)
                if (!TextUtils.isEmpty(itemData.lastMessageDate)) {
                    ret.add(itemData)
                }
            }
        }
        return ret
    }

    override fun provideFilteredDataSet(fullDataSet: List<ConversationListItemData>, query: String): List<ConversationListItemData> {
        return fullDataSet
    }

    override fun onParticipantMatched(matched: ParticipantMatch) {
        Log.d(TAG, "onParticipantMatched: ")
        mPendingMatch[matched.key]?.let {
            it.updateDisplayData(matched.displayName, matched.photo)
            it.isContactMatched = true
            mDataProviderListener.onDataLoaded(isEmpty)
        }
    }

    private fun findContactByNumberAsync(conversationListItemData: ConversationListItemData?, number: String) {
        // todo: care of disposable
        val disposable = Controllers.get().contacts
                .contactsApi
                .contact
                .byNumber(number)
                .asMaybe()
                .subscribeOn(Schedulers.io())
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe({ contact ->
                    if (conversationListItemData != null) {
                        conversationListItemData.updateDisplayData(contact.displayName, contact.getPhotoAsMaybe(context).blockingGet())
                        conversationListItemData.isContactMatched = true
                        mDataProviderListener.onDataLoaded(isEmpty)
                    }
                }, { throwable -> Log.d(TAG, "Error ", throwable) })
    }
}
