package com.bria.voip.ui.main.im.chatroom

import android.os.Bundle
import com.bria.common.controller.Controllers
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.IAccountsFilter
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.im.refactoring.roomdb.entities.getInitials
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.util.GlobalConstants
import com.bria.common.util.im.ParticipantsSet
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable

open class ChatRoomConfPresenter : AbstractPresenter() {

    companion object {
        const val MAX_ROOM_NAME_SIZE = 22
    }

    var dataValid: Boolean = false

    var chatRoomApi = Controllers.get().im.chatApi.chatRoomApi

    var chatRoomRepo = Controllers.get().im.imData.chatRoomRepo

    var chatRoom = ChatRoom()

    var disposable = CompositeDisposable()

    enum class Events : IPresenterEventTypeEnum {
        INVALIDATE_MENU,
        GO_TO_CHAT_SCREEN,
        ROOM_NAME_EMPTY,
        ROOM_NAME_TOO_LONG,
        ROOM_NAME_DUPLICATED,
        UPDATE_INITIALS;

        override fun getType(): IPresenterEventTypeEnum = this
    }

    init {
        //for Stretto we will consider single XMPP account case
        Controllers.get().accounts.getAccounts(IAccountsFilter.XMPP).firstOrNull()?.let {
            chatRoom.owners = ParticipantsSet(getSelfInfoKey(account = it))
            chatRoom.accountId = ImpsUtils.removePortFromDomain(it.getUserAtDomain())

        }
        disposable.add(chatRoomRepo.getOnRoomAddedObservable().subscribeOn(AndroidSchedulers.mainThread()).subscribe({
            it.firstOrNull { it.name == chatRoom.name }?.let {
                val bundle = Bundle(1)
                bundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, it.id)
                firePresenterEvent(Events.GO_TO_CHAT_SCREEN, bundle)
            }


        }, {
            //nvtd show error
        }))
    }


    var title: String = "Create Chat Room"

    fun handleCreateRoom() {
        //collectData
        chatRoom.name.trim()
        chatRoom.name.trimStart()
        validateRoom()
        if (dataValid) {
            chatRoom.creationTime = System.currentTimeMillis()
            chatRoomApi.createRoom(chatRoom = chatRoom)
        }
    }

    private fun validateRoom() {
        dataValid = when {
            chatRoom.name.isEmpty() -> {
                firePresenterEvent(Events.ROOM_NAME_EMPTY)
                false
            }
            chatRoom.name.length > MAX_ROOM_NAME_SIZE -> {
                firePresenterEvent(Events.ROOM_NAME_TOO_LONG)
                false
            }
            chatRoomRepo.roomNameInTheList(chatRoom.name) -> {
                firePresenterEvent(Events.ROOM_NAME_DUPLICATED)
                false
            }
            else -> true
        }
    }

    fun invitedMembersUpdated(buddyList: List<String>) {
        chatRoom.invited.clear()
        chatRoom.invited = ParticipantsSet(BuddyKeyUtils.getStringFromBuddyKeys(buddyList.toSet()))
        if (chatRoom.invited.isNotEmpty()) {
            chatRoom.accountId = BuddyKeyUtils.getAccountFromNewBuddyKey(chatRoom.invited.firstParticipantKey)
        }
    }

    fun getMembersText(): String = chatRoom.invited.getShortParticipantList(context)

    fun getInvitedMembers() : String {
        return chatRoom.invited.serializedParticipantKeys
    }

    fun getNameText(): String = chatRoom.name

    fun getOwnersText(): String = chatRoom.owners.participantsString

    val activeMembers: Set<String>
        get() = chatRoom.members.map { it.remoteAddress }.toSet()

    val tentativeMembers: Set<String>
        get() = chatRoom.invited.map { it.remoteAddress }.toSet()

    fun updateName(name: String) {
        chatRoom.name = name
        firePresenterEvent(Events.UPDATE_INITIALS, getInitials())
    }

    fun getInitials() : String {
        val initials = chatRoom.name.getInitials()
        return if (initials.isEmpty()) "#" else initials
    }

    private fun getSelfInfoKey(account: Account): String {
        return BuddyKeyUtils.getNewBuddyKey(EAccountType.Xmpp, account.getUserAtDomain(), account.getUserAtDomain())
    }

    fun updateDescription(description: String) {
        chatRoom.description = description
    }

    fun setPrivate(type: Int) {
        chatRoom.privateRoom = type
    }

    fun getType(): Int {
        return chatRoom.privateRoom
    }
}

fun Account.getUserAtDomain(): String = "${this.accountSettings.get(EAccountSetting.UserName)}@${this.accountSettings.get(EAccountSetting.Domain)}"

