package com.bria.voip.ui.main.im.chatroom

import android.os.Bundle
import android.support.annotation.MainThread
import android.support.design.widget.TextInputLayout
import android.text.Editable
import android.text.TextWatcher
import android.view.Gravity
import android.view.MenuItem
import android.view.View
import android.widget.EditText
import android.widget.ImageButton
import android.widget.Switch
import android.widget.TextView
import com.bria.common.controller.Controllers
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents
import com.bria.common.uiframework.annotations.Clickable
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.coloring.Coloring
import com.bria.common.uiframework.dialogs.ScreenHolderDialog
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.common.uiframework.screens.IScreenEnum
import com.bria.common.util.Log
import com.bria.common.util.im.ParticipantsSet
import com.bria.voip.R
import com.bria.voip.ui.main.im.chatroom.ChatRoomConfPresenter.Events
import com.bria.voip.ui.main.misc.EScreenList
import com.bria.voip.ui.shared.pickers.ChatRoomMembersPickerScreen
import com.bria.voip.ui.shared.pickers.generic.AbstractPickerScreen

@Layout(R.layout.chat_room_conf)
@com.bria.common.uiframework.annotations.Menu(R.menu.edit_menu)
class ChatRoomConfScreen<Presenter : ChatRoomConfPresenter> : AbstractScreen<Presenter>() {

    private companion object {
        private const val TAG = "ChatRoomConfScreen"

        private const val TYPE_KEY = "TYPE"
        private const val NAME_KEY = "NAME"
        private const val MEMBERS_KEY = "MEMBERS"
    }

    @Inject(R.id.chat_room_conf_type)
    private lateinit var typeSwitch: Switch

    @Inject(R.id.chat_room_conf_type_desc)
    private lateinit var typeDescription: TextView

    @Inject(R.id.chat_room_conf_initials)
    private lateinit var initials: TextView

    @Inject(R.id.chat_room_conf_name_container)
    private lateinit var nameContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_name)
    private lateinit var name: EditText

    @Inject(R.id.chat_room_conf_owners_container)
    private lateinit var ownersContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_owners)
    private lateinit var owners: EditText

    @Inject(R.id.chat_room_conf_owners_add)
    @Clickable
    private lateinit var addOwners: ImageButton

    @Inject(R.id.chat_room_conf_members_add)
    @Clickable
    private lateinit var addMembers: ImageButton

    @Inject(R.id.chat_room_conf_members_container)
    private lateinit var membersContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_members)
    private lateinit var members: EditText

    @Inject(R.id.chat_room_conf_description_container)
    private lateinit var descriptionContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_description)
    private lateinit var description: EditText

    @Inject(R.id.chat_room_conf_passcode_container)
    private lateinit var passwordContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_passcode)
    private lateinit var password: EditText

    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)
        addMembers.isEnabled = false
        membersContainer.isEnabled = false
        presenter.subscribe(this)
        typeSwitch.setOnCheckedChangeListener { button, isChecked ->
                updateType()
            if (!isChecked) {
                addMembers.isEnabled = false
                presenter.invitedMembersUpdated(ArrayList())
                members.setText(presenter.getMembersText())
                membersContainer.isEnabled = false
            } else {
                addMembers.isEnabled = true
                membersContainer.isEnabled = true
            }
        }
        name.afterChange { presenter.updateName(it.toString()) }
        description.afterChange { presenter.updateDescription(it) }
        bundle?.let {bundle ->
            if (bundle.containsKey(TYPE_KEY)) {
                typeSwitch.isChecked = bundle.getBoolean(TYPE_KEY)

            }
            if (bundle.containsKey(NAME_KEY)) {
                name.setText(bundle.getString(NAME_KEY, ""))
            }
            if (bundle.containsKey(MEMBERS_KEY)) {
                val membersString = bundle.getString(MEMBERS_KEY)
                val membersSet = ParticipantsSet(membersString)
                presenter.invitedMembersUpdated(membersSet.map { it.key})
                members.setText(presenter.getMembersText())
            }
        }
    }

    override fun onDestroy(finishing: Boolean) {
        super.onDestroy(finishing)
        presenter.unsubscribe()
        destroyPresenter()
    }

    override fun onSaveState(stateBundle: Bundle) {
        stateBundle.putBoolean(TYPE_KEY, presenter.getType() == 1)
        stateBundle.putString(NAME_KEY, presenter.getNameText())
        stateBundle.putString(MEMBERS_KEY, presenter.getInvitedMembers())
        super.onSaveState(stateBundle)
    }

    override fun onResume() {
        super.onResume()
        owners.setText(presenter.getOwnersText())
        members.setText(presenter.getMembersText())
        name.setText(presenter.getNameText())
        if (presenter.getNameText() != name.text.toString()) {
            name.setText(presenter.getNameText())
        }
        initials.text = presenter.getInitials()
    }

    private fun updateType() {
        if (typeSwitch.isChecked) {
            presenter.setPrivate(1)
            typeSwitch.text = getString(R.string.tPrivate)
            typeDescription.text = getString(R.string.tPrivateRoomDescription)
        } else {
            presenter.setPrivate(0)
            typeSwitch.text = getString(R.string.tPublic)
            typeDescription.text = getString(R.string.tPublicRoomDescription)
        }
    }

    override fun onMenuItemClick(menuItem: MenuItem?): Boolean {
        when (menuItem?.itemId) {
            R.id.mi_edit_save -> {
                presenter.handleCreateRoom()
            }
            R.id.mi_edit_discard -> {
                coordinator.flow().goBack()
            }
        }
        return super.onMenuItemClick(menuItem)

    }

    override fun onClick(v: View) {
        when (v.id) {
            R.id.chat_room_conf_members_add -> {
                val bundle = Bundle()
                bundle.putStringArrayList(ChatRoomMembersPickerScreen.FIXED_MEMBERS, ArrayList(presenter.activeMembers))

                val initiallySelectedIds = Controllers
                    .get()
                    .buddy
                    .getListOfBuddies(Buddy.EBuddyType.XMPP, IBuddyCtrlEvents.EBuddyGroupType.ALL, IBuddyCtrlEvents.EBuddyFilterType.ALL)
                    .filter { presenter.tentativeMembers.contains(it.imAddress) }
                    .map { it.uniqueIdentifier }
                    .toTypedArray()

                bundle.putStringArray(AbstractPickerScreen.INITIALLY_SELECTED_IDS, initiallySelectedIds)

                showScreenForResult(EScreenList.CHAT_ROOMS_MEMBERS_PICKER_SCREEN, bundle)
            }
        }
        super.onClick(v)

    }

    @MainThread
    override fun onNewMessage(message: Bundle, sender: IScreenEnum) {
        super.onNewMessage(message, sender)

        when (sender) {
            EScreenList.CHAT_ROOMS_MEMBERS_PICKER_SCREEN -> {
                val buddyList = message.getStringArrayList(ChatRoomMembersPickerScreen.CHOSEN_BUDDY_IDS)
                presenter.invitedMembersUpdated(buddyList)
                members.setText(presenter.getMembersText())
            }
        }
    }

    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        val type = event.type as ChatRoomConfPresenter.Events
        Log.d(TAG, "OnPresenterEvent ${type.name}")
        when (type) {
            Events.GO_TO_CHAT_SCREEN -> {
                coordinator.flow().withBundle(event.data as Bundle).goTo(EScreenList.CONVERSATION)
            }
            ChatRoomConfPresenter.Events.INVALIDATE_MENU -> TODO()
            ChatRoomConfPresenter.Events.UPDATE_INITIALS -> {
                initials.text = event.data as String
            }
            ChatRoomConfPresenter.Events.ROOM_NAME_EMPTY -> {
                nameContainer.error = getString(R.string.tChatRoomEmptyError)
            }
            ChatRoomConfPresenter.Events.ROOM_NAME_DUPLICATED -> {
                nameContainer.error = getString(R.string.tChatRoomNameDuplicated)
            }
            ChatRoomConfPresenter.Events.ROOM_NAME_TOO_LONG -> {
                nameContainer.error = getString(R.string.tChatRoomNameTooLong, ChatRoomConfPresenter.MAX_ROOM_NAME_SIZE)
            }
        }
    }

    override fun createDialogForResult(whichScreen: IScreenEnum, data: Bundle): ScreenHolderDialog {
        if (whichScreen === EScreenList.CHAT_ROOMS_MEMBERS_PICKER_SCREEN) {
            if (!isInTabletMode) {
                return ScreenHolderDialog
                    .builder(activity)
                    .screen(whichScreen)
                    .style(ScreenHolderDialog.DIALOG_MAX_SIZE_NO_MARGIN)
                    .bundle(data)
                    .build()
            } else {
                val color = Coloring.get().decodeColor(presenter.toolbarColor)
                return ScreenHolderDialog
                    .builder(activity)
                    .screen(whichScreen)
                    .anchorView(toolbar!!.findViewById(R.id.conversation_list_new))
                    .gravity(Gravity.BOTTOM)
                    .arrowColor(color)
                    .backgroundColor(color)
                    .bundle(data)
                    .build()
            }
        }

        return super.createDialogForResult(whichScreen, data)
    }

    @Suppress("UNCHECKED_CAST")
    override fun getPresenterClass(): Class<Presenter> = ChatRoomConfPresenter::class.java as Class<Presenter>

    override fun getTitle(): String? = presenter.title

}

fun EditText.afterChange(cb: (String) -> Unit) {
    this.addTextChangedListener(object: TextWatcher {
        override fun afterTextChanged(s: Editable?) { cb(s.toString()) }
        override fun beforeTextChanged(s: CharSequence?, start: Int, count: Int, after: Int) {}
        override fun onTextChanged(s: CharSequence?, start: Int, before: Int, count: Int) {}
    })
}
