package com.bria.voip.ui.main.im.chatroom

import android.support.v7.widget.RecyclerView
import android.view.View
import android.widget.Button
import android.widget.ImageView
import android.widget.TextView
import com.bria.common.R
import com.bria.common.controller.contacts.ContactNamesArrangement
import com.bria.common.controller.contacts.ContactsSortBy
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.ISettingsObserver
import com.bria.common.controller.settings.core.Settings
import com.bria.common.ui.util.ExpandableAreaController
import com.bria.common.ui.util.SearchQueryColorizer
import com.bria.common.uiframework.coloring.Coloring
import com.bria.common.uiframework.lists.recycler.AbstractMultiSelectRecyclerAdapter
import com.bria.common.uiframework.lists.recycler.AbstractRecyclerAdapter
import com.bria.common.uireusable.dataprovider.FilterQueryMatcher
import com.bria.common.uireusable.dataprovider.IFilterableDataProvider
import com.bria.voip.ui.main.contacts.ContactUiHelper

class ChatRoomMembersAdapter(
        private val settings: Settings,
        recyclerView: RecyclerView,
        private val actionsHandler: ActionsHandler,
        private val chatRoomMembersDataProvider: ChatRoomMembersDataProvider,
        private val state: State)
    : AbstractMultiSelectRecyclerAdapter<Buddy, ChatRoomMembersAdapter.ViewHolder>(
        recyclerView,
        chatRoomMembersDataProvider,
        R.layout.chat_room_members_list_item) {

    private var searchQueryColorizer = SearchQueryColorizer(Coloring.get().decodeColor(settings.getStr(ESetting.ColorBrandTint)))

    init {
        settings.attachObserver(object : ISettingsObserver {
            override fun onSettingsChanged(changedSettings: MutableSet<ESetting>?) {
                if (changedSettings?.contains(ESetting.ColorBrandTint) == true) {
                    searchQueryColorizer = SearchQueryColorizer(Coloring.get().decodeColor(settings.getStr(ESetting.ColorBrandTint)))
                }

                notifyDataChanged()
            }

            override fun onOwnerChanged() {}
        }, setOf(ESetting.ContactSortOrder, ESetting.ContactDisplayOrder, ESetting.ColorBrandTint))

        chatRoomMembersDataProvider.attachListener(object : IFilterableDataProvider.DataProviderListener {
            override fun onDataLoaded(isEmpty: Boolean) {
                notifyDataChanged()
            }

            override fun onDataLoading() {}
        })
    }

    interface ActionsHandler {
        fun onCallClicked(buddy: Buddy)
        fun onMessageClicked(buddy: Buddy)
        fun onRemoveClicked(buddy: Buddy)
    }

    inner class ViewHolder internal constructor(itemView: View)
        : AbstractRecyclerAdapter.AbstractViewHolder(itemView) {

        private val avatar: ImageView = itemView.findViewById(R.id.chat_room_members_item_avatar)
        private val presence: ImageView = itemView.findViewById(R.id.chat_room_members_item_presence_image)

        private val name1: TextView = itemView.findViewById(R.id.chat_room_members_item_name1)
        private val nameSpacer: TextView = itemView.findViewById(R.id.chat_room_members_item_name_spacer)
        private val name2: TextView = itemView.findViewById(R.id.chat_room_members_item_name2)

        private val firstName: TextView
        private val lastName: TextView

        private val moreArrow: View = itemView.findViewById(R.id.chat_room_members_item_more)
        private val moreContainer: View = itemView.findViewById(R.id.chat_room_members_item_more_container)

        private val expandableAreaController = ExpandableAreaController(moreArrow, moreContainer, ::onExpandChanged)

        private val btnCall: Button = itemView.findViewById(R.id.chat_room_members_item_action_call)
        private val btnMessage: Button = itemView.findViewById(R.id.chat_room_members_item_action_message)
        private val btnRemove: Button = itemView.findViewById(R.id.chat_room_members_item_action_remove)

        private var buddy: Buddy? = null

        init {
            val displayFirstNameFirst = ContactNamesArrangement.fromRawIntValue(settings.getInt(ESetting.ContactDisplayOrder)) == ContactNamesArrangement.FirstNameThenLastName

            firstName = if (displayFirstNameFirst) name1 else name2
            lastName = if (!displayFirstNameFirst) name1 else name2

            btnCall.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onCallClicked(buddy!!)
                }
            }

            btnMessage.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onMessageClicked(buddy!!)
                }
            }

            btnRemove.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onRemoveClicked(buddy!!)
                }
            }
        }

        fun bindToModel(buddy: Buddy) {
            this.buddy = buddy

            avatar.setImageBitmap(buddy.avatarIcon)
            presence.setImageDrawable(buddy.getPresenceIcon(presence.context))

            firstName.text = buddy.firstName
            lastName.text = buddy.lastName

            // if first info is empty, there will be a margin issue and the thing will look weird
            if (name1.text.isNullOrEmpty()) {
                name1.visibility = View.GONE
                nameSpacer.visibility = View.GONE
            } else {
                name1.visibility = View.VISIBLE
                nameSpacer.visibility = View.VISIBLE
            }

            ContactUiHelper.signifySortOrderWithTypeface(firstName, lastName, ContactsSortBy.fromRawIntValue(settings.getInt(ESetting.ContactSortOrder)))

            val filterQueryMatcher = FilterQueryMatcher(chatRoomMembersDataProvider.searchQuery)
            searchQueryColorizer.colorize(name1, filterQueryMatcher.queryTokens)
            searchQueryColorizer.colorize(name2, filterQueryMatcher.queryTokens)


            expandableAreaController.setExpanded(state.expandedItems.contains(buddy))
        }

        fun onExpandChanged(isExpanded: Boolean) {
            if (buddy == null) {
                return
            }

            if (isExpanded) {
                state.expandedItems.add(buddy!!)
            } else {
                state.expandedItems.remove(buddy!!)
            }
        }
    }

    override fun provideNewViewHolder(inflatedView: View, viewType: Int): ViewHolder {
        return ViewHolder(inflatedView)
    }

    override fun onNewViewHolderReady(viewHolder: ViewHolder, viewType: Int) {}

    override fun onUpdateNeeded(viewHolder: ViewHolder, position: Int) {
        val buddy = mDataProvider.getItemAt(position) as Buddy?
        if (buddy == null) {
            return
        }

        viewHolder.bindToModel(buddy)
    }

    class State {
        val expandedItems: MutableSet<Buddy> = mutableSetOf()
    }
}
