package com.bria.voip.ui.main.im.chatroom

import android.content.Context
import android.database.DataSetObservable
import android.database.DataSetObserver
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.*
import com.bria.common.controller.contacts.ContactNamesArrangement
import com.bria.common.controller.contacts.ContactsSortBy
import com.bria.common.controller.contacts.buddy.*
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.ISettingsObserver
import com.bria.common.controller.settings.core.Settings
import com.bria.common.util.AndroidUtils
import com.bria.common.util.IObservable
import com.bria.voip.R

class ChatRoomMentionsAdapter(
        private val context: Context,
        private val settings: Settings,
        private val buddyController: IBuddyCtrlEvents,
        buddyObservable: IObservable<IBuddyCtrlObserver>
) : ListAdapter, Filterable {

    private var namesArrangement: ContactNamesArrangement

    init {
        namesArrangement = ContactNamesArrangement.fromRawIntValue(settings.getInt(ESetting.ContactDisplayOrder))

        settings.attachObserver(
                object : ISettingsObserver {
                    override fun onSettingsChanged(changedSettings: MutableSet<ESetting>?) {
                        namesArrangement = ContactNamesArrangement.fromRawIntValue(settings.getInt(ESetting.ContactDisplayOrder))
                        notifyDataSetChanged()
                    }

                    override fun onOwnerChanged() {}
                },
                setOf(ESetting.ContactDisplayOrder))

        buddyObservable.attachObserver(object : IBuddyCtrlObserver {
            override fun onBuddyListUpdated() {
                notifyDataSetChanged()
            }

            override fun onBuddyPresenceChanged(buddy: Buddy) {
                val isDisplayed = mData.any { it.buddy == buddy }
                if (isDisplayed) {
                    notifyDataSetChanged()
                }
            }

            override fun onBuddyFilterChanged(aBuddyFilterType: IBuddyCtrlEvents.EBuddyFilterType?) {}
            override fun onSubscriptionRequest() {}
        })
    }

    private var buddyKeys: Set<String> = setOf()

    fun setBuddyKeys(buddyKeys: Set<String>) {
        this.buddyKeys = buddyKeys;

        mUnfilteredData = null

        notifyDataSetChanged()
    }


    override fun getView(position: Int, convertView: View?, parent: ViewGroup): View {
        val view = convertView ?: LayoutInflater.from(context).inflate(R.layout.chat_room_mention_item, parent, false)

        val item = getItem(position) as ChatRoomMentionItem

        bind(view, item)

        return view
    }

    private fun bind(view: View, item: ChatRoomMentionItem) {
        val buddyContainer = view.findViewById<View>(R.id.buddy_container)
        val atContainer = view.findViewById<View>(R.id.at_container)

        if (item.buddy != null) {
            buddyContainer.visibility = View.VISIBLE
            atContainer.visibility = View.GONE

            bindBuddy(view, item.buddy)
        } else {
            buddyContainer.visibility = View.GONE
            atContainer.visibility = View.VISIBLE

            bindAtItem(view, item.atItem!!)
        }
    }

    private fun bindBuddy(view: View, buddy: XmppBuddy) {
        val avatar = view.findViewById<ImageView>(R.id.avatar)
        val name1 = view.findViewById<TextView>(R.id.name1)
        val nameSpacer = view.findViewById<TextView>(R.id.name_spacer)
        val name2 = view.findViewById<TextView>(R.id.name2)
        val presence = view.findViewById<ImageView>(R.id.presence_image)

        if (buddy.avatarIcon != null) {
            avatar.setImageBitmap(buddy.avatarIcon)
        } else {
            avatar.setImageDrawable(AndroidUtils.getDrawable(context, R.drawable.default_avatar))
        }

        val firstName = if (namesArrangement == ContactNamesArrangement.FirstNameThenLastName) name1 else name2
        val lastName = if (namesArrangement == ContactNamesArrangement.LastNameThenFirstName) name1 else name2

        firstName.text = buddy.firstName
        lastName.text = buddy.lastName

        // if first info is empty, there will be a margin issue and the thing will look weird
        if (name1.text.isNullOrEmpty()) {
            name1.visibility = View.GONE
            nameSpacer.visibility = View.GONE
        } else {
            name1.visibility = View.VISIBLE
            nameSpacer.visibility = View.VISIBLE
        }

        presence.setImageDrawable(buddy.getPresenceIcon(context))
    }

    private fun bindAtItem(view: View, atItem: ChatRoomMentionAtItem) {
        val description = view.findViewById<TextView>(R.id.notify_whom_text)
        description.setText(atItem.description)

        val text = view.findViewById<TextView>(R.id.at_text)
        text.text = atItem.atText
    }


    private var mUnfilteredData: List<ChatRoomMentionItem>? = null

    private fun fetchData(): List<ChatRoomMentionItem>? {
        val buddies = buddyController
            .allBuddies
            .filter { it.buddyType == Buddy.EBuddyType.XMPP }
            .map { it as XmppBuddy }
            .filter { buddyKeys.contains(BuddyKeyUtils.getNewBuddyKey(it)) }
            .sortedWith(BuddyComparator(ContactsSortBy.fromRawIntValue(settings.getInt(ESetting.ContactSortOrder))))
            .take(5)
            .map { ChatRoomMentionItem(it, null) }

        // add @ values to the end
        val result = buddies +
                ChatRoomMentionAtItem
                    .getKnownItems(context)
                    .sortedBy { it.atText    }
                    .map { ChatRoomMentionItem(null, it) }

        return result
    }

    private var mData = listOf<ChatRoomMentionItem>()

    override fun getCount(): Int {
        return mData.size
    }

    override fun isEmpty(): Boolean {
        return mData.isEmpty()
    }

    override fun getItem(position: Int): Any {
        return mData.get(position)
    }

    companion object {
        private fun ChatRoomMentionItem.matches(prefixIncludesAt: CharSequence): Boolean {
            if (this.buddy != null) {
                return buddy.matches(prefixIncludesAt.trimStart('@'));
            } else {
                return this.atItem!!.matches(prefixIncludesAt);
            }
        }

        private fun XmppBuddy.matches(prefixWithoutAt: CharSequence): Boolean {
            return this.firstName.startsWith(prefixWithoutAt, true) ||
                    this.lastName.startsWith(prefixWithoutAt, true) ||
                    this.imAddress.startsWith(prefixWithoutAt, true)
        }

        private fun ChatRoomMentionAtItem.matches(prefixIncludesAt: CharSequence): Boolean {
            return this.atText.startsWith(prefixIncludesAt, true) ||
                    this.description.startsWith(prefixIncludesAt.trim('@'), true)
        }

        fun filter(list: List<ChatRoomMentionItem>, prefix: CharSequence?): List<ChatRoomMentionItem> {
            if (prefix?.startsWith("@") == true) {
                if (prefix == "@") {
                    // just @here, @channel etc.
                    return list.filter { it.atItem != null }
                } else {
                    return list.filter { it.matches(prefix) }
                }
            } else {
                return listOf()
            }
        }
    }

    inner class SimpleFilter : Filter() {

        override fun performFiltering(prefix: CharSequence?): Filter.FilterResults {
            val results = Filter.FilterResults()

            if (mUnfilteredData == null) {
                mUnfilteredData = fetchData()
            }

            val list = filter(mUnfilteredData!!, prefix)

            results.values = list
            results.count = list.size

            return results
        }

        override fun publishResults(constraint: CharSequence?, results: Filter.FilterResults) {
            mData = results.values as List<ChatRoomMentionItem>
            if (results.count > 0) {
                notifyDataSetChanged()
            } else {
                notifyDataSetInvalidated()
            }
        }
    }

    // mandatory implementations, no use case specific logic

    private val mDataSetObservable = DataSetObservable()

    override fun registerDataSetObserver(observer: DataSetObserver?) {
        mDataSetObservable.registerObserver(observer)
    }

    override fun unregisterDataSetObserver(observer: DataSetObserver?) {
        mDataSetObservable.unregisterObserver(observer)
    }

    private val mFilter = SimpleFilter()

    override fun getFilter(): Filter {
        return mFilter
    }


    override fun getItemId(position: Int): Long {
        return position.toLong()
    }

    override fun hasStableIds(): Boolean {
        return false
    }

    override fun areAllItemsEnabled(): Boolean {
        return true
    }

    override fun isEnabled(position: Int): Boolean {
        return true
    }

    fun notifyDataSetChanged() {
        mDataSetObservable.notifyChanged()
    }

    fun notifyDataSetInvalidated() {
        mDataSetObservable.notifyInvalidated()
    }

    override fun getViewTypeCount(): Int {
        return 1
    }

    override fun getItemViewType(position: Int): Int {
        return 0
    }
}
