package com.bria.voip.ui.main.settings.developer.settings2

import android.support.annotation.MainThread
import android.text.TextUtils
import com.bria.common.controller.settings.core.SettingData
import com.bria.common.controller.settings.core.Settings
import com.bria.common.controller.settings.core.storage.ISettingsStorage
import com.bria.common.controller.settings.core.types.SettingType
import com.bria.common.controller.settings.core.upgrade.SettingsUpgradeData
import com.bria.common.controller.settings.core.utils.Benchmark
import com.bria.common.controller.settings.core.utils.SettingsJsonEncoder
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.uireusable.dataprovider.ISimpleDataProvider
import com.bria.common.util.Log
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.rxkotlin.addTo
import io.reactivex.schedulers.Schedulers

class SettingsViewerPresenter : AbstractPresenter() {

    companion object {
        private val TAG = "SettingsViewerPresenter"
    }

    class SettingsViewerListItem() {
        var id: Long = 0
        var name: String = ""
        var type: String = ""
        var value: String = ""
        lateinit var settingData: SettingData

        constructor(setting: SettingData) : this() {
            this.id = setting.id
            this.name = setting.settingStr
            this.type = if (setting.setting != null) setting.setting.type.toString() else setting.dataTypeId
            this.value = if (setting.setting != null && setting.setting.type == SettingType.String()) {
                """"${SettingsJsonEncoder.unquote(setting.data)}""""
            } else {
                setting.data
            }
            this.settingData = setting
        }
    }

    enum class Events : IPresenterEventTypeEnum {
        SETTINGS_LOADING, SETTINGS_LOADED, OWNERS_LOADING, OWNERS_LOADED;

        override fun getType(): IPresenterEventTypeEnum = this
    }

    private lateinit var storage: ISettingsStorage
    var ownerDataItems: List<SettingsViewerListItem> = ArrayList()

    private val disposables = CompositeDisposable()
    private var settingsLoader: Observable<SettingsUpgradeData>? = null
    private var settingsFilterDisposable: Disposable? = null

    var owner: String = ""
        set(value) {
            if (value != field) {
                field = value
                filterSettings()
            }
        }
    var filter: String = ""
        set(value) {
            field = value
            filterSettings()
        }

    val dataProvider: ISimpleDataProvider<SettingsViewerListItem> = object : ISimpleDataProvider<SettingsViewerListItem> {
        override fun getItemsCount(): Int = ownerDataItems.size
        override fun getItemAt(position: Int): SettingsViewerListItem = ownerDataItems[position]
        override fun clean() {}
    }

    @MainThread
    override fun onCreate() {
        super.onCreate()
        storage = Settings.get(context).storage
        loadAllSettings()
    }

    private fun loadAllSettings() {
        disposables.clear()

        // prepare the settings loader observable
        settingsLoader = Observable
            .fromCallable {
                SettingsUpgradeData(storage)
            }
            .subscribeOn(Schedulers.io())
            .replay(1) // replay the last emitted event
            .autoConnect(0, { it.addTo(disposables) }) // auto-start it
    }

    private fun filterSettings() {
        firePresenterEvent(Events.SETTINGS_LOADING)
        firePresenterEvent(Events.OWNERS_LOADING)
        if (settingsLoader == null) {
            loadAllSettings()
        }

        dispose(settingsFilterDisposable)
        settingsFilterDisposable = settingsLoader
            ?.observeOn(AndroidSchedulers.mainThread())
            ?.map {
                // notify that users are loaded
                firePresenterEvent(Events.OWNERS_LOADED, it.owners.toMutableList())
                it
            }
            ?.subscribeOn(Schedulers.computation())
            ?.flatMap {
                // get a stream of OwnerSettingsData
                Observable.fromCallable {
                    it.getOwnerSettings(owner)
                }
            }
            ?.map { it.all.values } // get all settings for this owner, and convert from Map<String, SettingData> to Collection<SettingData>
            ?.flatMap { Observable.fromIterable(it) } // turn collection into a stream
            ?.filter { TextUtils.isEmpty(filter) || it.settingStr.contains(filter, true) } // keep only matching items
            ?.map { SettingsViewerListItem(it) } // convert to our new data item
            ?.toSortedList({ o1, o2 -> o1.name.compareTo(o2.name) }) // sort items by name
            ?.observeOn(AndroidSchedulers.mainThread())
            ?.subscribe({
                firePresenterEvent(Events.SETTINGS_LOADED, it)
            }, {
                Log.e(TAG, "Failed to filter settings on ${Thread.currentThread()}", it)
            })
            ?.addTo(disposables)
    }

    @MainThread
    override fun onDestroy() {
        super.onDestroy()
        disposables.dispose()
    }

    fun refresh() {
        loadAllSettings()
        filterSettings()
    }

    fun benchmark() {
        val benchmark = Benchmark()
        benchmark.startBenchmark(context)
    }

}
