package com.bria.voip.ui.main.settings.developer.settings2

import android.annotation.SuppressLint
import android.app.Dialog
import android.os.Bundle
import android.support.annotation.MainThread
import android.support.v7.app.AlertDialog
import android.support.v7.widget.LinearLayoutManager
import android.support.v7.widget.RecyclerView
import android.text.method.ScrollingMovementMethod
import android.view.MenuItem
import android.view.View
import android.widget.*
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.utils.SettingsLog
import com.bria.common.uiframework.annotations.Clickable
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.voip.R
import com.bria.voip.ui.main.settings.developer.settings2.SettingsViewerPresenter.Events.*
import com.google.gson.GsonBuilder
import com.google.gson.JsonParser
import com.jakewharton.rxbinding2.widget.RxTextView
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.rxkotlin.addTo
import java.util.concurrent.TimeUnit


@Layout(R.layout.settings_viewer_screen)
@com.bria.common.uiframework.annotations.Menu(R.menu.settings_viewer_menu)
class SettingsViewerScreen : AbstractScreen<SettingsViewerPresenter>() {

    companion object {
        val DEFAULT_USERNAME = SettingsLog.getOwnerDisplayStr("")!!
        val DIALOG_SETTING_DETAILS = 0xD001
        val DIALOG_INDEX_KEY = "index_key"
    }

    // <editor-fold desc="Injections">
    @Inject(R.id.settings_viewer_list)
    private lateinit var list: RecyclerView

    @Inject(id = R.id.settings_viewer_loading, fore = ESetting.ColorBrandTint)
    private lateinit var progressBar: ProgressBar

    @Inject(R.id.settings_viewer_filter)
    private lateinit var filterEditBox: EditText

    @Clickable
    @Inject(id = R.id.settings_viewer_clear, fore = ESetting.ColorBlack, back = ESetting.ColorBrandDefault)
    private lateinit var clearButton: ImageView

    @Inject(R.id.settings_viewer_dropdown)
    private lateinit var dropdown: Spinner
    // </editor-fold>

    private lateinit var adapter: SettingsViewerAdapter
    private val disposables = CompositeDisposable()

    @MainThread
    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        val eventType = event.type as SettingsViewerPresenter.Events

        @Suppress("UNCHECKED_CAST")
        when (eventType) {
            SETTINGS_LOADED -> {
                val settings = event.data as List<SettingsViewerPresenter.SettingsViewerListItem>
                presenter.ownerDataItems = settings
                adapter.notifyDataChanged()
                list.visibility = View.VISIBLE
                progressBar.visibility = View.GONE
            }
            SETTINGS_LOADING -> {
                list.visibility = View.GONE
                progressBar.visibility = View.VISIBLE
            }
            OWNERS_LOADING -> dropdown.visibility = View.GONE
            OWNERS_LOADED -> {
                val items = event.data as MutableList<String>
                val selectedOwner = items.indexOf(presenter.owner)
                items[items.indexOf("")] = DEFAULT_USERNAME

                dropdown.onItemSelectedListener = null
                dropdown.adapter = ArrayAdapter<String>(activity, R.layout.settings_viewer_dropdown_item, items)
                dropdown.setSelection(selectedOwner)
                dropdown.onItemSelectedListener = dropdownListener
                dropdown.visibility = View.VISIBLE
            }
        }
    }

    @SuppressLint("RxSubscribeOnError")
    @MainThread
    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)

        adapter = SettingsViewerAdapter(list, R.layout.settings_viewer_list_item)
        adapter.dataProvider = presenter.dataProvider
        list.layoutManager = LinearLayoutManager(activity, LinearLayoutManager.VERTICAL, false)
        list.adapter = adapter
        adapter.setOnItemClickListener { _, index ->
            val subBundle = Bundle()
            subBundle.putInt(DIALOG_INDEX_KEY, index)
            showDialog(DIALOG_SETTING_DETAILS, subBundle)
        }

        filterEditBox.setText(presenter.filter)
        RxTextView
            .afterTextChangeEvents(filterEditBox)
            .distinct()
            .debounce(300, TimeUnit.MILLISECONDS)
            .map { it -> it.editable().toString() }
            .subscribe {
                presenter.filter = it
            }
            .addTo(disposables)

        dropdown.onItemSelectedListener = dropdownListener
    }

    private val dropdownListener: AdapterView.OnItemSelectedListener = object : AdapterView.OnItemSelectedListener {
        override fun onNothingSelected(parent: AdapterView<*>?) {}
        override fun onItemSelected(parent: AdapterView<*>?, view: View?, index: Int, id: Long) {
            val picked = dropdown.adapter.getItem(index) as String
            presenter.owner = if (picked == DEFAULT_USERNAME) "" else picked
        }
    }

    @MainThread
    override fun onStart(bundle: Bundle?) {
        super.onStart(bundle)
        presenter.subscribe(this)
        presenter.filter = filterEditBox.text.toString()
    }

    @MainThread
    override fun onClick(v: View) {
        super.onClick(v)
        when (v.id) {
            R.id.settings_viewer_clear -> filterEditBox.setText("")
        }
    }

    @MainThread
    override fun onStop(finishing: Boolean) {
        super.onStop(finishing)
        presenter.unsubscribe()
        disposables.clear()
    }

    @MainThread
    override fun onDestroy(finishing: Boolean) {
        super.onDestroy(finishing)
        disposables.dispose()
        destroyPresenter()
    }

    @MainThread
    override fun createDialog(which: Int, data: Bundle?): Dialog? {
        if (which == DIALOG_SETTING_DETAILS) {
            val builder: AlertDialog.Builder = AlertDialog.Builder(activity)
            val view = View.inflate(activity, R.layout.settings_viewer_dialog, null)
            val settingData = presenter.dataProvider.getItemAt(data?.getInt(DIALOG_INDEX_KEY, 0)!!).settingData
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_name)?.text = settingData.settingStr
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_type)?.text = settingData.dataTypeId
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_owner_type)?.text = settingData.ownerTypeStr
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_owner)?.text =
                    if (settingData.owner.isEmpty()) "<global_owner>" else settingData.owner
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_origin)?.text = settingData.originStr
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_value)?.text = toPrettyJson(settingData.data)
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_value)?.movementMethod = ScrollingMovementMethod()
            view.findViewById<TextView>(R.id.settings_viewer_dialog_setting_id)?.text = settingData.id.toString()
            builder.setView(view)
            return builder.create()
        }
        return super.createDialog(which, data)
    }

    override fun getPresenterClass(): Class<out SettingsViewerPresenter> = SettingsViewerPresenter::class.java

    override fun getTitle(): String = "Settings Viewer"

    private fun toPrettyJson(jsonString: String): String {
        val parser = JsonParser()
        val json = parser.parse(jsonString)
        val gson = GsonBuilder().setPrettyPrinting().create()
        return gson.toJson(json)
    }

    @MainThread
    override fun onMenuItemClick(menuItem: MenuItem?): Boolean {
        if (menuItem?.itemId == R.id.settings_viewer_refresh) {
            presenter.refresh()
        } else if (menuItem?.itemId == R.id.settings_viewer_benchmark) {
            presenter.benchmark()
        }
        return super.onMenuItemClick(menuItem)
    }

}

