package com.bria.voip.ui.shared.pickers

import com.bria.common.controller.ClientConfig
import com.bria.common.controller.Controllers
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents.EBuddyFilterType
import com.bria.common.uireusable.dataprovider.AbstractFilterableListDataProvider
import com.bria.common.util.Log
import java.util.*

class ChatRoomMembersPickerProvider : AbstractFilterableListDataProvider<Buddy>() {

    /**
     * The presence status filter type - we're either searching for online or all buddies. Requests are not supported.
     *
     * @see EBuddyFilterType
     */
    /**
     * Sets for which availabilities we want to filter the list of buddies: all or just online.
     *
     * @param availability A [EBuddyFilterType]. Must not be Requests, as that's not supported, and it will throw an exception.
     */
    var onlineStatus = EBuddyFilterType.ALL
        set(newValue) {
            if (newValue === EBuddyFilterType.REQUESTS) {
                throw IllegalArgumentException("Requests are not supported! They shouldn't even be a part of EBuddyFilterType!")
            }

            field = newValue

            refilter()
        }

    /**
     * List of buddy IDs from buddies that are already active room members.
     */
    private var mAlreadyPresentBuddyIds = setOf<String>()

    override fun provideFullDataSet(): List<Buddy> {

        val buddies = Controllers.get().buddy.getListOfBuddies(Buddy.EBuddyType.XMPP, IBuddyCtrlEvents.EBuddyGroupType.ALL, onlineStatus)

        val iterator = buddies.iterator()
        while (iterator.hasNext()) {
            val buddy = iterator.next()

            if (mAlreadyPresentBuddyIds.contains(buddy.uniqueIdentifier)) {
                iterator.remove()
            }
        }

        return buddies
    }

    override fun provideFilteredDataSet(fullDataSet: List<Buddy>, query: String): List<Buddy> {
        // TODO: This should be extracted into a Utils method, as it can be used for any list and query.
        val filteredList = ArrayList<Buddy>()

        for (buddy in fullDataSet) {
            // Tokenize all the names and the query by
            // splitting them by whitespaces
            val nameTokens = getNameTokens(buddy)
            val queryTokens = getQueryTokens(query)

            var allMatched = true
            for (q in queryTokens) {
                var matched = false
                for (name in nameTokens) {
                    if (name.startsWith(q)) {
                        matched = true
                        break
                    }
                }

                // If any of the queries wasn't matched, the
                // buddy does not fit the filter criteria.
                if (!matched) {
                    allMatched = false
                    break
                }
            }

            if (allMatched) {
                filteredList.add(buddy)
            }
        }

        return filteredList
    }

    /**
     * Retrieves the first and last name(s) of a buddy in a list of strings.
     *
     * @param buddy The buddy whose first and last name(s) we are to extract.
     *
     * @return A [<] containing all the names of a buddy. The separation is being done by all whitespace characters: blank space, tab, etc.
     */
    private fun getNameTokens(buddy: Buddy): List<String> {
        val splitFirstNames = buddy.firstName.toLowerCase().split("\\s+".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()
        val splitLastNames = buddy.lastName.toLowerCase().split("\\s+".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()

        val nameTokens = ArrayList<String>()
        nameTokens.addAll(Arrays.asList(*splitFirstNames))
        nameTokens.addAll(Arrays.asList(*splitLastNames))

        return nameTokens
    }

    /**
     * Tokenizes the query
     *
     * @param query The query that the data should be filtered by.
     *
     * @return A [<] containing all the search strings. I.e. from "Buddy Spencer", you would get a list of two elements: ["Buddy", "Spencer"].
     * The separation is being done by all whitespace characters: blank space, tab, etc.
     */
    private fun getQueryTokens(query: String): List<String> {
        val splitQuery = query.toLowerCase().split("\\s+".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()

        return ArrayList(Arrays.asList(*splitQuery))
    }

    /**
     * Print a debug message.
     */
    private fun debug(message: String) {
        if (ClientConfig.get().isDebugMode) {
            Log.d(ChatRoomMembersPickerProvider::class.java.simpleName + "-DEBUG", message)
        }
    }

    /**
     * Sets a list of Buddy IDs as the list of currently selected buddies inside of the picker. The rest of the list will get filtered accordingly.
     *
     * @param alreadyPresentBuddyIds A list of Buddy IDs.
     */
    fun setAlreadyPresentBuddyIds(alreadyPresentBuddyIds: Set<String>) {
        mAlreadyPresentBuddyIds = alreadyPresentBuddyIds
        refilter()
    }
}
